/* Generated by Cython 0.29.28 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/core/include"
        ],
        "name": "biosaur2.cutils",
        "sources": [
            "biosaur2/cutils.pyx"
        ]
    },
    "module_name": "biosaur2.cutils"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_28"
#define CYTHON_HEX_VERSION 0x001D1CF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030B00A1)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__biosaur2__cutils
#define __PYX_HAVE_API__biosaur2__cutils
/* Early includes */
#include <omp.h>
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "biosaur2/cutils.pyx",
  "__init__.pxd",
  "type.pxd",
};

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr;
struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "biosaur2/cutils.pyx":30
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 */
struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr {
  PyObject_HEAD
  PyObject *__pyx_v_hill_idict_1;
};


/* "biosaur2/cutils.pyx":41
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
 *     if hill_sqrt_of_i_1 is None:
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))             # <<<<<<<<<<<<<<
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1
 * 
 */
struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *__pyx_outer_scope;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_FloorDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceFloorDivide(op1, op2) : PyNumber_FloorDivide(op1, op2))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'openmp' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'biosaur2.cutils' */
static PyTypeObject *__pyx_ptype_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr = 0;
static PyTypeObject *__pyx_ptype_8biosaur2_6cutils___pyx_scope_struct_1_genexpr = 0;
static PyObject *__pyx_v_8biosaur2_6cutils_charge_ban_map = 0;
#define __Pyx_MODULE_NAME "biosaur2.cutils"
extern int __pyx_module_is_main_biosaur2__cutils;
int __pyx_module_is_main_biosaur2__cutils = 0;

/* Implementation of 'biosaur2.cutils' */
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_a[] = "a";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_z[] = "z";
static const char __pyx_k_fi[] = "fi";
static const char __pyx_k_fm[] = "fm";
static const char __pyx_k_i1[] = "i1";
static const char __pyx_k_i2[] = "i2";
static const char __pyx_k_im[] = "im";
static const char __pyx_k_np[] = "np";
static const char __pyx_k__16[] = "_";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_dia[] = "dia";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_hvf[] = "hvf";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_l_r[] = "l_r";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_r_r[] = "r_r";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_top[] = "top";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_ch_v[] = "ch_v";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "dict";
static const char __pyx_k_htol[] = "htol";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_numb[] = "numb";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sqrt[] = "sqrt";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_FAIMS[] = "FAIMS";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_c_len[] = "c_len";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_flag1[] = "flag1";
static const char __pyx_k_flag2[] = "flag2";
static const char __pyx_k_flag3[] = "flag3";
static const char __pyx_k_idx_1[] = "idx_1";
static const char __pyx_k_idx_2[] = "idx_2";
static const char __pyx_k_l_idx[] = "l_idx";
static const char __pyx_k_minlh[] = "minlh";
static const char __pyx_k_nproc[] = "nproc";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ready[] = "ready";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_bottom[] = "bottom";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_cumsum[] = "cumsum";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_len_mz[] = "len_mz";
static const char __pyx_k_ma_vec[] = "ma_vec";
static const char __pyx_k_mz_cur[] = "mz_cur";
static const char __pyx_k_mz_tol[] = "mz_tol";
static const char __pyx_k_thresh[] = "thresh";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Counter[] = "Counter";
static const char __pyx_k_all_idx[] = "all_idx";
static const char __pyx_k_argsort[] = "argsort";
static const char __pyx_k_average[] = "average";
static const char __pyx_k_charges[] = "charges";
static const char __pyx_k_cos_cor[] = "cos_cor";
static const char __pyx_k_diahtol[] = "diahtol";
static const char __pyx_k_end_idx[] = "end_idx";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_idx_end[] = "idx_end";
static const char __pyx_k_im_mz_1[] = "im_mz_1";
static const char __pyx_k_im_mz_2[] = "im_mz_2";
static const char __pyx_k_int_val[] = "int_val";
static const char __pyx_k_min_idx[] = "min_idx";
static const char __pyx_k_min_val[] = "min_val";
static const char __pyx_k_mz_step[] = "mz_step";
static const char __pyx_k_product[] = "product";
static const char __pyx_k_tmp_val[] = "tmp_val";
static const char __pyx_k_weights[] = "weights";
static const char __pyx_k_best_cor[] = "best_cor";
static const char __pyx_k_best_pos[] = "best_pos";
static const char __pyx_k_cos_corr[] = "cos_corr";
static const char __pyx_k_diaminlh[] = "diaminlh";
static const char __pyx_k_hill_idx[] = "hill_idx";
static const char __pyx_k_idx_minl[] = "idx_minl";
static const char __pyx_k_idx_prev[] = "idx_prev";
static const char __pyx_k_idx_sort[] = "idx_sort";
static const char __pyx_k_im_array[] = "im_array";
static const char __pyx_k_isotopes[] = "isotopes";
static const char __pyx_k_last_idx[] = "last_idx";
static const char __pyx_k_meanfilt[] = "meanfilt";
static const char __pyx_k_mult_val[] = "mult_val";
static const char __pyx_k_paseftol[] = "paseftol";
static const char __pyx_k_prev_idx[] = "prev_idx";
static const char __pyx_k_scan_val[] = "scan_val";
static const char __pyx_k_spec_idx[] = "spec_idx";
static const char __pyx_k_suit_len[] = "suit_len";
static const char __pyx_k_all_prevs[] = "all_prevs";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_faims_val[] = "faims_val";
static const char __pyx_k_fast_dict[] = "fast_dict";
static const char __pyx_k_hill_mz_1[] = "hill_mz_1";
static const char __pyx_k_hill_mz_2[] = "hill_mz_2";
static const char __pyx_k_i_sum_tmp[] = "i_sum_tmp";
static const char __pyx_k_i_val_tmp[] = "i_val_tmp";
static const char __pyx_k_idx_start[] = "idx_start";
static const char __pyx_k_im_median[] = "im_median";
static const char __pyx_k_im_sorted[] = "im_sorted";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_m_z_array[] = "m/z array";
static const char __pyx_k_mz_median[] = "mz_median";
static const char __pyx_k_mz_sorted[] = "mz_sorted";
static const char __pyx_k_mzs_array[] = "mzs_array";
static const char __pyx_k_nIsotopes[] = "nIsotopes";
static const char __pyx_k_tmp_scans[] = "tmp_scans";
static const char __pyx_k_candidates[] = "candidates";
static const char __pyx_k_checked_id[] = "checked_id";
static const char __pyx_k_cos_cor_RT[] = "cos_cor_RT";
static const char __pyx_k_cumsum_vec[] = "cumsum_vec";
static const char __pyx_k_fast_array[] = "fast_array";
static const char __pyx_k_hill_idx_1[] = "hill_idx_1";
static const char __pyx_k_hill_idx_2[] = "hill_idx_2";
static const char __pyx_k_hills_dict[] = "hills_dict";
static const char __pyx_k_m_to_check[] = "m_to_check";
static const char __pyx_k_max_charge[] = "max_charge";
static const char __pyx_k_min_charge[] = "min_charge";
static const char __pyx_k_mz_val_tmp[] = "mz_val_tmp";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_cur_new_idx[] = "cur_new_idx";
static const char __pyx_k_defaultdict[] = "defaultdict";
static const char __pyx_k_hill_length[] = "hill_length";
static const char __pyx_k_hills_idict[] = "hills_idict";
static const char __pyx_k_isotope_idx[] = "isotope_idx";
static const char __pyx_k_recheck_idx[] = "recheck_idx";
static const char __pyx_k_recheck_r_r[] = "recheck_r_r";
static const char __pyx_k_scan_id_val[] = "scan_id_val";
static const char __pyx_k_split_peaks[] = "split_peaks";
static const char __pyx_k_detect_hills[] = "detect_hills";
static const char __pyx_k_fast_dict_im[] = "fast_dict_im";
static const char __pyx_k_hill_idict_1[] = "hill_idict_1";
static const char __pyx_k_hill_idict_2[] = "hill_idict_2";
static const char __pyx_k_hill_scans_1[] = "hill_scans_1";
static const char __pyx_k_hill_scans_2[] = "hill_scans_2";
static const char __pyx_k_idx_for_sort[] = "idx_for_sort";
static const char __pyx_k_intersection[] = "intersection";
static const char __pyx_k_ma_vec_array[] = "ma_vec_array";
static const char __pyx_k_min_idx_list[] = "min_idx_list";
static const char __pyx_k_neutral_mass[] = "neutral_mass";
static const char __pyx_k_orig_idx_val[] = "orig_idx_val";
static const char __pyx_k_tmp_im_array[] = "tmp_im_array";
static const char __pyx_k_tmp_mz_array[] = "tmp_mz_array";
static const char __pyx_k_tmp_orig_idx[] = "tmp_orig_idx";
static const char __pyx_k_window_width[] = "window_width";
static const char __pyx_k_best_idx_prev[] = "best_idx_prev";
static const char __pyx_k_cur_intensity[] = "cur_intensity";
static const char __pyx_k_cur_mass_diff[] = "cur_mass_diff";
static const char __pyx_k_fast_array_im[] = "fast_array_im";
static const char __pyx_k_get_fast_dict[] = "get_fast_dict";
static const char __pyx_k_hills_lengths[] = "hills_lengths";
static const char __pyx_k_im_median_int[] = "im_median_int";
static const char __pyx_k_intensity_val[] = "intensity_val";
static const char __pyx_k_isotopes_list[] = "isotopes_list";
static const char __pyx_k_mass_diff_abs[] = "mass_diff_abs";
static const char __pyx_k_mass_diff_ppm[] = "mass_diff_ppm";
static const char __pyx_k_mz_median_int[] = "mz_median_int";
static const char __pyx_k_process_hills[] = "process_hills";
static const char __pyx_k_tmp_intensity[] = "tmp_intensity";
static const char __pyx_k_tmp_scans_set[] = "tmp_scans_set";
static const char __pyx_k_banned_charges[] = "banned_charges";
static const char __pyx_k_best_intensity[] = "best_intensity";
static const char __pyx_k_best_mass_diff[] = "best_mass_diff";
static const char __pyx_k_hill_sqrt_of_i[] = "hill_sqrt_of_i";
static const char __pyx_k_include_factor[] = "include_factor";
static const char __pyx_k_isotope_number[] = "isotope_number";
static const char __pyx_k_local_res_dict[] = "local_res_dict";
static const char __pyx_k_new_index_list[] = "new_index_list";
static const char __pyx_k_orig_idx_array[] = "orig_idx_array";
static const char __pyx_k_prev_fast_dict[] = "prev_fast_dict";
static const char __pyx_k_scan_idx_array[] = "scan_idx_array";
static const char __pyx_k_tmp_candidates[] = "tmp_candidates";
static const char __pyx_k_tmp_scans_list[] = "tmp_scans_list";
static const char __pyx_k_basic_id_sorted[] = "basic_id_sorted";
static const char __pyx_k_biosaur2_cutils[] = "biosaur2.cutils";
static const char __pyx_k_cos_correlation[] = "cos_correlation";
static const char __pyx_k_hills_idx_array[] = "hills_idx_array";
static const char __pyx_k_hills_im_median[] = "hills_im_median";
static const char __pyx_k_hills_mz_median[] = "hills_mz_median";
static const char __pyx_k_hills_scan_apex[] = "hills_scan_apex";
static const char __pyx_k_hills_scan_sets[] = "hills_scan_sets";
static const char __pyx_k_intensity_array[] = "intensity array";
static const char __pyx_k_iter_candidates[] = "iter_candidates";
static const char __pyx_k_m_to_check_fast[] = "m_to_check_fast";
static const char __pyx_k_min_length_hill[] = "min_length_hill";
static const char __pyx_k_monoisotope_idx[] = "monoisotope idx";
static const char __pyx_k_split_peaks_old[] = "split_peaks_old";
static const char __pyx_k_theor_total_sum[] = "theor_total_sum";
static const char __pyx_k_tmp_hill_length[] = "tmp_hill_length";
static const char __pyx_k_total_mass_diff[] = "total_mass_diff";
static const char __pyx_k_total_num_hills[] = "total_num_hills";
static const char __pyx_k_cos_cor_isotopes[] = "cos_cor_isotopes";
static const char __pyx_k_hillValleyFactor[] = "hillValleyFactor";
static const char __pyx_k_hill_scan_apex_1[] = "hill_scan_apex_1";
static const char __pyx_k_hill_scan_apex_2[] = "hill_scan_apex_2";
static const char __pyx_k_hill_sqrt_of_i_1[] = "hill_sqrt_of_i_1";
static const char __pyx_k_hill_sqrt_of_i_2[] = "hill_sqrt_of_i_2";
static const char __pyx_k_hills_scan_lists[] = "hills_scan_lists";
static const char __pyx_k_im_to_check_fast[] = "im_to_check_fast";
static const char __pyx_k_isotope_hill_idx[] = "isotope_hill_idx";
static const char __pyx_k_theoretical_list[] = "theoretical_list";
static const char __pyx_k_all_exp_intensity[] = "all_exp_intensity";
static const char __pyx_k_counter_hills_idx[] = "counter_hills_idx";
static const char __pyx_k_experimental_list[] = "experimental_list";
static const char __pyx_k_hill_scans_1_list[] = "hill_scans_1_list";
static const char __pyx_k_hill_scans_2_list[] = "hill_scans_2_list";
static const char __pyx_k_intensity_array_2[] = "intensity_array";
static const char __pyx_k_prev_fast_dict_im[] = "prev_fast_dict_im";
static const char __pyx_k_smothed_intensity[] = "smothed_intensity";
static const char __pyx_k_averagineExplained[] = "averagineExplained";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_hill_mass_accuracy[] = "hill_mass_accuracy";
static const char __pyx_k_all_theoretical_int[] = "all_theoretical_int";
static const char __pyx_k_banned_prev_idx_set[] = "banned_prev_idx_set";
static const char __pyx_k_biosaur2_cutils_pyx[] = "biosaur2/cutils.pyx";
static const char __pyx_k_cos_correlation_new[] = "cos_correlation_new";
static const char __pyx_k_local_isotopes_dict[] = "local_isotopes_dict";
static const char __pyx_k_averagineCorrelation[] = "averagineCorrelation";
static const char __pyx_k_data_for_analyse_tmp[] = "data_for_analyse_tmp";
static const char __pyx_k_get_initial_isotopes[] = "get_initial_isotopes";
static const char __pyx_k_hills_intensity_apex[] = "hills_intensity_apex";
static const char __pyx_k_monoisotope_hill_idx[] = "monoisotope hill idx";
static const char __pyx_k_hill_intensity_apex_1[] = "hill_intensity_apex_1";
static const char __pyx_k_hill_intensity_apex_2[] = "hill_intensity_apex_2";
static const char __pyx_k_hills_intensity_array[] = "hills_intensity_array";
static const char __pyx_k_mono_hills_scan_lists[] = "mono_hills_scan_lists";
static const char __pyx_k_hill_scans_1_list_last[] = "hill_scans_1_list_last";
static const char __pyx_k_hill_scans_2_list_last[] = "hill_scans_2_list_last";
static const char __pyx_k_hills_idx_array_unique[] = "hills_idx_array_unique";
static const char __pyx_k_isotopes_mass_accuracy[] = "isotopes_mass_accuracy";
static const char __pyx_k_cur_mass_diff_with_sign[] = "cur_mass_diff_with_sign";
static const char __pyx_k_hill_scans_1_list_first[] = "hill_scans_1_list_first";
static const char __pyx_k_hill_scans_2_list_first[] = "hill_scans_2_list_first";
static const char __pyx_k_spec_mean_mass_accuracy[] = "spec_mean_mass_accuracy";
static const char __pyx_k_sorted_idx_child_process[] = "sorted_idx_child_process";
static const char __pyx_k_hills_im_median_fast_dict[] = "hills_im_median_fast_dict";
static const char __pyx_k_hills_mz_median_fast_dict[] = "hills_mz_median_fast_dict";
static const char __pyx_k_mono_hills_intensity_list[] = "mono_hills_intensity_list";
static const char __pyx_k_number_of_passed_isotopes[] = "number_of_passed_isotopes";
static const char __pyx_k_intensity_array_for_cos_corr[] = "intensity_array_for_cos_corr";
static const char __pyx_k_sorted_idx_array_child_process[] = "sorted_idx_array_child_process";
static const char __pyx_k_get_and_calc_apex_intensity_and[] = "get_and_calc_apex_intensity_and_scan";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_checking_cos_correlation_for_car[] = "checking_cos_correlation_for_carbon";
static const char __pyx_k_get_and_calc_values_for_cos_corr[] = "get_and_calc_values_for_cos_corr.<locals>.genexpr";
static const char __pyx_k_mean_inverse_reduced_ion_mobilit[] = "mean inverse reduced ion mobility array";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_get_and_calc_values_for_cos_corr_2[] = "get_and_calc_values_for_cos_corr";
static PyObject *__pyx_n_s_Counter;
static PyObject *__pyx_n_s_FAIMS;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s__16;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_all_exp_intensity;
static PyObject *__pyx_n_s_all_idx;
static PyObject *__pyx_n_s_all_prevs;
static PyObject *__pyx_n_s_all_theoretical_int;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_argsort;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_average;
static PyObject *__pyx_n_s_averagineCorrelation;
static PyObject *__pyx_n_s_averagineExplained;
static PyObject *__pyx_n_s_banned_charges;
static PyObject *__pyx_n_s_banned_prev_idx_set;
static PyObject *__pyx_n_s_basic_id_sorted;
static PyObject *__pyx_n_s_best_cor;
static PyObject *__pyx_n_s_best_idx_prev;
static PyObject *__pyx_n_s_best_intensity;
static PyObject *__pyx_n_s_best_mass_diff;
static PyObject *__pyx_n_s_best_pos;
static PyObject *__pyx_n_s_biosaur2_cutils;
static PyObject *__pyx_kp_s_biosaur2_cutils_pyx;
static PyObject *__pyx_n_s_bottom;
static PyObject *__pyx_n_s_c_len;
static PyObject *__pyx_n_s_candidates;
static PyObject *__pyx_n_s_ch_v;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_charges;
static PyObject *__pyx_n_s_checked_id;
static PyObject *__pyx_n_s_checking_cos_correlation_for_car;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_cos_cor;
static PyObject *__pyx_n_s_cos_cor_RT;
static PyObject *__pyx_n_s_cos_cor_isotopes;
static PyObject *__pyx_n_s_cos_corr;
static PyObject *__pyx_n_s_cos_correlation;
static PyObject *__pyx_n_s_cos_correlation_new;
static PyObject *__pyx_n_s_counter_hills_idx;
static PyObject *__pyx_n_s_cumsum;
static PyObject *__pyx_n_s_cumsum_vec;
static PyObject *__pyx_n_s_cur_intensity;
static PyObject *__pyx_n_s_cur_mass_diff;
static PyObject *__pyx_n_s_cur_mass_diff_with_sign;
static PyObject *__pyx_n_s_cur_new_idx;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_for_analyse_tmp;
static PyObject *__pyx_n_s_defaultdict;
static PyObject *__pyx_n_s_detect_hills;
static PyObject *__pyx_n_s_dia;
static PyObject *__pyx_n_s_diahtol;
static PyObject *__pyx_n_s_diaminlh;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_end_idx;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_experimental_list;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_faims_val;
static PyObject *__pyx_n_s_fast_array;
static PyObject *__pyx_n_s_fast_array_im;
static PyObject *__pyx_n_s_fast_dict;
static PyObject *__pyx_n_s_fast_dict_im;
static PyObject *__pyx_n_s_fi;
static PyObject *__pyx_n_s_flag1;
static PyObject *__pyx_n_s_flag2;
static PyObject *__pyx_n_s_flag3;
static PyObject *__pyx_n_s_fm;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_and_calc_apex_intensity_and;
static PyObject *__pyx_n_s_get_and_calc_values_for_cos_corr;
static PyObject *__pyx_n_s_get_and_calc_values_for_cos_corr_2;
static PyObject *__pyx_n_s_get_fast_dict;
static PyObject *__pyx_n_s_get_initial_isotopes;
static PyObject *__pyx_n_s_hillValleyFactor;
static PyObject *__pyx_n_s_hill_idict_1;
static PyObject *__pyx_n_s_hill_idict_2;
static PyObject *__pyx_n_s_hill_idx;
static PyObject *__pyx_n_s_hill_idx_1;
static PyObject *__pyx_n_s_hill_idx_2;
static PyObject *__pyx_n_s_hill_intensity_apex_1;
static PyObject *__pyx_n_s_hill_intensity_apex_2;
static PyObject *__pyx_n_s_hill_length;
static PyObject *__pyx_n_s_hill_mass_accuracy;
static PyObject *__pyx_n_s_hill_mz_1;
static PyObject *__pyx_n_s_hill_mz_2;
static PyObject *__pyx_n_s_hill_scan_apex_1;
static PyObject *__pyx_n_s_hill_scan_apex_2;
static PyObject *__pyx_n_s_hill_scans_1;
static PyObject *__pyx_n_s_hill_scans_1_list;
static PyObject *__pyx_n_s_hill_scans_1_list_first;
static PyObject *__pyx_n_s_hill_scans_1_list_last;
static PyObject *__pyx_n_s_hill_scans_2;
static PyObject *__pyx_n_s_hill_scans_2_list;
static PyObject *__pyx_n_s_hill_scans_2_list_first;
static PyObject *__pyx_n_s_hill_scans_2_list_last;
static PyObject *__pyx_n_s_hill_sqrt_of_i;
static PyObject *__pyx_n_s_hill_sqrt_of_i_1;
static PyObject *__pyx_n_s_hill_sqrt_of_i_2;
static PyObject *__pyx_n_s_hills_dict;
static PyObject *__pyx_n_s_hills_idict;
static PyObject *__pyx_n_s_hills_idx_array;
static PyObject *__pyx_n_s_hills_idx_array_unique;
static PyObject *__pyx_n_s_hills_im_median;
static PyObject *__pyx_n_s_hills_im_median_fast_dict;
static PyObject *__pyx_n_s_hills_intensity_apex;
static PyObject *__pyx_n_s_hills_intensity_array;
static PyObject *__pyx_n_s_hills_lengths;
static PyObject *__pyx_n_s_hills_mz_median;
static PyObject *__pyx_n_s_hills_mz_median_fast_dict;
static PyObject *__pyx_n_s_hills_scan_apex;
static PyObject *__pyx_n_s_hills_scan_lists;
static PyObject *__pyx_n_s_hills_scan_sets;
static PyObject *__pyx_n_s_htol;
static PyObject *__pyx_n_s_hvf;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_i1;
static PyObject *__pyx_n_s_i2;
static PyObject *__pyx_n_s_i_sum_tmp;
static PyObject *__pyx_n_s_i_val_tmp;
static PyObject *__pyx_n_s_idx;
static PyObject *__pyx_n_s_idx_1;
static PyObject *__pyx_n_s_idx_2;
static PyObject *__pyx_n_s_idx_end;
static PyObject *__pyx_n_s_idx_for_sort;
static PyObject *__pyx_n_s_idx_minl;
static PyObject *__pyx_n_s_idx_prev;
static PyObject *__pyx_n_s_idx_sort;
static PyObject *__pyx_n_s_idx_start;
static PyObject *__pyx_n_s_im;
static PyObject *__pyx_n_s_im_array;
static PyObject *__pyx_n_s_im_median;
static PyObject *__pyx_n_s_im_median_int;
static PyObject *__pyx_n_s_im_mz_1;
static PyObject *__pyx_n_s_im_mz_2;
static PyObject *__pyx_n_s_im_sorted;
static PyObject *__pyx_n_s_im_to_check_fast;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_include_factor;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_int_val;
static PyObject *__pyx_kp_s_intensity_array;
static PyObject *__pyx_n_s_intensity_array_2;
static PyObject *__pyx_n_s_intensity_array_for_cos_corr;
static PyObject *__pyx_n_s_intensity_val;
static PyObject *__pyx_n_s_intersection;
static PyObject *__pyx_n_s_isotope_hill_idx;
static PyObject *__pyx_n_s_isotope_idx;
static PyObject *__pyx_n_s_isotope_number;
static PyObject *__pyx_n_s_isotopes;
static PyObject *__pyx_n_s_isotopes_list;
static PyObject *__pyx_n_s_isotopes_mass_accuracy;
static PyObject *__pyx_n_s_iter_candidates;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_l_idx;
static PyObject *__pyx_n_s_l_r;
static PyObject *__pyx_n_s_last_idx;
static PyObject *__pyx_n_s_len_mz;
static PyObject *__pyx_n_s_local_isotopes_dict;
static PyObject *__pyx_n_s_local_res_dict;
static PyObject *__pyx_n_s_m_to_check;
static PyObject *__pyx_n_s_m_to_check_fast;
static PyObject *__pyx_kp_s_m_z_array;
static PyObject *__pyx_n_s_ma_vec;
static PyObject *__pyx_n_s_ma_vec_array;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mass_diff_abs;
static PyObject *__pyx_n_s_mass_diff_ppm;
static PyObject *__pyx_n_s_math;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max_charge;
static PyObject *__pyx_kp_s_mean_inverse_reduced_ion_mobilit;
static PyObject *__pyx_n_s_meanfilt;
static PyObject *__pyx_n_s_min_charge;
static PyObject *__pyx_n_s_min_idx;
static PyObject *__pyx_n_s_min_idx_list;
static PyObject *__pyx_n_s_min_length_hill;
static PyObject *__pyx_n_s_min_val;
static PyObject *__pyx_n_s_minlh;
static PyObject *__pyx_n_s_mono_hills_intensity_list;
static PyObject *__pyx_n_s_mono_hills_scan_lists;
static PyObject *__pyx_kp_s_monoisotope_hill_idx;
static PyObject *__pyx_kp_s_monoisotope_idx;
static PyObject *__pyx_n_s_mult_val;
static PyObject *__pyx_n_s_mz_cur;
static PyObject *__pyx_n_s_mz_median;
static PyObject *__pyx_n_s_mz_median_int;
static PyObject *__pyx_n_s_mz_sorted;
static PyObject *__pyx_n_s_mz_step;
static PyObject *__pyx_n_s_mz_tol;
static PyObject *__pyx_n_s_mz_val_tmp;
static PyObject *__pyx_n_s_mzs_array;
static PyObject *__pyx_n_s_nIsotopes;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_neutral_mass;
static PyObject *__pyx_n_s_new_index_list;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nproc;
static PyObject *__pyx_n_s_numb;
static PyObject *__pyx_n_s_number_of_passed_isotopes;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_orig_idx_array;
static PyObject *__pyx_n_s_orig_idx_val;
static PyObject *__pyx_n_s_paseftol;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_prev_fast_dict;
static PyObject *__pyx_n_s_prev_fast_dict_im;
static PyObject *__pyx_n_s_prev_idx;
static PyObject *__pyx_n_s_process_hills;
static PyObject *__pyx_n_s_product;
static PyObject *__pyx_n_s_r_r;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ready;
static PyObject *__pyx_n_s_recheck_idx;
static PyObject *__pyx_n_s_recheck_r_r;
static PyObject *__pyx_n_s_scan_id_val;
static PyObject *__pyx_n_s_scan_idx_array;
static PyObject *__pyx_n_s_scan_val;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_smothed_intensity;
static PyObject *__pyx_n_s_sorted_idx_array_child_process;
static PyObject *__pyx_n_s_sorted_idx_child_process;
static PyObject *__pyx_n_s_spec_idx;
static PyObject *__pyx_n_s_spec_mean_mass_accuracy;
static PyObject *__pyx_n_s_split_peaks;
static PyObject *__pyx_n_s_split_peaks_old;
static PyObject *__pyx_n_s_sqrt;
static PyObject *__pyx_n_s_suit_len;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theor_total_sum;
static PyObject *__pyx_n_s_theoretical_list;
static PyObject *__pyx_n_s_thresh;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tmp_candidates;
static PyObject *__pyx_n_s_tmp_hill_length;
static PyObject *__pyx_n_s_tmp_im_array;
static PyObject *__pyx_n_s_tmp_intensity;
static PyObject *__pyx_n_s_tmp_mz_array;
static PyObject *__pyx_n_s_tmp_orig_idx;
static PyObject *__pyx_n_s_tmp_scans;
static PyObject *__pyx_n_s_tmp_scans_list;
static PyObject *__pyx_n_s_tmp_scans_set;
static PyObject *__pyx_n_s_tmp_val;
static PyObject *__pyx_n_s_top;
static PyObject *__pyx_n_s_total_mass_diff;
static PyObject *__pyx_n_s_total_num_hills;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_weights;
static PyObject *__pyx_n_s_window_width;
static PyObject *__pyx_n_s_z;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_get_and_calc_values_for_cos_corr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, int __pyx_v_idx_1); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_2get_and_calc_apex_intensity_and_scan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, int __pyx_v_idx_1); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_4cos_correlation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hill_scans_1, PyObject *__pyx_v_hill_idict_1, float __pyx_v_hill_sqrt_of_i_1, PyObject *__pyx_v_hill_scans_2, PyObject *__pyx_v_hill_idict_2, float __pyx_v_hill_sqrt_of_i_2); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_6get_initial_isotopes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, float __pyx_v_isotopes_mass_accuracy, PyObject *__pyx_v_isotopes_list, PyObject *__pyx_v_a, int __pyx_v_min_charge, int __pyx_v_max_charge, float __pyx_v_mz_step, int __pyx_v_paseftol, float __pyx_v_faims_val, PyObject *__pyx_v_sorted_idx_child_process); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_8cos_correlation_new(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_theoretical_list, PyObject *__pyx_v_experimental_list); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_10checking_cos_correlation_for_carbon(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_theoretical_list, PyObject *__pyx_v_experimental_list, float __pyx_v_thresh); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_12get_fast_dict(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mz_sorted, float __pyx_v_mz_step, PyObject *__pyx_v_basic_id_sorted); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_14meanfilt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, int __pyx_v_window_width); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_16split_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, PyObject *__pyx_v_data_for_analyse_tmp, PyObject *__pyx_v_args, PyObject *__pyx_v_counter_hills_idx, PyObject *__pyx_v_sorted_idx_child_process, PyArrayObject *__pyx_v_sorted_idx_array_child_process, CYTHON_UNUSED int __pyx_v_nproc, int __pyx_v_checked_id); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_18split_peaks_old(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, PyObject *__pyx_v_data_for_analyse_tmp, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_20detect_hills(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data_for_analyse_tmp, PyObject *__pyx_v_args, float __pyx_v_mz_step, int __pyx_v_paseftol, int __pyx_v_dia); /* proto */
static PyObject *__pyx_pf_8biosaur2_6cutils_22process_hills(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, PyObject *__pyx_v_data_for_analyse_tmp, float __pyx_v_mz_step, int __pyx_v_paseftol, PyObject *__pyx_v_args, int __pyx_v_dia); /* proto */
static PyObject *__pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_intersection = {0, &__pyx_n_s_intersection, 0, 0, 0};
static PyObject *__pyx_float_0_6;
static PyObject *__pyx_float_1e6;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_slice_;
static PyObject *__pyx_slice__3;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
/* Late includes */

/* "biosaur2/cutils.pyx":30
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_1get_and_calc_values_for_cos_corr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_get_and_calc_values_for_cos_corr[] = "get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_1get_and_calc_values_for_cos_corr = {"get_and_calc_values_for_cos_corr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_1get_and_calc_values_for_cos_corr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_get_and_calc_values_for_cos_corr};
static PyObject *__pyx_pw_8biosaur2_6cutils_1get_and_calc_values_for_cos_corr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hills_dict = 0;
  int __pyx_v_idx_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_and_calc_values_for_cos_corr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hills_dict,&__pyx_n_s_idx_1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hills_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_idx_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_and_calc_values_for_cos_corr", 1, 2, 2, 1); __PYX_ERR(0, 30, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_and_calc_values_for_cos_corr") < 0)) __PYX_ERR(0, 30, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_hills_dict = ((PyObject*)values[0]);
    __pyx_v_idx_1 = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_idx_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 30, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_and_calc_values_for_cos_corr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 30, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.get_and_calc_values_for_cos_corr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hills_dict), (&PyDict_Type), 1, "hills_dict", 1))) __PYX_ERR(0, 30, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_get_and_calc_values_for_cos_corr(__pyx_self, __pyx_v_hills_dict, __pyx_v_idx_1);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "biosaur2/cutils.pyx":41
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
 *     if hill_sqrt_of_i_1 is None:
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))             # <<<<<<<<<<<<<<
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1
 * 
 */

static PyObject *__pyx_pf_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *)__pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct_1_genexpr(__pyx_ptype_8biosaur2_6cutils___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 41, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_get_and_calc_values_for_cos_corr, __pyx_n_s_biosaur2_cutils); if (unlikely(!gen)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("biosaur2.cutils.get_and_calc_values_for_cos_corr.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 41, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_hill_idict_1)) { __Pyx_RaiseClosureNameError("hill_idict_1"); __PYX_ERR(0, 41, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_hill_idict_1, __pyx_n_s_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 41, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 41, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 41, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Power(__pyx_cur_scope->__pyx_v_v, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 41, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":30
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 */

static PyObject *__pyx_pf_8biosaur2_6cutils_get_and_calc_values_for_cos_corr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, int __pyx_v_idx_1) {
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *__pyx_cur_scope;
  PyObject *__pyx_v_scan_id_val = NULL;
  PyObject *__pyx_v_intensity_val = NULL;
  PyObject *__pyx_v_hill_sqrt_of_i_1 = NULL;
  PyObject *__pyx_gb_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_and_calc_values_for_cos_corr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *)__pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr(__pyx_ptype_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 30, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "biosaur2/cutils.pyx":32
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]             # <<<<<<<<<<<<<<
 *     if hill_idict_1 is None:
 *         hill_idict_1 = dict()
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 32, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_hill_idict_1 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":33
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 *     if hill_idict_1 is None:             # <<<<<<<<<<<<<<
 *         hill_idict_1 = dict()
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):
 */
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_hill_idict_1 == Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "biosaur2/cutils.pyx":34
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 *     if hill_idict_1 is None:
 *         hill_idict_1 = dict()             # <<<<<<<<<<<<<<
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):
 *             hill_idict_1[scan_id_val] = intensity_val
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_hill_idict_1);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_hill_idict_1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "biosaur2/cutils.pyx":35
 *     if hill_idict_1 is None:
 *         hill_idict_1 = dict()
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):             # <<<<<<<<<<<<<<
 *             hill_idict_1[scan_id_val] = intensity_val
 *         hills_dict['hills_idict'][idx_1] = hill_idict_1
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 35, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_lists); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 35, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_2 = __pyx_t_5; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 35, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 35, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 35, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 35, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 35, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 35, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 35, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_1)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 35, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_scan_id_val, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_intensity_val, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "biosaur2/cutils.pyx":36
 *         hill_idict_1 = dict()
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):
 *             hill_idict_1[scan_id_val] = intensity_val             # <<<<<<<<<<<<<<
 *         hills_dict['hills_idict'][idx_1] = hill_idict_1
 * 
 */
      if (unlikely(PyObject_SetItem(__pyx_cur_scope->__pyx_v_hill_idict_1, __pyx_v_scan_id_val, __pyx_v_intensity_val) < 0)) __PYX_ERR(0, 36, __pyx_L1_error)

      /* "biosaur2/cutils.pyx":35
 *     if hill_idict_1 is None:
 *         hill_idict_1 = dict()
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):             # <<<<<<<<<<<<<<
 *             hill_idict_1[scan_id_val] = intensity_val
 *         hills_dict['hills_idict'][idx_1] = hill_idict_1
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "biosaur2/cutils.pyx":37
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):
 *             hill_idict_1[scan_id_val] = intensity_val
 *         hills_dict['hills_idict'][idx_1] = hill_idict_1             # <<<<<<<<<<<<<<
 * 
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 37, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_SetItemInt(__pyx_t_2, __pyx_v_idx_1, __pyx_cur_scope->__pyx_v_hill_idict_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "biosaur2/cutils.pyx":33
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 *     if hill_idict_1 is None:             # <<<<<<<<<<<<<<
 *         hill_idict_1 = dict()
 *         for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):
 */
  }

  /* "biosaur2/cutils.pyx":39
 *         hills_dict['hills_idict'][idx_1] = hill_idict_1
 * 
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]             # <<<<<<<<<<<<<<
 *     if hill_sqrt_of_i_1 is None:
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 39, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hill_sqrt_of_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_hill_sqrt_of_i_1 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "biosaur2/cutils.pyx":40
 * 
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
 *     if hill_sqrt_of_i_1 is None:             # <<<<<<<<<<<<<<
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1
 */
  __pyx_t_4 = (__pyx_v_hill_sqrt_of_i_1 == Py_None);
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "biosaur2/cutils.pyx":41
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
 *     if hill_sqrt_of_i_1 is None:
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))             # <<<<<<<<<<<<<<
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_math); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_pf_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF_SET(__pyx_v_hill_sqrt_of_i_1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "biosaur2/cutils.pyx":42
 *     if hill_sqrt_of_i_1 is None:
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1             # <<<<<<<<<<<<<<
 * 
 *     return hills_dict, hill_idict_1, hill_sqrt_of_i_1
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 42, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hill_sqrt_of_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_SetItemInt(__pyx_t_5, __pyx_v_idx_1, __pyx_v_hill_sqrt_of_i_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "biosaur2/cutils.pyx":40
 * 
 *     hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
 *     if hill_sqrt_of_i_1 is None:             # <<<<<<<<<<<<<<
 *         hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1
 */
  }

  /* "biosaur2/cutils.pyx":44
 *         hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1
 * 
 *     return hills_dict, hill_idict_1, hill_sqrt_of_i_1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_hills_dict);
  __Pyx_GIVEREF(__pyx_v_hills_dict);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_hills_dict);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_hill_idict_1);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_hill_idict_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_cur_scope->__pyx_v_hill_idict_1);
  __Pyx_INCREF(__pyx_v_hill_sqrt_of_i_1);
  __Pyx_GIVEREF(__pyx_v_hill_sqrt_of_i_1);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_hill_sqrt_of_i_1);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":30
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("biosaur2.cutils.get_and_calc_values_for_cos_corr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_scan_id_val);
  __Pyx_XDECREF(__pyx_v_intensity_val);
  __Pyx_XDECREF(__pyx_v_hill_sqrt_of_i_1);
  __Pyx_XDECREF(__pyx_gb_8biosaur2_6cutils_32get_and_calc_values_for_cos_corr_2generator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":50
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_3get_and_calc_apex_intensity_and_scan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_2get_and_calc_apex_intensity_and_scan[] = "get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_3get_and_calc_apex_intensity_and_scan = {"get_and_calc_apex_intensity_and_scan", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_3get_and_calc_apex_intensity_and_scan, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_2get_and_calc_apex_intensity_and_scan};
static PyObject *__pyx_pw_8biosaur2_6cutils_3get_and_calc_apex_intensity_and_scan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hills_dict = 0;
  int __pyx_v_idx_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_and_calc_apex_intensity_and_scan (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hills_dict,&__pyx_n_s_idx_1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hills_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_idx_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_and_calc_apex_intensity_and_scan", 1, 2, 2, 1); __PYX_ERR(0, 50, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_and_calc_apex_intensity_and_scan") < 0)) __PYX_ERR(0, 50, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_hills_dict = ((PyObject*)values[0]);
    __pyx_v_idx_1 = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_idx_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 50, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_and_calc_apex_intensity_and_scan", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 50, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.get_and_calc_apex_intensity_and_scan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hills_dict), (&PyDict_Type), 1, "hills_dict", 1))) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_2get_and_calc_apex_intensity_and_scan(__pyx_self, __pyx_v_hills_dict, __pyx_v_idx_1);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_2get_and_calc_apex_intensity_and_scan(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, int __pyx_v_idx_1) {
  PyObject *__pyx_v_hill_intensity_apex_1 = NULL;
  PyObject *__pyx_v_hill_scan_apex_1 = NULL;
  PyObject *__pyx_v_int_val = NULL;
  PyObject *__pyx_v_scan_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_and_calc_apex_intensity_and_scan", 0);

  /* "biosaur2/cutils.pyx":52
 * def get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1):
 * 
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]             # <<<<<<<<<<<<<<
 *     hill_scan_apex_1 = hills_dict['hills_scan_apex'][idx_1]
 *     if hill_intensity_apex_1 is None:
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 52, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_apex); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hill_intensity_apex_1 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":53
 * 
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 *     hill_scan_apex_1 = hills_dict['hills_scan_apex'][idx_1]             # <<<<<<<<<<<<<<
 *     if hill_intensity_apex_1 is None:
 * 
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 53, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_apex); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_hill_scan_apex_1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":54
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 *     hill_scan_apex_1 = hills_dict['hills_scan_apex'][idx_1]
 *     if hill_intensity_apex_1 is None:             # <<<<<<<<<<<<<<
 * 
 *         hill_intensity_apex_1 = 0
 */
  __pyx_t_3 = (__pyx_v_hill_intensity_apex_1 == Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "biosaur2/cutils.pyx":56
 *     if hill_intensity_apex_1 is None:
 * 
 *         hill_intensity_apex_1 = 0             # <<<<<<<<<<<<<<
 * 
 *         for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_DECREF_SET(__pyx_v_hill_intensity_apex_1, __pyx_int_0);

    /* "biosaur2/cutils.pyx":58
 *         hill_intensity_apex_1 = 0
 * 
 *         for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):             # <<<<<<<<<<<<<<
 *             if int_val > hill_intensity_apex_1:
 *                 hill_intensity_apex_1 = int_val
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 58, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 58, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_lists); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_1 = __pyx_t_5; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 58, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 58, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 58, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 58, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 58, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 58, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 58, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_int_val, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_scan_val, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "biosaur2/cutils.pyx":59
 * 
 *         for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):
 *             if int_val > hill_intensity_apex_1:             # <<<<<<<<<<<<<<
 *                 hill_intensity_apex_1 = int_val
 *                 hill_scan_apex_1 = scan_val
 */
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_int_val, __pyx_v_hill_intensity_apex_1, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 59, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 59, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_4) {

        /* "biosaur2/cutils.pyx":60
 *         for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):
 *             if int_val > hill_intensity_apex_1:
 *                 hill_intensity_apex_1 = int_val             # <<<<<<<<<<<<<<
 *                 hill_scan_apex_1 = scan_val
 * 
 */
        __Pyx_INCREF(__pyx_v_int_val);
        __Pyx_DECREF_SET(__pyx_v_hill_intensity_apex_1, __pyx_v_int_val);

        /* "biosaur2/cutils.pyx":61
 *             if int_val > hill_intensity_apex_1:
 *                 hill_intensity_apex_1 = int_val
 *                 hill_scan_apex_1 = scan_val             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['hills_intensity_apex'][idx_1] = hill_intensity_apex_1
 */
        __Pyx_INCREF(__pyx_v_scan_val);
        __Pyx_DECREF_SET(__pyx_v_hill_scan_apex_1, __pyx_v_scan_val);

        /* "biosaur2/cutils.pyx":59
 * 
 *         for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):
 *             if int_val > hill_intensity_apex_1:             # <<<<<<<<<<<<<<
 *                 hill_intensity_apex_1 = int_val
 *                 hill_scan_apex_1 = scan_val
 */
      }

      /* "biosaur2/cutils.pyx":58
 *         hill_intensity_apex_1 = 0
 * 
 *         for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):             # <<<<<<<<<<<<<<
 *             if int_val > hill_intensity_apex_1:
 *                 hill_intensity_apex_1 = int_val
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":63
 *                 hill_scan_apex_1 = scan_val
 * 
 *         hills_dict['hills_intensity_apex'][idx_1] = hill_intensity_apex_1             # <<<<<<<<<<<<<<
 *         hills_dict['hills_scan_apex'][idx_1] = hill_scan_apex_1
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 63, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_apex); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_t_1, __pyx_v_idx_1, __pyx_v_hill_intensity_apex_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":64
 * 
 *         hills_dict['hills_intensity_apex'][idx_1] = hill_intensity_apex_1
 *         hills_dict['hills_scan_apex'][idx_1] = hill_scan_apex_1             # <<<<<<<<<<<<<<
 * 
 *     return hills_dict, hill_intensity_apex_1, hill_scan_apex_1
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 64, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_apex); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_SetItemInt(__pyx_t_1, __pyx_v_idx_1, __pyx_v_hill_scan_apex_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0) < 0)) __PYX_ERR(0, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":54
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 *     hill_scan_apex_1 = hills_dict['hills_scan_apex'][idx_1]
 *     if hill_intensity_apex_1 is None:             # <<<<<<<<<<<<<<
 * 
 *         hill_intensity_apex_1 = 0
 */
  }

  /* "biosaur2/cutils.pyx":66
 *         hills_dict['hills_scan_apex'][idx_1] = hill_scan_apex_1
 * 
 *     return hills_dict, hill_intensity_apex_1, hill_scan_apex_1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_hills_dict);
  __Pyx_GIVEREF(__pyx_v_hills_dict);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_hills_dict);
  __Pyx_INCREF(__pyx_v_hill_intensity_apex_1);
  __Pyx_GIVEREF(__pyx_v_hill_intensity_apex_1);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_hill_intensity_apex_1);
  __Pyx_INCREF(__pyx_v_hill_scan_apex_1);
  __Pyx_GIVEREF(__pyx_v_hill_scan_apex_1);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_hill_scan_apex_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":50
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("biosaur2.cutils.get_and_calc_apex_intensity_and_scan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hill_intensity_apex_1);
  __Pyx_XDECREF(__pyx_v_hill_scan_apex_1);
  __Pyx_XDECREF(__pyx_v_int_val);
  __Pyx_XDECREF(__pyx_v_scan_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":72
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation(set hill_scans_1, dict hill_idict_1, float hill_sqrt_of_i_1, set hill_scans_2, dict hill_idict_2, float hill_sqrt_of_i_2):             # <<<<<<<<<<<<<<
 * 
 *     cdef int i
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_5cos_correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_4cos_correlation[] = "cos_correlation(set hill_scans_1, dict hill_idict_1, float hill_sqrt_of_i_1, set hill_scans_2, dict hill_idict_2, float hill_sqrt_of_i_2)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_5cos_correlation = {"cos_correlation", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_5cos_correlation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_4cos_correlation};
static PyObject *__pyx_pw_8biosaur2_6cutils_5cos_correlation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hill_scans_1 = 0;
  PyObject *__pyx_v_hill_idict_1 = 0;
  float __pyx_v_hill_sqrt_of_i_1;
  PyObject *__pyx_v_hill_scans_2 = 0;
  PyObject *__pyx_v_hill_idict_2 = 0;
  float __pyx_v_hill_sqrt_of_i_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cos_correlation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hill_scans_1,&__pyx_n_s_hill_idict_1,&__pyx_n_s_hill_sqrt_of_i_1,&__pyx_n_s_hill_scans_2,&__pyx_n_s_hill_idict_2,&__pyx_n_s_hill_sqrt_of_i_2,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hill_scans_1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hill_idict_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cos_correlation", 1, 6, 6, 1); __PYX_ERR(0, 72, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hill_sqrt_of_i_1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cos_correlation", 1, 6, 6, 2); __PYX_ERR(0, 72, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hill_scans_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cos_correlation", 1, 6, 6, 3); __PYX_ERR(0, 72, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hill_idict_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cos_correlation", 1, 6, 6, 4); __PYX_ERR(0, 72, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hill_sqrt_of_i_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cos_correlation", 1, 6, 6, 5); __PYX_ERR(0, 72, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cos_correlation") < 0)) __PYX_ERR(0, 72, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_hill_scans_1 = ((PyObject*)values[0]);
    __pyx_v_hill_idict_1 = ((PyObject*)values[1]);
    __pyx_v_hill_sqrt_of_i_1 = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_hill_sqrt_of_i_1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 72, __pyx_L3_error)
    __pyx_v_hill_scans_2 = ((PyObject*)values[3]);
    __pyx_v_hill_idict_2 = ((PyObject*)values[4]);
    __pyx_v_hill_sqrt_of_i_2 = __pyx_PyFloat_AsFloat(values[5]); if (unlikely((__pyx_v_hill_sqrt_of_i_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 72, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cos_correlation", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 72, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.cos_correlation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hill_scans_1), (&PySet_Type), 1, "hill_scans_1", 1))) __PYX_ERR(0, 72, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hill_idict_1), (&PyDict_Type), 1, "hill_idict_1", 1))) __PYX_ERR(0, 72, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hill_scans_2), (&PySet_Type), 1, "hill_scans_2", 1))) __PYX_ERR(0, 72, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hill_idict_2), (&PyDict_Type), 1, "hill_idict_2", 1))) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_4cos_correlation(__pyx_self, __pyx_v_hill_scans_1, __pyx_v_hill_idict_1, __pyx_v_hill_sqrt_of_i_1, __pyx_v_hill_scans_2, __pyx_v_hill_idict_2, __pyx_v_hill_sqrt_of_i_2);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_4cos_correlation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hill_scans_1, PyObject *__pyx_v_hill_idict_1, float __pyx_v_hill_sqrt_of_i_1, PyObject *__pyx_v_hill_scans_2, PyObject *__pyx_v_hill_idict_2, float __pyx_v_hill_sqrt_of_i_2) {
  int __pyx_v_i;
  float __pyx_v_top;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  float __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cos_correlation", 0);

  /* "biosaur2/cutils.pyx":77
 *     cdef float top
 * 
 *     top = 0             # <<<<<<<<<<<<<<
 *     for i in  hill_scans_1.intersection(hill_scans_2):
 *         top += hill_idict_1.get(i, 0) * hill_idict_2.get(i, 0)
 */
  __pyx_v_top = 0.0;

  /* "biosaur2/cutils.pyx":78
 * 
 *     top = 0
 *     for i in  hill_scans_1.intersection(hill_scans_2):             # <<<<<<<<<<<<<<
 *         top += hill_idict_1.get(i, 0) * hill_idict_2.get(i, 0)
 * 
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_intersection, __pyx_v_hill_scans_1, __pyx_v_hill_scans_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 78, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 78, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 78, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 78, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_i = __pyx_t_5;

    /* "biosaur2/cutils.pyx":79
 *     top = 0
 *     for i in  hill_scans_1.intersection(hill_scans_2):
 *         top += hill_idict_1.get(i, 0) * hill_idict_2.get(i, 0)             # <<<<<<<<<<<<<<
 * 
 *     return top / (hill_sqrt_of_i_1 * hill_sqrt_of_i_2)
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_top); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_hill_idict_1 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
      __PYX_ERR(0, 79, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyDict_GetItemDefault(__pyx_v_hill_idict_1, __pyx_t_6, __pyx_int_0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v_hill_idict_2 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
      __PYX_ERR(0, 79, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyDict_GetItemDefault(__pyx_v_hill_idict_2, __pyx_t_6, __pyx_int_0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __pyx_PyFloat_AsFloat(__pyx_t_8); if (unlikely((__pyx_t_9 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_top = __pyx_t_9;

    /* "biosaur2/cutils.pyx":78
 * 
 *     top = 0
 *     for i in  hill_scans_1.intersection(hill_scans_2):             # <<<<<<<<<<<<<<
 *         top += hill_idict_1.get(i, 0) * hill_idict_2.get(i, 0)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":81
 *         top += hill_idict_1.get(i, 0) * hill_idict_2.get(i, 0)
 * 
 *     return top / (hill_sqrt_of_i_1 * hill_sqrt_of_i_2)             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_top / (__pyx_v_hill_sqrt_of_i_1 * __pyx_v_hill_sqrt_of_i_2))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":72
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation(set hill_scans_1, dict hill_idict_1, float hill_sqrt_of_i_1, set hill_scans_2, dict hill_idict_2, float hill_sqrt_of_i_2):             # <<<<<<<<<<<<<<
 * 
 *     cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("biosaur2.cutils.cos_correlation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":86
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_initial_isotopes(dict hills_dict, float isotopes_mass_accuracy, list isotopes_list, dict a, int min_charge, int max_charge, float mz_step, bint paseftol, float faims_val, list sorted_idx_child_process):             # <<<<<<<<<<<<<<
 * 
 *     cdef list ready, charges, hill_scans_1_list, hill_scans_2_list, candidates, tmp_candidates
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_7get_initial_isotopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_6get_initial_isotopes[] = "get_initial_isotopes(dict hills_dict, float isotopes_mass_accuracy, list isotopes_list, dict a, int min_charge, int max_charge, float mz_step, bool paseftol, float faims_val, list sorted_idx_child_process)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_7get_initial_isotopes = {"get_initial_isotopes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_7get_initial_isotopes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_6get_initial_isotopes};
static PyObject *__pyx_pw_8biosaur2_6cutils_7get_initial_isotopes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hills_dict = 0;
  float __pyx_v_isotopes_mass_accuracy;
  PyObject *__pyx_v_isotopes_list = 0;
  PyObject *__pyx_v_a = 0;
  int __pyx_v_min_charge;
  int __pyx_v_max_charge;
  float __pyx_v_mz_step;
  int __pyx_v_paseftol;
  float __pyx_v_faims_val;
  PyObject *__pyx_v_sorted_idx_child_process = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_initial_isotopes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hills_dict,&__pyx_n_s_isotopes_mass_accuracy,&__pyx_n_s_isotopes_list,&__pyx_n_s_a,&__pyx_n_s_min_charge,&__pyx_n_s_max_charge,&__pyx_n_s_mz_step,&__pyx_n_s_paseftol,&__pyx_n_s_faims_val,&__pyx_n_s_sorted_idx_child_process,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hills_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_isotopes_mass_accuracy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 1); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_isotopes_list)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 2); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 3); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 4); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 5); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 6); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_paseftol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 7); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_faims_val)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 8); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sorted_idx_child_process)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, 9); __PYX_ERR(0, 86, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_initial_isotopes") < 0)) __PYX_ERR(0, 86, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_hills_dict = ((PyObject*)values[0]);
    __pyx_v_isotopes_mass_accuracy = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_isotopes_mass_accuracy == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L3_error)
    __pyx_v_isotopes_list = ((PyObject*)values[2]);
    __pyx_v_a = ((PyObject*)values[3]);
    __pyx_v_min_charge = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_min_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L3_error)
    __pyx_v_max_charge = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_max_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L3_error)
    __pyx_v_mz_step = __pyx_PyFloat_AsFloat(values[6]); if (unlikely((__pyx_v_mz_step == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L3_error)
    __pyx_v_paseftol = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_paseftol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L3_error)
    __pyx_v_faims_val = __pyx_PyFloat_AsFloat(values[8]); if (unlikely((__pyx_v_faims_val == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L3_error)
    __pyx_v_sorted_idx_child_process = ((PyObject*)values[9]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_initial_isotopes", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 86, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.get_initial_isotopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hills_dict), (&PyDict_Type), 1, "hills_dict", 1))) __PYX_ERR(0, 86, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_isotopes_list), (&PyList_Type), 1, "isotopes_list", 1))) __PYX_ERR(0, 86, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_a), (&PyDict_Type), 1, "a", 1))) __PYX_ERR(0, 86, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sorted_idx_child_process), (&PyList_Type), 1, "sorted_idx_child_process", 1))) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_6get_initial_isotopes(__pyx_self, __pyx_v_hills_dict, __pyx_v_isotopes_mass_accuracy, __pyx_v_isotopes_list, __pyx_v_a, __pyx_v_min_charge, __pyx_v_max_charge, __pyx_v_mz_step, __pyx_v_paseftol, __pyx_v_faims_val, __pyx_v_sorted_idx_child_process);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_6get_initial_isotopes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, float __pyx_v_isotopes_mass_accuracy, PyObject *__pyx_v_isotopes_list, PyObject *__pyx_v_a, int __pyx_v_min_charge, int __pyx_v_max_charge, float __pyx_v_mz_step, int __pyx_v_paseftol, float __pyx_v_faims_val, PyObject *__pyx_v_sorted_idx_child_process) {
  PyObject *__pyx_v_ready = 0;
  PyObject *__pyx_v_charges = 0;
  PyObject *__pyx_v_hill_scans_1_list = 0;
  PyObject *__pyx_v_candidates = 0;
  PyObject *__pyx_v_tmp_candidates = 0;
  int __pyx_v_idx_1;
  int __pyx_v_hill_idx_1;
  int __pyx_v_idx_2;
  int __pyx_v_hill_idx_2;
  int __pyx_v_im_to_check_fast;
  int __pyx_v_hill_scans_1_list_first;
  int __pyx_v_hill_scans_1_list_last;
  int __pyx_v_hill_scans_2_list_first;
  int __pyx_v_hill_scans_2_list_last;
  int __pyx_v_charge;
  int __pyx_v_isotope_number;
  int __pyx_v_m_to_check_fast;
  float __pyx_v_im_mz_1;
  float __pyx_v_hill_mz_1;
  float __pyx_v_hill_mz_2;
  float __pyx_v_m_to_check;
  float __pyx_v_mass_diff_abs;
  float __pyx_v_cos_cor_RT;
  float __pyx_v_hill_sqrt_of_i_1;
  float __pyx_v_hill_sqrt_of_i_2;
  PyObject *__pyx_v_banned_charges = 0;
  PyObject *__pyx_v_hill_idict_1 = 0;
  PyObject *__pyx_v_hill_idict_2 = 0;
  PyObject *__pyx_v_local_isotopes_dict = 0;
  PyObject *__pyx_v_hill_scans_1 = 0;
  PyObject *__pyx_v_hill_scans_2 = 0;
  double __pyx_v_mz_tol;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_neutral_mass = NULL;
  PyObject *__pyx_v_tmp_intensity = NULL;
  PyObject *__pyx_v_hill_intensity_apex_1 = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_hill_scan_apex_1 = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_mono_hills_scan_lists = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_mono_hills_intensity_list = NULL;
  PyObject *__pyx_v_all_theoretical_int = NULL;
  PyObject *__pyx_v_iter_candidates = NULL;
  PyObject *__pyx_v_all_exp_intensity = NULL;
  PyObject *__pyx_v_hill_intensity_apex_2 = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_hill_scan_apex_2 = NULL;
  PyObject *__pyx_v_cos_corr = NULL;
  PyObject *__pyx_v_number_of_passed_isotopes = NULL;
  PyObject *__pyx_v_local_res_dict = NULL;
  PyObject *__pyx_v_ch_v = NULL;
  PyObject *__pyx_v_z = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  float __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  PyObject *(*__pyx_t_26)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_initial_isotopes", 0);
  __Pyx_INCREF(__pyx_v_hills_dict);

  /* "biosaur2/cutils.pyx":96
 *     cdef set hill_scans_1, hill_scans_2
 * 
 *     ready = []             # <<<<<<<<<<<<<<
 * 
 *     charges = list(range(min_charge, max_charge + 1, 1)[::-1])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ready = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":98
 *     ready = []
 * 
 *     charges = list(range(min_charge, max_charge + 1, 1)[::-1])             # <<<<<<<<<<<<<<
 * 
 *     for idx_1 in sorted_idx_child_process:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_min_charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_long((__pyx_v_max_charge + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_int_1);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_slice_); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_charges = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":100
 *     charges = list(range(min_charge, max_charge + 1, 1)[::-1])
 * 
 *     for idx_1 in sorted_idx_child_process:             # <<<<<<<<<<<<<<
 *         hill_idx_1 = hills_dict['hills_idx_array_unique'][idx_1]
 *         hill_mz_1 = hills_dict['hills_mz_median'][idx_1]
 */
  if (unlikely(__pyx_v_sorted_idx_child_process == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 100, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_sorted_idx_child_process; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 100, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_idx_1 = __pyx_t_5;

    /* "biosaur2/cutils.pyx":101
 * 
 *     for idx_1 in sorted_idx_child_process:
 *         hill_idx_1 = hills_dict['hills_idx_array_unique'][idx_1]             # <<<<<<<<<<<<<<
 *         hill_mz_1 = hills_dict['hills_mz_median'][idx_1]
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 101, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_hill_idx_1 = __pyx_t_5;

    /* "biosaur2/cutils.pyx":102
 *     for idx_1 in sorted_idx_child_process:
 *         hill_idx_1 = hills_dict['hills_idx_array_unique'][idx_1]
 *         hill_mz_1 = hills_dict['hills_mz_median'][idx_1]             # <<<<<<<<<<<<<<
 * 
 *     #for (idx_1, hill_idx_1), hill_mz_1 in sorted(list(zip(list(enumerate(hills_dict['hills_idx_array_unique'])), hills_dict['hills_mz_median'])), key=lambda x: x[-1]):
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 102, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_hill_mz_1 = __pyx_t_6;

    /* "biosaur2/cutils.pyx":106
 *     #for (idx_1, hill_idx_1), hill_mz_1 in sorted(list(zip(list(enumerate(hills_dict['hills_idx_array_unique'])), hills_dict['hills_mz_median'])), key=lambda x: x[-1]):
 * 
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             im_mz_1 = hills_dict['hills_im_median'][idx_1]
 *             im_to_check_fast = int(im_mz_1 / paseftol)
 */
    __pyx_t_7 = ((__pyx_v_paseftol != 0) != 0);
    if (__pyx_t_7) {

      /* "biosaur2/cutils.pyx":107
 * 
 *         if paseftol is not False:
 *             im_mz_1 = hills_dict['hills_im_median'][idx_1]             # <<<<<<<<<<<<<<
 *             im_to_check_fast = int(im_mz_1 / paseftol)
 * 
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 107, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_im_mz_1 = __pyx_t_6;

      /* "biosaur2/cutils.pyx":108
 *         if paseftol is not False:
 *             im_mz_1 = hills_dict['hills_im_median'][idx_1]
 *             im_to_check_fast = int(im_mz_1 / paseftol)             # <<<<<<<<<<<<<<
 * 
 *         banned_charges = dict()
 */
      __pyx_v_im_to_check_fast = ((int)(__pyx_v_im_mz_1 / __pyx_v_paseftol));

      /* "biosaur2/cutils.pyx":106
 *     #for (idx_1, hill_idx_1), hill_mz_1 in sorted(list(zip(list(enumerate(hills_dict['hills_idx_array_unique'])), hills_dict['hills_mz_median'])), key=lambda x: x[-1]):
 * 
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             im_mz_1 = hills_dict['hills_im_median'][idx_1]
 *             im_to_check_fast = int(im_mz_1 / paseftol)
 */
    }

    /* "biosaur2/cutils.pyx":110
 *             im_to_check_fast = int(im_mz_1 / paseftol)
 * 
 *         banned_charges = dict()             # <<<<<<<<<<<<<<
 * 
 *         hill_scans_1 = hills_dict['hills_scan_sets'][idx_1]
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_banned_charges, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":112
 *         banned_charges = dict()
 * 
 *         hill_scans_1 = hills_dict['hills_scan_sets'][idx_1]             # <<<<<<<<<<<<<<
 *         hill_scans_1_list = hills_dict['hills_scan_lists'][idx_1]
 *         hill_scans_1_list_first, hill_scans_1_list_last = hill_scans_1_list[0], hill_scans_1_list[-1]
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 112, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_sets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PySet_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_hill_scans_1, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":113
 * 
 *         hill_scans_1 = hills_dict['hills_scan_sets'][idx_1]
 *         hill_scans_1_list = hills_dict['hills_scan_lists'][idx_1]             # <<<<<<<<<<<<<<
 *         hill_scans_1_list_first, hill_scans_1_list_last = hill_scans_1_list[0], hill_scans_1_list[-1]
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 113, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_lists); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_hill_scans_1_list, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":114
 *         hill_scans_1 = hills_dict['hills_scan_sets'][idx_1]
 *         hill_scans_1_list = hills_dict['hills_scan_lists'][idx_1]
 *         hill_scans_1_list_first, hill_scans_1_list_last = hill_scans_1_list[0], hill_scans_1_list[-1]             # <<<<<<<<<<<<<<
 * 
 *         mz_tol = isotopes_mass_accuracy * 1e-6 * hill_mz_1
 */
    if (unlikely(__pyx_v_hill_scans_1_list == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 114, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyInt_As_int(PyList_GET_ITEM(__pyx_v_hill_scans_1_list, 0)); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)
    if (unlikely(__pyx_v_hill_scans_1_list == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 114, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_hill_scans_1_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_hill_scans_1_list_first = __pyx_t_5;
    __pyx_v_hill_scans_1_list_last = __pyx_t_8;

    /* "biosaur2/cutils.pyx":116
 *         hill_scans_1_list_first, hill_scans_1_list_last = hill_scans_1_list[0], hill_scans_1_list[-1]
 * 
 *         mz_tol = isotopes_mass_accuracy * 1e-6 * hill_mz_1             # <<<<<<<<<<<<<<
 * 
 *         for charge in charges:
 */
    __pyx_v_mz_tol = ((__pyx_v_isotopes_mass_accuracy * 1e-6) * __pyx_v_hill_mz_1);

    /* "biosaur2/cutils.pyx":118
 *         mz_tol = isotopes_mass_accuracy * 1e-6 * hill_mz_1
 * 
 *         for charge in charges:             # <<<<<<<<<<<<<<
 * 
 *             candidates = []
 */
    __pyx_t_1 = __pyx_v_charges; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    for (;;) {
      if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
      #else
      __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_charge = __pyx_t_8;

      /* "biosaur2/cutils.pyx":120
 *         for charge in charges:
 * 
 *             candidates = []             # <<<<<<<<<<<<<<
 * 
 *             for isotope_number in isotopes_list[1:]:
 */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_candidates, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "biosaur2/cutils.pyx":122
 *             candidates = []
 * 
 *             for isotope_number in isotopes_list[1:]:             # <<<<<<<<<<<<<<
 * 
 *                 tmp_candidates = []
 */
      if (unlikely(__pyx_v_isotopes_list == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 122, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyList_GetSlice(__pyx_v_isotopes_list, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = __pyx_t_3; __Pyx_INCREF(__pyx_t_10); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_10, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_isotope_number = __pyx_t_8;

        /* "biosaur2/cutils.pyx":124
 *             for isotope_number in isotopes_list[1:]:
 * 
 *                 tmp_candidates = []             # <<<<<<<<<<<<<<
 * 
 *                 m_to_check = hill_mz_1 + (1.00335 * isotope_number / charge)
 */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 124, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_tmp_candidates, ((PyObject*)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "biosaur2/cutils.pyx":126
 *                 tmp_candidates = []
 * 
 *                 m_to_check = hill_mz_1 + (1.00335 * isotope_number / charge)             # <<<<<<<<<<<<<<
 *                 m_to_check_fast = int(m_to_check/mz_step)
 * 
 */
        __pyx_v_m_to_check = (__pyx_v_hill_mz_1 + ((1.00335 * __pyx_v_isotope_number) / __pyx_v_charge));

        /* "biosaur2/cutils.pyx":127
 * 
 *                 m_to_check = hill_mz_1 + (1.00335 * isotope_number / charge)
 *                 m_to_check_fast = int(m_to_check/mz_step)             # <<<<<<<<<<<<<<
 * 
 *                 for idx_2, hill_scans_2_list_first, hill_scans_2_list_last in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:
 */
        __pyx_v_m_to_check_fast = ((int)(__pyx_v_m_to_check / __pyx_v_mz_step));

        /* "biosaur2/cutils.pyx":129
 *                 m_to_check_fast = int(m_to_check/mz_step)
 * 
 *                 for idx_2, hill_scans_2_list_first, hill_scans_2_list_last in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:             # <<<<<<<<<<<<<<
 * 
 *                     if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 129, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median_fast_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_m_to_check_fast, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 129, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (likely(PyList_CheckExact(__pyx_t_12)) || PyTuple_CheckExact(__pyx_t_12)) {
          __pyx_t_3 = __pyx_t_12; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_14 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 129, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_12 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_12); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 129, __pyx_L1_error)
              #else
              __pyx_t_12 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 129, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              #endif
            } else {
              if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_12); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 129, __pyx_L1_error)
              #else
              __pyx_t_12 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 129, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              #endif
            }
          } else {
            __pyx_t_12 = __pyx_t_14(__pyx_t_3);
            if (unlikely(!__pyx_t_12)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 129, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_12);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_12))) || (PyList_CheckExact(__pyx_t_12))) {
            PyObject* sequence = __pyx_t_12;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 3)) {
              if (size > 3) __Pyx_RaiseTooManyValuesError(3);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 129, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_15 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
              __pyx_t_17 = PyTuple_GET_ITEM(sequence, 2); 
            } else {
              __pyx_t_15 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
              __pyx_t_17 = PyList_GET_ITEM(sequence, 2); 
            }
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_17);
            #else
            __pyx_t_15 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 129, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 129, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_17 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 129, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            #endif
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_18 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 129, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_19 = Py_TYPE(__pyx_t_18)->tp_iternext;
            index = 0; __pyx_t_15 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_15)) goto __pyx_L12_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_15);
            index = 1; __pyx_t_16 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_16)) goto __pyx_L12_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_16);
            index = 2; __pyx_t_17 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_17)) goto __pyx_L12_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_17);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_18), 3) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
            __pyx_t_19 = NULL;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            goto __pyx_L13_unpacking_done;
            __pyx_L12_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_19 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 129, __pyx_L1_error)
            __pyx_L13_unpacking_done:;
          }
          __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_15); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_16); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_20 = __Pyx_PyInt_As_int(__pyx_t_17); if (unlikely((__pyx_t_20 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_v_idx_2 = __pyx_t_8;
          __pyx_v_hill_scans_2_list_first = __pyx_t_5;
          __pyx_v_hill_scans_2_list_last = __pyx_t_20;

          /* "biosaur2/cutils.pyx":131
 *                 for idx_2, hill_scans_2_list_first, hill_scans_2_list_last in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:
 * 
 *                     if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:             # <<<<<<<<<<<<<<
 * 
 *                         if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:
 */
          __pyx_t_21 = ((__pyx_v_paseftol == 0) != 0);
          if (!__pyx_t_21) {
          } else {
            __pyx_t_7 = __pyx_t_21;
            goto __pyx_L15_bool_binop_done;
          }
          __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_idx_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 131, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (unlikely(__pyx_v_hills_dict == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 131, __pyx_L1_error)
          }
          __pyx_t_17 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median_fast_dict); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 131, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_16 = __Pyx_GetItemInt(__pyx_t_17, __pyx_v_im_to_check_fast, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 131, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_21 = (__Pyx_PySequence_ContainsTF(__pyx_t_12, __pyx_t_16, Py_EQ)); if (unlikely(__pyx_t_21 < 0)) __PYX_ERR(0, 131, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_22 = (__pyx_t_21 != 0);
          __pyx_t_7 = __pyx_t_22;
          __pyx_L15_bool_binop_done:;
          if (__pyx_t_7) {

            /* "biosaur2/cutils.pyx":133
 *                     if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:
 * 
 *                         if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:             # <<<<<<<<<<<<<<
 *                             hill_mz_2 = hills_dict['hills_mz_median'][idx_2]
 *                             mass_diff_abs = hill_mz_2 - m_to_check
 */
            __pyx_t_22 = ((!((__pyx_v_hill_scans_1_list_last < __pyx_v_hill_scans_2_list_first) != 0)) != 0);
            if (__pyx_t_22) {
            } else {
              __pyx_t_7 = __pyx_t_22;
              goto __pyx_L18_bool_binop_done;
            }
            __pyx_t_22 = ((!((__pyx_v_hill_scans_2_list_last < __pyx_v_hill_scans_1_list_first) != 0)) != 0);
            __pyx_t_7 = __pyx_t_22;
            __pyx_L18_bool_binop_done:;
            if (__pyx_t_7) {

              /* "biosaur2/cutils.pyx":134
 * 
 *                         if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:
 *                             hill_mz_2 = hills_dict['hills_mz_median'][idx_2]             # <<<<<<<<<<<<<<
 *                             mass_diff_abs = hill_mz_2 - m_to_check
 * 
 */
              if (unlikely(__pyx_v_hills_dict == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 134, __pyx_L1_error)
              }
              __pyx_t_16 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 134, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_idx_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 134, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_12); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_v_hill_mz_2 = __pyx_t_6;

              /* "biosaur2/cutils.pyx":135
 *                         if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:
 *                             hill_mz_2 = hills_dict['hills_mz_median'][idx_2]
 *                             mass_diff_abs = hill_mz_2 - m_to_check             # <<<<<<<<<<<<<<
 * 
 *                             if abs(mass_diff_abs) <= mz_tol:
 */
              __pyx_v_mass_diff_abs = (__pyx_v_hill_mz_2 - __pyx_v_m_to_check);

              /* "biosaur2/cutils.pyx":137
 *                             mass_diff_abs = hill_mz_2 - m_to_check
 * 
 *                             if abs(mass_diff_abs) <= mz_tol:             # <<<<<<<<<<<<<<
 *                                 hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]
 *                                 if len(hill_scans_1.intersection(hill_scans_2)) >= 1:
 */
              __pyx_t_7 = ((fabsf(__pyx_v_mass_diff_abs) <= __pyx_v_mz_tol) != 0);
              if (__pyx_t_7) {

                /* "biosaur2/cutils.pyx":138
 * 
 *                             if abs(mass_diff_abs) <= mz_tol:
 *                                 hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]             # <<<<<<<<<<<<<<
 *                                 if len(hill_scans_1.intersection(hill_scans_2)) >= 1:
 * 
 */
                if (unlikely(__pyx_v_hills_dict == Py_None)) {
                  PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                  __PYX_ERR(0, 138, __pyx_L1_error)
                }
                __pyx_t_12 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_sets); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 138, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_12);
                __pyx_t_16 = __Pyx_GetItemInt(__pyx_t_12, __pyx_v_idx_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 138, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_16);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (!(likely(PySet_CheckExact(__pyx_t_16))||((__pyx_t_16) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_16)->tp_name), 0))) __PYX_ERR(0, 138, __pyx_L1_error)
                __Pyx_XDECREF_SET(__pyx_v_hill_scans_2, ((PyObject*)__pyx_t_16));
                __pyx_t_16 = 0;

                /* "biosaur2/cutils.pyx":139
 *                             if abs(mass_diff_abs) <= mz_tol:
 *                                 hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]
 *                                 if len(hill_scans_1.intersection(hill_scans_2)) >= 1:             # <<<<<<<<<<<<<<
 * 
 * 
 */
                __pyx_t_16 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_intersection, __pyx_v_hill_scans_1, __pyx_v_hill_scans_2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 139, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_16);
                __pyx_t_23 = PyObject_Length(__pyx_t_16); if (unlikely(__pyx_t_23 == ((Py_ssize_t)-1))) __PYX_ERR(0, 139, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                __pyx_t_7 = ((__pyx_t_23 >= 1) != 0);
                if (__pyx_t_7) {

                  /* "biosaur2/cutils.pyx":142
 * 
 * 
 *                                     hills_dict, hill_idict_1, hill_sqrt_of_i_1 = get_and_calc_values_for_cos_corr(hills_dict, idx_1)             # <<<<<<<<<<<<<<
 *                                     hills_dict, hill_idict_2, hill_sqrt_of_i_2 = get_and_calc_values_for_cos_corr(hills_dict, idx_2)
 * 
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_get_and_calc_values_for_cos_corr_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 142, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_12);
                  __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 142, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __pyx_t_15 = NULL;
                  __pyx_t_20 = 0;
                  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
                    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_12);
                    if (likely(__pyx_t_15)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                      __Pyx_INCREF(__pyx_t_15);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_12, function);
                      __pyx_t_20 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_12)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_v_hills_dict, __pyx_t_17};
                    __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_v_hills_dict, __pyx_t_17};
                    __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  } else
                  #endif
                  {
                    __pyx_t_18 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    if (__pyx_t_15) {
                      __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_15); __pyx_t_15 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_hills_dict);
                    __Pyx_GIVEREF(__pyx_v_hills_dict);
                    PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_20, __pyx_v_hills_dict);
                    __Pyx_GIVEREF(__pyx_t_17);
                    PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_20, __pyx_t_17);
                    __pyx_t_17 = 0;
                    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_18, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                  if ((likely(PyTuple_CheckExact(__pyx_t_16))) || (PyList_CheckExact(__pyx_t_16))) {
                    PyObject* sequence = __pyx_t_16;
                    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                    if (unlikely(size != 3)) {
                      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                      __PYX_ERR(0, 142, __pyx_L1_error)
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    if (likely(PyTuple_CheckExact(sequence))) {
                      __pyx_t_12 = PyTuple_GET_ITEM(sequence, 0); 
                      __pyx_t_18 = PyTuple_GET_ITEM(sequence, 1); 
                      __pyx_t_17 = PyTuple_GET_ITEM(sequence, 2); 
                    } else {
                      __pyx_t_12 = PyList_GET_ITEM(sequence, 0); 
                      __pyx_t_18 = PyList_GET_ITEM(sequence, 1); 
                      __pyx_t_17 = PyList_GET_ITEM(sequence, 2); 
                    }
                    __Pyx_INCREF(__pyx_t_12);
                    __Pyx_INCREF(__pyx_t_18);
                    __Pyx_INCREF(__pyx_t_17);
                    #else
                    __pyx_t_12 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_18 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __pyx_t_17 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    #endif
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                  } else {
                    Py_ssize_t index = -1;
                    __pyx_t_15 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 142, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                    __pyx_t_19 = Py_TYPE(__pyx_t_15)->tp_iternext;
                    index = 0; __pyx_t_12 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_12)) goto __pyx_L22_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_12);
                    index = 1; __pyx_t_18 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_18)) goto __pyx_L22_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_18);
                    index = 2; __pyx_t_17 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_17)) goto __pyx_L22_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_17);
                    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_15), 3) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
                    __pyx_t_19 = NULL;
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    goto __pyx_L23_unpacking_done;
                    __pyx_L22_unpacking_failed:;
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __pyx_t_19 = NULL;
                    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                    __PYX_ERR(0, 142, __pyx_L1_error)
                    __pyx_L23_unpacking_done:;
                  }
                  if (!(likely(PyDict_CheckExact(__pyx_t_12))||((__pyx_t_12) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_12)->tp_name), 0))) __PYX_ERR(0, 142, __pyx_L1_error)
                  if (!(likely(PyDict_CheckExact(__pyx_t_18))||((__pyx_t_18) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_18)->tp_name), 0))) __PYX_ERR(0, 142, __pyx_L1_error)
                  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_17); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 142, __pyx_L1_error)
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  __Pyx_DECREF_SET(__pyx_v_hills_dict, ((PyObject*)__pyx_t_12));
                  __pyx_t_12 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_hill_idict_1, ((PyObject*)__pyx_t_18));
                  __pyx_t_18 = 0;
                  __pyx_v_hill_sqrt_of_i_1 = __pyx_t_6;

                  /* "biosaur2/cutils.pyx":143
 * 
 *                                     hills_dict, hill_idict_1, hill_sqrt_of_i_1 = get_and_calc_values_for_cos_corr(hills_dict, idx_1)
 *                                     hills_dict, hill_idict_2, hill_sqrt_of_i_2 = get_and_calc_values_for_cos_corr(hills_dict, idx_2)             # <<<<<<<<<<<<<<
 * 
 *                                     cos_cor_RT = cos_correlation(hill_scans_1, hill_idict_1, hill_sqrt_of_i_1, hill_scans_2, hill_idict_2, hill_sqrt_of_i_2)
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_get_and_calc_values_for_cos_corr_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 143, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_idx_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 143, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_12 = NULL;
                  __pyx_t_20 = 0;
                  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
                    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_17);
                    if (likely(__pyx_t_12)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                      __Pyx_INCREF(__pyx_t_12);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_17, function);
                      __pyx_t_20 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_17)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_hills_dict, __pyx_t_18};
                    __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_hills_dict, __pyx_t_18};
                    __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  } else
                  #endif
                  {
                    __pyx_t_15 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    if (__pyx_t_12) {
                      __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_12); __pyx_t_12 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_hills_dict);
                    __Pyx_GIVEREF(__pyx_v_hills_dict);
                    PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_20, __pyx_v_hills_dict);
                    __Pyx_GIVEREF(__pyx_t_18);
                    PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_20, __pyx_t_18);
                    __pyx_t_18 = 0;
                    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_15, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  if ((likely(PyTuple_CheckExact(__pyx_t_16))) || (PyList_CheckExact(__pyx_t_16))) {
                    PyObject* sequence = __pyx_t_16;
                    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                    if (unlikely(size != 3)) {
                      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                      __PYX_ERR(0, 143, __pyx_L1_error)
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    if (likely(PyTuple_CheckExact(sequence))) {
                      __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
                      __pyx_t_15 = PyTuple_GET_ITEM(sequence, 1); 
                      __pyx_t_18 = PyTuple_GET_ITEM(sequence, 2); 
                    } else {
                      __pyx_t_17 = PyList_GET_ITEM(sequence, 0); 
                      __pyx_t_15 = PyList_GET_ITEM(sequence, 1); 
                      __pyx_t_18 = PyList_GET_ITEM(sequence, 2); 
                    }
                    __Pyx_INCREF(__pyx_t_17);
                    __Pyx_INCREF(__pyx_t_15);
                    __Pyx_INCREF(__pyx_t_18);
                    #else
                    __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __pyx_t_15 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    __pyx_t_18 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    #endif
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                  } else {
                    Py_ssize_t index = -1;
                    __pyx_t_12 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 143, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                    __pyx_t_19 = Py_TYPE(__pyx_t_12)->tp_iternext;
                    index = 0; __pyx_t_17 = __pyx_t_19(__pyx_t_12); if (unlikely(!__pyx_t_17)) goto __pyx_L24_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_17);
                    index = 1; __pyx_t_15 = __pyx_t_19(__pyx_t_12); if (unlikely(!__pyx_t_15)) goto __pyx_L24_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_15);
                    index = 2; __pyx_t_18 = __pyx_t_19(__pyx_t_12); if (unlikely(!__pyx_t_18)) goto __pyx_L24_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_18);
                    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_12), 3) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
                    __pyx_t_19 = NULL;
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    goto __pyx_L25_unpacking_done;
                    __pyx_L24_unpacking_failed:;
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __pyx_t_19 = NULL;
                    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                    __PYX_ERR(0, 143, __pyx_L1_error)
                    __pyx_L25_unpacking_done:;
                  }
                  if (!(likely(PyDict_CheckExact(__pyx_t_17))||((__pyx_t_17) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_17)->tp_name), 0))) __PYX_ERR(0, 143, __pyx_L1_error)
                  if (!(likely(PyDict_CheckExact(__pyx_t_15))||((__pyx_t_15) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_15)->tp_name), 0))) __PYX_ERR(0, 143, __pyx_L1_error)
                  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_18); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 143, __pyx_L1_error)
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  __Pyx_DECREF_SET(__pyx_v_hills_dict, ((PyObject*)__pyx_t_17));
                  __pyx_t_17 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_hill_idict_2, ((PyObject*)__pyx_t_15));
                  __pyx_t_15 = 0;
                  __pyx_v_hill_sqrt_of_i_2 = __pyx_t_6;

                  /* "biosaur2/cutils.pyx":145
 *                                     hills_dict, hill_idict_2, hill_sqrt_of_i_2 = get_and_calc_values_for_cos_corr(hills_dict, idx_2)
 * 
 *                                     cos_cor_RT = cos_correlation(hill_scans_1, hill_idict_1, hill_sqrt_of_i_1, hill_scans_2, hill_idict_2, hill_sqrt_of_i_2)             # <<<<<<<<<<<<<<
 * 
 *                                     if cos_cor_RT >= 0.6:
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_cos_correlation); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 145, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_15 = PyFloat_FromDouble(__pyx_v_hill_sqrt_of_i_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 145, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_15);
                  __pyx_t_17 = PyFloat_FromDouble(__pyx_v_hill_sqrt_of_i_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 145, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __pyx_t_12 = NULL;
                  __pyx_t_20 = 0;
                  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
                    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_18);
                    if (likely(__pyx_t_12)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                      __Pyx_INCREF(__pyx_t_12);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_18, function);
                      __pyx_t_20 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_18)) {
                    PyObject *__pyx_temp[7] = {__pyx_t_12, __pyx_v_hill_scans_1, __pyx_v_hill_idict_1, __pyx_t_15, __pyx_v_hill_scans_2, __pyx_v_hill_idict_2, __pyx_t_17};
                    __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_18, __pyx_temp+1-__pyx_t_20, 6+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 145, __pyx_L1_error)
                    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_18)) {
                    PyObject *__pyx_temp[7] = {__pyx_t_12, __pyx_v_hill_scans_1, __pyx_v_hill_idict_1, __pyx_t_15, __pyx_v_hill_scans_2, __pyx_v_hill_idict_2, __pyx_t_17};
                    __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_18, __pyx_temp+1-__pyx_t_20, 6+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 145, __pyx_L1_error)
                    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  } else
                  #endif
                  {
                    __pyx_t_24 = PyTuple_New(6+__pyx_t_20); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 145, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    if (__pyx_t_12) {
                      __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_12); __pyx_t_12 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_hill_scans_1);
                    __Pyx_GIVEREF(__pyx_v_hill_scans_1);
                    PyTuple_SET_ITEM(__pyx_t_24, 0+__pyx_t_20, __pyx_v_hill_scans_1);
                    __Pyx_INCREF(__pyx_v_hill_idict_1);
                    __Pyx_GIVEREF(__pyx_v_hill_idict_1);
                    PyTuple_SET_ITEM(__pyx_t_24, 1+__pyx_t_20, __pyx_v_hill_idict_1);
                    __Pyx_GIVEREF(__pyx_t_15);
                    PyTuple_SET_ITEM(__pyx_t_24, 2+__pyx_t_20, __pyx_t_15);
                    __Pyx_INCREF(__pyx_v_hill_scans_2);
                    __Pyx_GIVEREF(__pyx_v_hill_scans_2);
                    PyTuple_SET_ITEM(__pyx_t_24, 3+__pyx_t_20, __pyx_v_hill_scans_2);
                    __Pyx_INCREF(__pyx_v_hill_idict_2);
                    __Pyx_GIVEREF(__pyx_v_hill_idict_2);
                    PyTuple_SET_ITEM(__pyx_t_24, 4+__pyx_t_20, __pyx_v_hill_idict_2);
                    __Pyx_GIVEREF(__pyx_t_17);
                    PyTuple_SET_ITEM(__pyx_t_24, 5+__pyx_t_20, __pyx_t_17);
                    __pyx_t_15 = 0;
                    __pyx_t_17 = 0;
                    __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_t_24, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 145, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_16); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L1_error)
                  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                  __pyx_v_cos_cor_RT = __pyx_t_6;

                  /* "biosaur2/cutils.pyx":147
 *                                     cos_cor_RT = cos_correlation(hill_scans_1, hill_idict_1, hill_sqrt_of_i_1, hill_scans_2, hill_idict_2, hill_sqrt_of_i_2)
 * 
 *                                     if cos_cor_RT >= 0.6:             # <<<<<<<<<<<<<<
 * 
 *                                         hill_idx_2 = hills_dict['hills_idx_array_unique'][idx_2]
 */
                  __pyx_t_7 = ((__pyx_v_cos_cor_RT >= 0.6) != 0);
                  if (__pyx_t_7) {

                    /* "biosaur2/cutils.pyx":149
 *                                     if cos_cor_RT >= 0.6:
 * 
 *                                         hill_idx_2 = hills_dict['hills_idx_array_unique'][idx_2]             # <<<<<<<<<<<<<<
 * 
 *                                         hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
 */
                    if (unlikely(__pyx_v_hills_dict == Py_None)) {
                      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                      __PYX_ERR(0, 149, __pyx_L1_error)
                    }
                    __pyx_t_16 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 149, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_idx_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 149, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                    __pyx_t_20 = __Pyx_PyInt_As_int(__pyx_t_18); if (unlikely((__pyx_t_20 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    __pyx_v_hill_idx_2 = __pyx_t_20;

                    /* "biosaur2/cutils.pyx":151
 *                                         hill_idx_2 = hills_dict['hills_idx_array_unique'][idx_2]
 * 
 *                                         hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)             # <<<<<<<<<<<<<<
 *                                         hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)
 * 
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_get_and_calc_apex_intensity_and); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 151, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    __pyx_t_24 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 151, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    __pyx_t_17 = NULL;
                    __pyx_t_20 = 0;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
                      __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);
                      if (likely(__pyx_t_17)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
                        __Pyx_INCREF(__pyx_t_17);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_16, function);
                        __pyx_t_20 = 1;
                      }
                    }
                    #if CYTHON_FAST_PYCALL
                    if (PyFunction_Check(__pyx_t_16)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_v_hills_dict, __pyx_t_24};
                      __pyx_t_18 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __Pyx_GOTREF(__pyx_t_18);
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    } else
                    #endif
                    #if CYTHON_FAST_PYCCALL
                    if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_v_hills_dict, __pyx_t_24};
                      __pyx_t_18 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __Pyx_GOTREF(__pyx_t_18);
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    } else
                    #endif
                    {
                      __pyx_t_15 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_15);
                      if (__pyx_t_17) {
                        __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_17); __pyx_t_17 = NULL;
                      }
                      __Pyx_INCREF(__pyx_v_hills_dict);
                      __Pyx_GIVEREF(__pyx_v_hills_dict);
                      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_20, __pyx_v_hills_dict);
                      __Pyx_GIVEREF(__pyx_t_24);
                      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_20, __pyx_t_24);
                      __pyx_t_24 = 0;
                      __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_15, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_18);
                      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                    if ((likely(PyTuple_CheckExact(__pyx_t_18))) || (PyList_CheckExact(__pyx_t_18))) {
                      PyObject* sequence = __pyx_t_18;
                      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                      if (unlikely(size != 3)) {
                        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                        __PYX_ERR(0, 151, __pyx_L1_error)
                      }
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      if (likely(PyTuple_CheckExact(sequence))) {
                        __pyx_t_16 = PyTuple_GET_ITEM(sequence, 0); 
                        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 1); 
                        __pyx_t_24 = PyTuple_GET_ITEM(sequence, 2); 
                      } else {
                        __pyx_t_16 = PyList_GET_ITEM(sequence, 0); 
                        __pyx_t_15 = PyList_GET_ITEM(sequence, 1); 
                        __pyx_t_24 = PyList_GET_ITEM(sequence, 2); 
                      }
                      __Pyx_INCREF(__pyx_t_16);
                      __Pyx_INCREF(__pyx_t_15);
                      __Pyx_INCREF(__pyx_t_24);
                      #else
                      __pyx_t_16 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_16);
                      __pyx_t_15 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_15);
                      __pyx_t_24 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_24);
                      #endif
                      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    } else {
                      Py_ssize_t index = -1;
                      __pyx_t_17 = PyObject_GetIter(__pyx_t_18); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 151, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_17);
                      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                      __pyx_t_19 = Py_TYPE(__pyx_t_17)->tp_iternext;
                      index = 0; __pyx_t_16 = __pyx_t_19(__pyx_t_17); if (unlikely(!__pyx_t_16)) goto __pyx_L27_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_16);
                      index = 1; __pyx_t_15 = __pyx_t_19(__pyx_t_17); if (unlikely(!__pyx_t_15)) goto __pyx_L27_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_15);
                      index = 2; __pyx_t_24 = __pyx_t_19(__pyx_t_17); if (unlikely(!__pyx_t_24)) goto __pyx_L27_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_24);
                      if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_17), 3) < 0) __PYX_ERR(0, 151, __pyx_L1_error)
                      __pyx_t_19 = NULL;
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                      goto __pyx_L28_unpacking_done;
                      __pyx_L27_unpacking_failed:;
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __pyx_t_19 = NULL;
                      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                      __PYX_ERR(0, 151, __pyx_L1_error)
                      __pyx_L28_unpacking_done:;
                    }
                    if (!(likely(PyDict_CheckExact(__pyx_t_16))||((__pyx_t_16) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_16)->tp_name), 0))) __PYX_ERR(0, 151, __pyx_L1_error)
                    __Pyx_DECREF_SET(__pyx_v_hills_dict, ((PyObject*)__pyx_t_16));
                    __pyx_t_16 = 0;
                    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_15);
                    __pyx_t_15 = 0;
                    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_24);
                    __pyx_t_24 = 0;

                    /* "biosaur2/cutils.pyx":152
 * 
 *                                         hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
 *                                         hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)             # <<<<<<<<<<<<<<
 * 
 *                                         local_isotopes_dict = {
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_get_and_calc_apex_intensity_and); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 152, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_idx_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 152, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    __pyx_t_16 = NULL;
                    __pyx_t_20 = 0;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_24))) {
                      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_24);
                      if (likely(__pyx_t_16)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_24);
                        __Pyx_INCREF(__pyx_t_16);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_24, function);
                        __pyx_t_20 = 1;
                      }
                    }
                    #if CYTHON_FAST_PYCALL
                    if (PyFunction_Check(__pyx_t_24)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_v_hills_dict, __pyx_t_15};
                      __pyx_t_18 = __Pyx_PyFunction_FastCall(__pyx_t_24, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                      __Pyx_GOTREF(__pyx_t_18);
                      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    } else
                    #endif
                    #if CYTHON_FAST_PYCCALL
                    if (__Pyx_PyFastCFunction_Check(__pyx_t_24)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_v_hills_dict, __pyx_t_15};
                      __pyx_t_18 = __Pyx_PyCFunction_FastCall(__pyx_t_24, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                      __Pyx_GOTREF(__pyx_t_18);
                      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    } else
                    #endif
                    {
                      __pyx_t_17 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_17);
                      if (__pyx_t_16) {
                        __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_16); __pyx_t_16 = NULL;
                      }
                      __Pyx_INCREF(__pyx_v_hills_dict);
                      __Pyx_GIVEREF(__pyx_v_hills_dict);
                      PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_20, __pyx_v_hills_dict);
                      __Pyx_GIVEREF(__pyx_t_15);
                      PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_20, __pyx_t_15);
                      __pyx_t_15 = 0;
                      __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_17, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_18);
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    if ((likely(PyTuple_CheckExact(__pyx_t_18))) || (PyList_CheckExact(__pyx_t_18))) {
                      PyObject* sequence = __pyx_t_18;
                      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                      if (unlikely(size != 3)) {
                        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                        __PYX_ERR(0, 152, __pyx_L1_error)
                      }
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      if (likely(PyTuple_CheckExact(sequence))) {
                        __pyx_t_24 = PyTuple_GET_ITEM(sequence, 0); 
                        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
                        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 2); 
                      } else {
                        __pyx_t_24 = PyList_GET_ITEM(sequence, 0); 
                        __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
                        __pyx_t_15 = PyList_GET_ITEM(sequence, 2); 
                      }
                      __Pyx_INCREF(__pyx_t_24);
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(__pyx_t_15);
                      #else
                      __pyx_t_24 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_24);
                      __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_17);
                      __pyx_t_15 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_15);
                      #endif
                      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    } else {
                      Py_ssize_t index = -1;
                      __pyx_t_16 = PyObject_GetIter(__pyx_t_18); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 152, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_16);
                      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                      __pyx_t_19 = Py_TYPE(__pyx_t_16)->tp_iternext;
                      index = 0; __pyx_t_24 = __pyx_t_19(__pyx_t_16); if (unlikely(!__pyx_t_24)) goto __pyx_L29_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_24);
                      index = 1; __pyx_t_17 = __pyx_t_19(__pyx_t_16); if (unlikely(!__pyx_t_17)) goto __pyx_L29_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_17);
                      index = 2; __pyx_t_15 = __pyx_t_19(__pyx_t_16); if (unlikely(!__pyx_t_15)) goto __pyx_L29_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_15);
                      if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_16), 3) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
                      __pyx_t_19 = NULL;
                      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                      goto __pyx_L30_unpacking_done;
                      __pyx_L29_unpacking_failed:;
                      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                      __pyx_t_19 = NULL;
                      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                      __PYX_ERR(0, 152, __pyx_L1_error)
                      __pyx_L30_unpacking_done:;
                    }
                    if (!(likely(PyDict_CheckExact(__pyx_t_24))||((__pyx_t_24) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_24)->tp_name), 0))) __PYX_ERR(0, 152, __pyx_L1_error)
                    __Pyx_DECREF_SET(__pyx_v_hills_dict, ((PyObject*)__pyx_t_24));
                    __pyx_t_24 = 0;
                    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_17);
                    __pyx_t_17 = 0;
                    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_15);
                    __pyx_t_15 = 0;

                    /* "biosaur2/cutils.pyx":155
 * 
 *                                         local_isotopes_dict = {
 *                                             'isotope_number': isotope_number,             # <<<<<<<<<<<<<<
 *                                             'isotope_hill_idx': hill_idx_2,
 *                                             'isotope_idx': idx_2,
 */
                    __pyx_t_18 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_isotope_number); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_isotope_number, __pyx_t_15) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                    /* "biosaur2/cutils.pyx":156
 *                                         local_isotopes_dict = {
 *                                             'isotope_number': isotope_number,
 *                                             'isotope_hill_idx': hill_idx_2,             # <<<<<<<<<<<<<<
 *                                             'isotope_idx': idx_2,
 *                                             'cos_cor': cos_cor_RT,
 */
                    __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_hill_idx_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 156, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_isotope_hill_idx, __pyx_t_15) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                    /* "biosaur2/cutils.pyx":157
 *                                             'isotope_number': isotope_number,
 *                                             'isotope_hill_idx': hill_idx_2,
 *                                             'isotope_idx': idx_2,             # <<<<<<<<<<<<<<
 *                                             'cos_cor': cos_cor_RT,
 *                                             'mass_diff_ppm': mass_diff_abs/m_to_check*1e6,
 */
                    __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_idx_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 157, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_isotope_idx, __pyx_t_15) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                    /* "biosaur2/cutils.pyx":158
 *                                             'isotope_hill_idx': hill_idx_2,
 *                                             'isotope_idx': idx_2,
 *                                             'cos_cor': cos_cor_RT,             # <<<<<<<<<<<<<<
 *                                             'mass_diff_ppm': mass_diff_abs/m_to_check*1e6,
 *                                         }
 */
                    __pyx_t_15 = PyFloat_FromDouble(__pyx_v_cos_cor_RT); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 158, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_cos_cor, __pyx_t_15) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                    /* "biosaur2/cutils.pyx":159
 *                                             'isotope_idx': idx_2,
 *                                             'cos_cor': cos_cor_RT,
 *                                             'mass_diff_ppm': mass_diff_abs/m_to_check*1e6,             # <<<<<<<<<<<<<<
 *                                         }
 * 
 */
                    __pyx_t_15 = PyFloat_FromDouble(((__pyx_v_mass_diff_abs / __pyx_v_m_to_check) * 1e6)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 159, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_15);
                    if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_mass_diff_ppm, __pyx_t_15) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                    __Pyx_XDECREF_SET(__pyx_v_local_isotopes_dict, ((PyObject*)__pyx_t_18));
                    __pyx_t_18 = 0;

                    /* "biosaur2/cutils.pyx":162
 *                                         }
 * 
 *                                         tmp_candidates.append(local_isotopes_dict)             # <<<<<<<<<<<<<<
 * 
 * 
 */
                    __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_tmp_candidates, __pyx_v_local_isotopes_dict); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 162, __pyx_L1_error)

                    /* "biosaur2/cutils.pyx":147
 *                                     cos_cor_RT = cos_correlation(hill_scans_1, hill_idict_1, hill_sqrt_of_i_1, hill_scans_2, hill_idict_2, hill_sqrt_of_i_2)
 * 
 *                                     if cos_cor_RT >= 0.6:             # <<<<<<<<<<<<<<
 * 
 *                                         hill_idx_2 = hills_dict['hills_idx_array_unique'][idx_2]
 */
                  }

                  /* "biosaur2/cutils.pyx":139
 *                             if abs(mass_diff_abs) <= mz_tol:
 *                                 hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]
 *                                 if len(hill_scans_1.intersection(hill_scans_2)) >= 1:             # <<<<<<<<<<<<<<
 * 
 * 
 */
                }

                /* "biosaur2/cutils.pyx":137
 *                             mass_diff_abs = hill_mz_2 - m_to_check
 * 
 *                             if abs(mass_diff_abs) <= mz_tol:             # <<<<<<<<<<<<<<
 *                                 hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]
 *                                 if len(hill_scans_1.intersection(hill_scans_2)) >= 1:
 */
              }

              /* "biosaur2/cutils.pyx":133
 *                     if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:
 * 
 *                         if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:             # <<<<<<<<<<<<<<
 *                             hill_mz_2 = hills_dict['hills_mz_median'][idx_2]
 *                             mass_diff_abs = hill_mz_2 - m_to_check
 */
            }

            /* "biosaur2/cutils.pyx":131
 *                 for idx_2, hill_scans_2_list_first, hill_scans_2_list_last in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:
 * 
 *                     if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:             # <<<<<<<<<<<<<<
 * 
 *                         if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:
 */
          }

          /* "biosaur2/cutils.pyx":129
 *                 m_to_check_fast = int(m_to_check/mz_step)
 * 
 *                 for idx_2, hill_scans_2_list_first, hill_scans_2_list_last in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:             # <<<<<<<<<<<<<<
 * 
 *                     if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "biosaur2/cutils.pyx":166
 * 
 * 
 *                 if len(tmp_candidates):             # <<<<<<<<<<<<<<
 *                     candidates.append(tmp_candidates)
 * 
 */
        __pyx_t_13 = PyList_GET_SIZE(__pyx_v_tmp_candidates); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 166, __pyx_L1_error)
        __pyx_t_7 = (__pyx_t_13 != 0);
        if (__pyx_t_7) {

          /* "biosaur2/cutils.pyx":167
 * 
 *                 if len(tmp_candidates):
 *                     candidates.append(tmp_candidates)             # <<<<<<<<<<<<<<
 * 
 *                 if len(candidates) < isotope_number:
 */
          __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_tmp_candidates); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 167, __pyx_L1_error)

          /* "biosaur2/cutils.pyx":166
 * 
 * 
 *                 if len(tmp_candidates):             # <<<<<<<<<<<<<<
 *                     candidates.append(tmp_candidates)
 * 
 */
        }

        /* "biosaur2/cutils.pyx":169
 *                     candidates.append(tmp_candidates)
 * 
 *                 if len(candidates) < isotope_number:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_13 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 169, __pyx_L1_error)
        __pyx_t_7 = ((__pyx_t_13 < __pyx_v_isotope_number) != 0);
        if (__pyx_t_7) {

          /* "biosaur2/cutils.pyx":170
 * 
 *                 if len(candidates) < isotope_number:
 *                     break             # <<<<<<<<<<<<<<
 * 
 * 
 */
          goto __pyx_L9_break;

          /* "biosaur2/cutils.pyx":169
 *                     candidates.append(tmp_candidates)
 * 
 *                 if len(candidates) < isotope_number:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        }

        /* "biosaur2/cutils.pyx":122
 *             candidates = []
 * 
 *             for isotope_number in isotopes_list[1:]:             # <<<<<<<<<<<<<<
 * 
 *                 tmp_candidates = []
 */
      }
      __pyx_L9_break:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "biosaur2/cutils.pyx":174
 * 
 * 
 *             if len(candidates) >= banned_charges.get(charge, 1):             # <<<<<<<<<<<<<<
 * 
 *                 neutral_mass = hill_mz_1 * charge
 */
      __pyx_t_11 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 174, __pyx_L1_error)
      __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_18 = __Pyx_PyDict_GetItemDefault(__pyx_v_banned_charges, __pyx_t_3, __pyx_int_1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_10, __pyx_t_18, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {

        /* "biosaur2/cutils.pyx":176
 *             if len(candidates) >= banned_charges.get(charge, 1):
 * 
 *                 neutral_mass = hill_mz_1 * charge             # <<<<<<<<<<<<<<
 * 
 *                 tmp_intensity = a[int(100 * (neutral_mass // 100))]
 */
        __pyx_t_3 = PyFloat_FromDouble((__pyx_v_hill_mz_1 * __pyx_v_charge)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_neutral_mass, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "biosaur2/cutils.pyx":178
 *                 neutral_mass = hill_mz_1 * charge
 * 
 *                 tmp_intensity = a[int(100 * (neutral_mass // 100))]             # <<<<<<<<<<<<<<
 * 
 *                 _, hill_intensity_apex_1, hill_scan_apex_1 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
 */
        if (unlikely(__pyx_v_a == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 178, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_PyInt_FloorDivideObjC(__pyx_v_neutral_mass, __pyx_int_100, 0x64, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_18 = PyNumber_Multiply(__pyx_int_100, __pyx_t_3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_a, __pyx_t_3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_tmp_intensity, __pyx_t_18);
        __pyx_t_18 = 0;

        /* "biosaur2/cutils.pyx":180
 *                 tmp_intensity = a[int(100 * (neutral_mass // 100))]
 * 
 *                 _, hill_intensity_apex_1, hill_scan_apex_1 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)             # <<<<<<<<<<<<<<
 *                 mono_hills_scan_lists =  hills_dict['hills_scan_lists'][idx_1]
 *                 mono_hills_intensity_list =  hills_dict['hills_intensity_array'][idx_1]
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_get_and_calc_apex_intensity_and); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_15 = NULL;
        __pyx_t_20 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_20 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_v_hills_dict, __pyx_t_10};
          __pyx_t_18 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_v_hills_dict, __pyx_t_10};
          __pyx_t_18 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else
        #endif
        {
          __pyx_t_17 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          if (__pyx_t_15) {
            __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_15); __pyx_t_15 = NULL;
          }
          __Pyx_INCREF(__pyx_v_hills_dict);
          __Pyx_GIVEREF(__pyx_v_hills_dict);
          PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_20, __pyx_v_hills_dict);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_20, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_17, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_18))) || (PyList_CheckExact(__pyx_t_18))) {
          PyObject* sequence = __pyx_t_18;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 180, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_10 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_10 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_10);
          #else
          __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_10 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_15 = PyObject_GetIter(__pyx_t_18); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 180, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __pyx_t_19 = Py_TYPE(__pyx_t_15)->tp_iternext;
          index = 0; __pyx_t_3 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_3)) goto __pyx_L34_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          index = 1; __pyx_t_17 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_17)) goto __pyx_L34_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_17);
          index = 2; __pyx_t_10 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_10)) goto __pyx_L34_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_10);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_15), 3) < 0) __PYX_ERR(0, 180, __pyx_L1_error)
          __pyx_t_19 = NULL;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L35_unpacking_done;
          __pyx_L34_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_19 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 180, __pyx_L1_error)
          __pyx_L35_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_hill_intensity_apex_1, __pyx_t_17);
        __pyx_t_17 = 0;
        __Pyx_XDECREF_SET(__pyx_v_hill_scan_apex_1, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "biosaur2/cutils.pyx":181
 * 
 *                 _, hill_intensity_apex_1, hill_scan_apex_1 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
 *                 mono_hills_scan_lists =  hills_dict['hills_scan_lists'][idx_1]             # <<<<<<<<<<<<<<
 *                 mono_hills_intensity_list =  hills_dict['hills_intensity_array'][idx_1]
 * 
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 181, __pyx_L1_error)
        }
        __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_lists); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_18, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF_SET(__pyx_v_mono_hills_scan_lists, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "biosaur2/cutils.pyx":182
 *                 _, hill_intensity_apex_1, hill_scan_apex_1 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
 *                 mono_hills_scan_lists =  hills_dict['hills_scan_lists'][idx_1]
 *                 mono_hills_intensity_list =  hills_dict['hills_intensity_array'][idx_1]             # <<<<<<<<<<<<<<
 * 
 *                 all_theoretical_int = [
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 182, __pyx_L1_error)
        }
        __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_10, __pyx_v_idx_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF_SET(__pyx_v_mono_hills_intensity_list, __pyx_t_18);
        __pyx_t_18 = 0;

        /* "biosaur2/cutils.pyx":184
 *                 mono_hills_intensity_list =  hills_dict['hills_intensity_array'][idx_1]
 * 
 *                 all_theoretical_int = [             # <<<<<<<<<<<<<<
 *                     hill_intensity_apex_1 *
 *                     tmp_intensity[z] /
 */
        __pyx_t_18 = PyList_New(0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);

        /* "biosaur2/cutils.pyx":187
 *                     hill_intensity_apex_1 *
 *                     tmp_intensity[z] /
 *                     tmp_intensity[0] for z in isotopes_list]             # <<<<<<<<<<<<<<
 * 
 *                 for iter_candidates in itertools.product(*candidates):
 */
        if (unlikely(__pyx_v_isotopes_list == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 187, __pyx_L1_error)
        }
        __pyx_t_10 = __pyx_v_isotopes_list; __Pyx_INCREF(__pyx_t_10); __pyx_t_11 = 0;
        for (;;) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_17 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_11); __Pyx_INCREF(__pyx_t_17); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 187, __pyx_L1_error)
          #else
          __pyx_t_17 = PySequence_ITEM(__pyx_t_10, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 187, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          #endif
          __Pyx_XDECREF_SET(__pyx_v_z, __pyx_t_17);
          __pyx_t_17 = 0;

          /* "biosaur2/cutils.pyx":186
 *                 all_theoretical_int = [
 *                     hill_intensity_apex_1 *
 *                     tmp_intensity[z] /             # <<<<<<<<<<<<<<
 *                     tmp_intensity[0] for z in isotopes_list]
 * 
 */
          __pyx_t_17 = __Pyx_PyObject_GetItem(__pyx_v_tmp_intensity, __pyx_v_z); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 186, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);

          /* "biosaur2/cutils.pyx":185
 * 
 *                 all_theoretical_int = [
 *                     hill_intensity_apex_1 *             # <<<<<<<<<<<<<<
 *                     tmp_intensity[z] /
 *                     tmp_intensity[0] for z in isotopes_list]
 */
          __pyx_t_3 = PyNumber_Multiply(__pyx_v_hill_intensity_apex_1, __pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

          /* "biosaur2/cutils.pyx":187
 *                     hill_intensity_apex_1 *
 *                     tmp_intensity[z] /
 *                     tmp_intensity[0] for z in isotopes_list]             # <<<<<<<<<<<<<<
 * 
 *                 for iter_candidates in itertools.product(*candidates):
 */
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_tmp_intensity, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 187, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);

          /* "biosaur2/cutils.pyx":186
 *                 all_theoretical_int = [
 *                     hill_intensity_apex_1 *
 *                     tmp_intensity[z] /             # <<<<<<<<<<<<<<
 *                     tmp_intensity[0] for z in isotopes_list]
 * 
 */
          __pyx_t_15 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_17); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 186, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_18, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "biosaur2/cutils.pyx":187
 *                     hill_intensity_apex_1 *
 *                     tmp_intensity[z] /
 *                     tmp_intensity[0] for z in isotopes_list]             # <<<<<<<<<<<<<<
 * 
 *                 for iter_candidates in itertools.product(*candidates):
 */
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF_SET(__pyx_v_all_theoretical_int, ((PyObject*)__pyx_t_18));
        __pyx_t_18 = 0;

        /* "biosaur2/cutils.pyx":189
 *                     tmp_intensity[0] for z in isotopes_list]
 * 
 *                 for iter_candidates in itertools.product(*candidates):             # <<<<<<<<<<<<<<
 * 
 *                     all_exp_intensity = [hill_intensity_apex_1, ]
 */
        __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_itertools); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_product); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = PySequence_Tuple(__pyx_v_candidates); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_18, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
          __pyx_t_18 = __pyx_t_15; __Pyx_INCREF(__pyx_t_18); __pyx_t_11 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_11 = -1; __pyx_t_18 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 189, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_14 = Py_TYPE(__pyx_t_18)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 189, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_18))) {
              if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_18)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyList_GET_ITEM(__pyx_t_18, __pyx_t_11); __Pyx_INCREF(__pyx_t_15); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_18, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 189, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            } else {
              if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_18)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_18, __pyx_t_11); __Pyx_INCREF(__pyx_t_15); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_18, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 189, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            }
          } else {
            __pyx_t_15 = __pyx_t_14(__pyx_t_18);
            if (unlikely(!__pyx_t_15)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 189, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_15);
          }
          __Pyx_XDECREF_SET(__pyx_v_iter_candidates, __pyx_t_15);
          __pyx_t_15 = 0;

          /* "biosaur2/cutils.pyx":191
 *                 for iter_candidates in itertools.product(*candidates):
 * 
 *                     all_exp_intensity = [hill_intensity_apex_1, ]             # <<<<<<<<<<<<<<
 * 
 *                     for local_isotopes_dict in iter_candidates:
 */
          __pyx_t_15 = PyList_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_INCREF(__pyx_v_hill_intensity_apex_1);
          __Pyx_GIVEREF(__pyx_v_hill_intensity_apex_1);
          PyList_SET_ITEM(__pyx_t_15, 0, __pyx_v_hill_intensity_apex_1);
          __Pyx_XDECREF_SET(__pyx_v_all_exp_intensity, ((PyObject*)__pyx_t_15));
          __pyx_t_15 = 0;

          /* "biosaur2/cutils.pyx":193
 *                     all_exp_intensity = [hill_intensity_apex_1, ]
 * 
 *                     for local_isotopes_dict in iter_candidates:             # <<<<<<<<<<<<<<
 * 
 *                         idx_2 = local_isotopes_dict['isotope_idx']
 */
          if (likely(PyList_CheckExact(__pyx_v_iter_candidates)) || PyTuple_CheckExact(__pyx_v_iter_candidates)) {
            __pyx_t_15 = __pyx_v_iter_candidates; __Pyx_INCREF(__pyx_t_15); __pyx_t_13 = 0;
            __pyx_t_26 = NULL;
          } else {
            __pyx_t_13 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_v_iter_candidates); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 193, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_26 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 193, __pyx_L1_error)
          }
          for (;;) {
            if (likely(!__pyx_t_26)) {
              if (likely(PyList_CheckExact(__pyx_t_15))) {
                if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_15)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_10 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_13); __Pyx_INCREF(__pyx_t_10); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)
                #else
                __pyx_t_10 = PySequence_ITEM(__pyx_t_15, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 193, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_10);
                #endif
              } else {
                if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_13); __Pyx_INCREF(__pyx_t_10); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)
                #else
                __pyx_t_10 = PySequence_ITEM(__pyx_t_15, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 193, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_10);
                #endif
              }
            } else {
              __pyx_t_10 = __pyx_t_26(__pyx_t_15);
              if (unlikely(!__pyx_t_10)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 193, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_10);
            }
            if (!(likely(PyDict_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 193, __pyx_L1_error)
            __Pyx_XDECREF_SET(__pyx_v_local_isotopes_dict, ((PyObject*)__pyx_t_10));
            __pyx_t_10 = 0;

            /* "biosaur2/cutils.pyx":195
 *                     for local_isotopes_dict in iter_candidates:
 * 
 *                         idx_2 = local_isotopes_dict['isotope_idx']             # <<<<<<<<<<<<<<
 * 
 *                         _, hill_intensity_apex_2, hill_scan_apex_2 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)
 */
            if (unlikely(__pyx_v_local_isotopes_dict == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 195, __pyx_L1_error)
            }
            __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_local_isotopes_dict, __pyx_n_s_isotope_idx); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 195, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_20 = __Pyx_PyInt_As_int(__pyx_t_10); if (unlikely((__pyx_t_20 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_v_idx_2 = __pyx_t_20;

            /* "biosaur2/cutils.pyx":197
 *                         idx_2 = local_isotopes_dict['isotope_idx']
 * 
 *                         _, hill_intensity_apex_2, hill_scan_apex_2 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)             # <<<<<<<<<<<<<<
 * 
 *                         all_exp_intensity.append(hill_intensity_apex_2)
 */
            __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_get_and_calc_apex_intensity_and); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 197, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_idx_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_24 = NULL;
            __pyx_t_20 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_17);
              if (likely(__pyx_t_24)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                __Pyx_INCREF(__pyx_t_24);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_17, function);
                __pyx_t_20 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_v_hills_dict, __pyx_t_3};
              __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_v_hills_dict, __pyx_t_3};
              __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_16 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_16);
              if (__pyx_t_24) {
                __Pyx_GIVEREF(__pyx_t_24); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_24); __pyx_t_24 = NULL;
              }
              __Pyx_INCREF(__pyx_v_hills_dict);
              __Pyx_GIVEREF(__pyx_v_hills_dict);
              PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_20, __pyx_v_hills_dict);
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_20, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_16, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
              PyObject* sequence = __pyx_t_10;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 3)) {
                if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 197, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
                __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
              } else {
                __pyx_t_17 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
                __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
              }
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_3);
              #else
              __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_24 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 197, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_24);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_19 = Py_TYPE(__pyx_t_24)->tp_iternext;
              index = 0; __pyx_t_17 = __pyx_t_19(__pyx_t_24); if (unlikely(!__pyx_t_17)) goto __pyx_L42_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_17);
              index = 1; __pyx_t_16 = __pyx_t_19(__pyx_t_24); if (unlikely(!__pyx_t_16)) goto __pyx_L42_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_16);
              index = 2; __pyx_t_3 = __pyx_t_19(__pyx_t_24); if (unlikely(!__pyx_t_3)) goto __pyx_L42_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_3);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_24), 3) < 0) __PYX_ERR(0, 197, __pyx_L1_error)
              __pyx_t_19 = NULL;
              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              goto __pyx_L43_unpacking_done;
              __pyx_L42_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              __pyx_t_19 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 197, __pyx_L1_error)
              __pyx_L43_unpacking_done:;
            }
            __Pyx_DECREF_SET(__pyx_v__, __pyx_t_17);
            __pyx_t_17 = 0;
            __Pyx_XDECREF_SET(__pyx_v_hill_intensity_apex_2, __pyx_t_16);
            __pyx_t_16 = 0;
            __Pyx_XDECREF_SET(__pyx_v_hill_scan_apex_2, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "biosaur2/cutils.pyx":199
 *                         _, hill_intensity_apex_2, hill_scan_apex_2 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)
 * 
 *                         all_exp_intensity.append(hill_intensity_apex_2)             # <<<<<<<<<<<<<<
 *                     cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)
 * 
 */
            __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_all_exp_intensity, __pyx_v_hill_intensity_apex_2); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 199, __pyx_L1_error)

            /* "biosaur2/cutils.pyx":193
 *                     all_exp_intensity = [hill_intensity_apex_1, ]
 * 
 *                     for local_isotopes_dict in iter_candidates:             # <<<<<<<<<<<<<<
 * 
 *                         idx_2 = local_isotopes_dict['isotope_idx']
 */
          }
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "biosaur2/cutils.pyx":200
 * 
 *                         all_exp_intensity.append(hill_intensity_apex_2)
 *                     cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)             # <<<<<<<<<<<<<<
 * 
 *                     if cos_corr:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_checking_cos_correlation_for_car); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 200, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_3 = NULL;
          __pyx_t_20 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_20 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_all_theoretical_int, __pyx_v_all_exp_intensity, __pyx_float_0_6};
            __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_20, 3+__pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_15);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_all_theoretical_int, __pyx_v_all_exp_intensity, __pyx_float_0_6};
            __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_20, 3+__pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_15);
          } else
          #endif
          {
            __pyx_t_16 = PyTuple_New(3+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_all_theoretical_int);
            __Pyx_GIVEREF(__pyx_v_all_theoretical_int);
            PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_20, __pyx_v_all_theoretical_int);
            __Pyx_INCREF(__pyx_v_all_exp_intensity);
            __Pyx_GIVEREF(__pyx_v_all_exp_intensity);
            PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_20, __pyx_v_all_exp_intensity);
            __Pyx_INCREF(__pyx_float_0_6);
            __Pyx_GIVEREF(__pyx_float_0_6);
            PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_20, __pyx_float_0_6);
            __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_16, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_15))) || (PyList_CheckExact(__pyx_t_15))) {
            PyObject* sequence = __pyx_t_15;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 200, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_16);
            #else
            __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            #endif
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_3 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_19 = Py_TYPE(__pyx_t_3)->tp_iternext;
            index = 0; __pyx_t_10 = __pyx_t_19(__pyx_t_3); if (unlikely(!__pyx_t_10)) goto __pyx_L44_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_10);
            index = 1; __pyx_t_16 = __pyx_t_19(__pyx_t_3); if (unlikely(!__pyx_t_16)) goto __pyx_L44_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_16);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_3), 2) < 0) __PYX_ERR(0, 200, __pyx_L1_error)
            __pyx_t_19 = NULL;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            goto __pyx_L45_unpacking_done;
            __pyx_L44_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_19 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 200, __pyx_L1_error)
            __pyx_L45_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_cos_corr, __pyx_t_10);
          __pyx_t_10 = 0;
          __Pyx_XDECREF_SET(__pyx_v_number_of_passed_isotopes, __pyx_t_16);
          __pyx_t_16 = 0;

          /* "biosaur2/cutils.pyx":202
 *                     cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)
 * 
 *                     if cos_corr:             # <<<<<<<<<<<<<<
 * 
 *                         iter_candidates = iter_candidates[:number_of_passed_isotopes]
 */
          __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_cos_corr); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 202, __pyx_L1_error)
          if (__pyx_t_7) {

            /* "biosaur2/cutils.pyx":204
 *                     if cos_corr:
 * 
 *                         iter_candidates = iter_candidates[:number_of_passed_isotopes]             # <<<<<<<<<<<<<<
 * 
 *                         local_res_dict = {
 */
            __pyx_t_15 = __Pyx_PyObject_GetSlice(__pyx_v_iter_candidates, 0, 0, NULL, &__pyx_v_number_of_passed_isotopes, NULL, 0, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 204, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF_SET(__pyx_v_iter_candidates, __pyx_t_15);
            __pyx_t_15 = 0;

            /* "biosaur2/cutils.pyx":207
 * 
 *                         local_res_dict = {
 *                             'monoisotope hill idx': hill_idx_1,             # <<<<<<<<<<<<<<
 *                             'monoisotope idx': idx_1,
 *                             'cos_cor_isotopes': cos_corr,
 */
            __pyx_t_15 = __Pyx_PyDict_NewPresized(10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_hill_idx_1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (PyDict_SetItem(__pyx_t_15, __pyx_kp_s_monoisotope_hill_idx, __pyx_t_16) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "biosaur2/cutils.pyx":208
 *                         local_res_dict = {
 *                             'monoisotope hill idx': hill_idx_1,
 *                             'monoisotope idx': idx_1,             # <<<<<<<<<<<<<<
 *                             'cos_cor_isotopes': cos_corr,
 *                             'hill_mz_1': hill_mz_1,
 */
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 208, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (PyDict_SetItem(__pyx_t_15, __pyx_kp_s_monoisotope_idx, __pyx_t_16) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "biosaur2/cutils.pyx":209
 *                             'monoisotope hill idx': hill_idx_1,
 *                             'monoisotope idx': idx_1,
 *                             'cos_cor_isotopes': cos_corr,             # <<<<<<<<<<<<<<
 *                             'hill_mz_1': hill_mz_1,
 *                             'isotopes': iter_candidates,
 */
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_cos_cor_isotopes, __pyx_v_cos_corr) < 0) __PYX_ERR(0, 207, __pyx_L1_error)

            /* "biosaur2/cutils.pyx":210
 *                             'monoisotope idx': idx_1,
 *                             'cos_cor_isotopes': cos_corr,
 *                             'hill_mz_1': hill_mz_1,             # <<<<<<<<<<<<<<
 *                             'isotopes': iter_candidates,
 *                             'nIsotopes': number_of_passed_isotopes,
 */
            __pyx_t_16 = PyFloat_FromDouble(__pyx_v_hill_mz_1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 210, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_hill_mz_1, __pyx_t_16) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "biosaur2/cutils.pyx":211
 *                             'cos_cor_isotopes': cos_corr,
 *                             'hill_mz_1': hill_mz_1,
 *                             'isotopes': iter_candidates,             # <<<<<<<<<<<<<<
 *                             'nIsotopes': number_of_passed_isotopes,
 *                             'charge': charge,
 */
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_isotopes, __pyx_v_iter_candidates) < 0) __PYX_ERR(0, 207, __pyx_L1_error)

            /* "biosaur2/cutils.pyx":212
 *                             'hill_mz_1': hill_mz_1,
 *                             'isotopes': iter_candidates,
 *                             'nIsotopes': number_of_passed_isotopes,             # <<<<<<<<<<<<<<
 *                             'charge': charge,
 *                             'FAIMS': faims_val,
 */
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_nIsotopes, __pyx_v_number_of_passed_isotopes) < 0) __PYX_ERR(0, 207, __pyx_L1_error)

            /* "biosaur2/cutils.pyx":213
 *                             'isotopes': iter_candidates,
 *                             'nIsotopes': number_of_passed_isotopes,
 *                             'charge': charge,             # <<<<<<<<<<<<<<
 *                             'FAIMS': faims_val,
 *                             'im': 0 if paseftol is False else im_mz_1,
 */
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 213, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_charge, __pyx_t_16) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "biosaur2/cutils.pyx":214
 *                             'nIsotopes': number_of_passed_isotopes,
 *                             'charge': charge,
 *                             'FAIMS': faims_val,             # <<<<<<<<<<<<<<
 *                             'im': 0 if paseftol is False else im_mz_1,
 *                             'intensity_array_for_cos_corr': [all_theoretical_int[:number_of_passed_isotopes+1], all_exp_intensity[:number_of_passed_isotopes+1]],
 */
            __pyx_t_16 = PyFloat_FromDouble(__pyx_v_faims_val); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 214, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_FAIMS, __pyx_t_16) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "biosaur2/cutils.pyx":215
 *                             'charge': charge,
 *                             'FAIMS': faims_val,
 *                             'im': 0 if paseftol is False else im_mz_1,             # <<<<<<<<<<<<<<
 *                             'intensity_array_for_cos_corr': [all_theoretical_int[:number_of_passed_isotopes+1], all_exp_intensity[:number_of_passed_isotopes+1]],
 *                         }
 */
            if (((__pyx_v_paseftol == 0) != 0)) {
              __Pyx_INCREF(__pyx_int_0);
              __pyx_t_16 = __pyx_int_0;
            } else {
              __pyx_t_10 = PyFloat_FromDouble(__pyx_v_im_mz_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 215, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_16 = __pyx_t_10;
              __pyx_t_10 = 0;
            }
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_im, __pyx_t_16) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "biosaur2/cutils.pyx":216
 *                             'FAIMS': faims_val,
 *                             'im': 0 if paseftol is False else im_mz_1,
 *                             'intensity_array_for_cos_corr': [all_theoretical_int[:number_of_passed_isotopes+1], all_exp_intensity[:number_of_passed_isotopes+1]],             # <<<<<<<<<<<<<<
 *                         }
 * 
 */
            __pyx_t_16 = __Pyx_PyInt_AddObjC(__pyx_v_number_of_passed_isotopes, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 216, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_7 = (__pyx_t_16 == Py_None);
            if (__pyx_t_7) {
              __pyx_t_13 = PY_SSIZE_T_MAX;
            } else {
              __pyx_t_23 = __Pyx_PyIndex_AsSsize_t(__pyx_t_16); if (unlikely((__pyx_t_23 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L1_error)
              __pyx_t_13 = __pyx_t_23;
            }
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_16 = __Pyx_PyList_GetSlice(__pyx_v_all_theoretical_int, 0, __pyx_t_13); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 216, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_v_number_of_passed_isotopes, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 216, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_7 = (__pyx_t_10 == Py_None);
            if (__pyx_t_7) {
              __pyx_t_13 = PY_SSIZE_T_MAX;
            } else {
              __pyx_t_23 = __Pyx_PyIndex_AsSsize_t(__pyx_t_10); if (unlikely((__pyx_t_23 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L1_error)
              __pyx_t_13 = __pyx_t_23;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyList_GetSlice(__pyx_v_all_exp_intensity, 0, __pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 216, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_10);
            __pyx_t_16 = 0;
            __pyx_t_10 = 0;
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_intensity_array_for_cos_corr, __pyx_t_3) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF_SET(__pyx_v_local_res_dict, ((PyObject*)__pyx_t_15));
            __pyx_t_15 = 0;

            /* "biosaur2/cutils.pyx":219
 *                         }
 * 
 *                         ready.append(local_res_dict)             # <<<<<<<<<<<<<<
 * 
 *                         for ch_v in charge_ban_map[charge]:
 */
            __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_ready, __pyx_v_local_res_dict); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 219, __pyx_L1_error)

            /* "biosaur2/cutils.pyx":221
 *                         ready.append(local_res_dict)
 * 
 *                         for ch_v in charge_ban_map[charge]:             # <<<<<<<<<<<<<<
 *                             banned_charges[ch_v] = max(number_of_passed_isotopes, banned_charges.get(ch_v, 1))
 *     return ready
 */
            if (unlikely(__pyx_v_8biosaur2_6cutils_charge_ban_map == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 221, __pyx_L1_error)
            }
            __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_charge); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 221, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_8biosaur2_6cutils_charge_ban_map, __pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
              __pyx_t_15 = __pyx_t_3; __Pyx_INCREF(__pyx_t_15); __pyx_t_13 = 0;
              __pyx_t_26 = NULL;
            } else {
              __pyx_t_13 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 221, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_26 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 221, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            for (;;) {
              if (likely(!__pyx_t_26)) {
                if (likely(PyList_CheckExact(__pyx_t_15))) {
                  if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_15)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_3 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 221, __pyx_L1_error)
                  #else
                  __pyx_t_3 = PySequence_ITEM(__pyx_t_15, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  #endif
                } else {
                  if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 221, __pyx_L1_error)
                  #else
                  __pyx_t_3 = PySequence_ITEM(__pyx_t_15, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  #endif
                }
              } else {
                __pyx_t_3 = __pyx_t_26(__pyx_t_15);
                if (unlikely(!__pyx_t_3)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 221, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_3);
              }
              __Pyx_XDECREF_SET(__pyx_v_ch_v, __pyx_t_3);
              __pyx_t_3 = 0;

              /* "biosaur2/cutils.pyx":222
 * 
 *                         for ch_v in charge_ban_map[charge]:
 *                             banned_charges[ch_v] = max(number_of_passed_isotopes, banned_charges.get(ch_v, 1))             # <<<<<<<<<<<<<<
 *     return ready
 * 
 */
              __pyx_t_3 = __Pyx_PyDict_GetItemDefault(__pyx_v_banned_charges, __pyx_v_ch_v, __pyx_int_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_INCREF(__pyx_v_number_of_passed_isotopes);
              __pyx_t_10 = __pyx_v_number_of_passed_isotopes;
              __pyx_t_17 = PyObject_RichCompare(__pyx_t_3, __pyx_t_10, Py_GT); __Pyx_XGOTREF(__pyx_t_17); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 222, __pyx_L1_error)
              __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 222, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (__pyx_t_7) {
                __Pyx_INCREF(__pyx_t_3);
                __pyx_t_16 = __pyx_t_3;
              } else {
                __Pyx_INCREF(__pyx_t_10);
                __pyx_t_16 = __pyx_t_10;
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __pyx_t_16;
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(PyDict_SetItem(__pyx_v_banned_charges, __pyx_v_ch_v, __pyx_t_3) < 0)) __PYX_ERR(0, 222, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "biosaur2/cutils.pyx":221
 *                         ready.append(local_res_dict)
 * 
 *                         for ch_v in charge_ban_map[charge]:             # <<<<<<<<<<<<<<
 *                             banned_charges[ch_v] = max(number_of_passed_isotopes, banned_charges.get(ch_v, 1))
 *     return ready
 */
            }
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

            /* "biosaur2/cutils.pyx":202
 *                     cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)
 * 
 *                     if cos_corr:             # <<<<<<<<<<<<<<
 * 
 *                         iter_candidates = iter_candidates[:number_of_passed_isotopes]
 */
          }

          /* "biosaur2/cutils.pyx":189
 *                     tmp_intensity[0] for z in isotopes_list]
 * 
 *                 for iter_candidates in itertools.product(*candidates):             # <<<<<<<<<<<<<<
 * 
 *                     all_exp_intensity = [hill_intensity_apex_1, ]
 */
        }
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

        /* "biosaur2/cutils.pyx":174
 * 
 * 
 *             if len(candidates) >= banned_charges.get(charge, 1):             # <<<<<<<<<<<<<<
 * 
 *                 neutral_mass = hill_mz_1 * charge
 */
      }

      /* "biosaur2/cutils.pyx":118
 *         mz_tol = isotopes_mass_accuracy * 1e-6 * hill_mz_1
 * 
 *         for charge in charges:             # <<<<<<<<<<<<<<
 * 
 *             candidates = []
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":100
 *     charges = list(range(min_charge, max_charge + 1, 1)[::-1])
 * 
 *     for idx_1 in sorted_idx_child_process:             # <<<<<<<<<<<<<<
 *         hill_idx_1 = hills_dict['hills_idx_array_unique'][idx_1]
 *         hill_mz_1 = hills_dict['hills_mz_median'][idx_1]
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":223
 *                         for ch_v in charge_ban_map[charge]:
 *                             banned_charges[ch_v] = max(number_of_passed_isotopes, banned_charges.get(ch_v, 1))
 *     return ready             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ready);
  __pyx_r = __pyx_v_ready;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":86
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_initial_isotopes(dict hills_dict, float isotopes_mass_accuracy, list isotopes_list, dict a, int min_charge, int max_charge, float mz_step, bint paseftol, float faims_val, list sorted_idx_child_process):             # <<<<<<<<<<<<<<
 * 
 *     cdef list ready, charges, hill_scans_1_list, hill_scans_2_list, candidates, tmp_candidates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_AddTraceback("biosaur2.cutils.get_initial_isotopes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ready);
  __Pyx_XDECREF(__pyx_v_charges);
  __Pyx_XDECREF(__pyx_v_hill_scans_1_list);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_tmp_candidates);
  __Pyx_XDECREF(__pyx_v_banned_charges);
  __Pyx_XDECREF(__pyx_v_hill_idict_1);
  __Pyx_XDECREF(__pyx_v_hill_idict_2);
  __Pyx_XDECREF(__pyx_v_local_isotopes_dict);
  __Pyx_XDECREF(__pyx_v_hill_scans_1);
  __Pyx_XDECREF(__pyx_v_hill_scans_2);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_neutral_mass);
  __Pyx_XDECREF(__pyx_v_tmp_intensity);
  __Pyx_XDECREF(__pyx_v_hill_intensity_apex_1);
  __Pyx_XDECREF(__pyx_v_hill_scan_apex_1);
  __Pyx_XDECREF(__pyx_v_mono_hills_scan_lists);
  __Pyx_XDECREF(__pyx_v_mono_hills_intensity_list);
  __Pyx_XDECREF(__pyx_v_all_theoretical_int);
  __Pyx_XDECREF(__pyx_v_iter_candidates);
  __Pyx_XDECREF(__pyx_v_all_exp_intensity);
  __Pyx_XDECREF(__pyx_v_hill_intensity_apex_2);
  __Pyx_XDECREF(__pyx_v_hill_scan_apex_2);
  __Pyx_XDECREF(__pyx_v_cos_corr);
  __Pyx_XDECREF(__pyx_v_number_of_passed_isotopes);
  __Pyx_XDECREF(__pyx_v_local_res_dict);
  __Pyx_XDECREF(__pyx_v_ch_v);
  __Pyx_XDECREF(__pyx_v_z);
  __Pyx_XDECREF(__pyx_v_hills_dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":228
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation_new(list theoretical_list, list experimental_list):             # <<<<<<<<<<<<<<
 *     cdef float theor_total_sum, top, i1, i2, bottom, numb, averagineExplained
 *     cdef int suit_len
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_9cos_correlation_new(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_8cos_correlation_new[] = "cos_correlation_new(list theoretical_list, list experimental_list)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_9cos_correlation_new = {"cos_correlation_new", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_9cos_correlation_new, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_8cos_correlation_new};
static PyObject *__pyx_pw_8biosaur2_6cutils_9cos_correlation_new(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_theoretical_list = 0;
  PyObject *__pyx_v_experimental_list = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cos_correlation_new (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_theoretical_list,&__pyx_n_s_experimental_list,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical_list)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental_list)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cos_correlation_new", 1, 2, 2, 1); __PYX_ERR(0, 228, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cos_correlation_new") < 0)) __PYX_ERR(0, 228, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_theoretical_list = ((PyObject*)values[0]);
    __pyx_v_experimental_list = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cos_correlation_new", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 228, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.cos_correlation_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical_list), (&PyList_Type), 1, "theoretical_list", 1))) __PYX_ERR(0, 228, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental_list), (&PyList_Type), 1, "experimental_list", 1))) __PYX_ERR(0, 228, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_8cos_correlation_new(__pyx_self, __pyx_v_theoretical_list, __pyx_v_experimental_list);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_8cos_correlation_new(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_theoretical_list, PyObject *__pyx_v_experimental_list) {
  float __pyx_v_theor_total_sum;
  float __pyx_v_top;
  float __pyx_v_i1;
  float __pyx_v_i2;
  float __pyx_v_bottom;
  float __pyx_v_numb;
  float __pyx_v_averagineExplained;
  int __pyx_v_suit_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  float __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  float __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cos_correlation_new", 0);
  __Pyx_INCREF(__pyx_v_theoretical_list);
  __Pyx_INCREF(__pyx_v_experimental_list);

  /* "biosaur2/cutils.pyx":233
 * 
 * 
 *     theor_total_sum = sum(theoretical_list)             # <<<<<<<<<<<<<<
 *     suit_len = min(len(theoretical_list), len(experimental_list))
 *     theoretical_list = theoretical_list[:suit_len]
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_v_theoretical_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_theor_total_sum = __pyx_t_2;

  /* "biosaur2/cutils.pyx":234
 * 
 *     theor_total_sum = sum(theoretical_list)
 *     suit_len = min(len(theoretical_list), len(experimental_list))             # <<<<<<<<<<<<<<
 *     theoretical_list = theoretical_list[:suit_len]
 *     experimental_list = experimental_list[:suit_len]
 */
  if (unlikely(__pyx_v_experimental_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 234, __pyx_L1_error)
  }
  __pyx_t_3 = PyList_GET_SIZE(__pyx_v_experimental_list); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 234, __pyx_L1_error)
  if (unlikely(__pyx_v_theoretical_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 234, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_theoretical_list); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 234, __pyx_L1_error)
  if (((__pyx_t_3 < __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_suit_len = __pyx_t_5;

  /* "biosaur2/cutils.pyx":235
 *     theor_total_sum = sum(theoretical_list)
 *     suit_len = min(len(theoretical_list), len(experimental_list))
 *     theoretical_list = theoretical_list[:suit_len]             # <<<<<<<<<<<<<<
 *     experimental_list = experimental_list[:suit_len]
 * 
 */
  if (unlikely(__pyx_v_theoretical_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 235, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_theoretical_list, 0, __pyx_v_suit_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_theoretical_list, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":236
 *     suit_len = min(len(theoretical_list), len(experimental_list))
 *     theoretical_list = theoretical_list[:suit_len]
 *     experimental_list = experimental_list[:suit_len]             # <<<<<<<<<<<<<<
 * 
 *     top = 0
 */
  if (unlikely(__pyx_v_experimental_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 236, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_experimental_list, 0, __pyx_v_suit_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_experimental_list, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":238
 *     experimental_list = experimental_list[:suit_len]
 * 
 *     top = 0             # <<<<<<<<<<<<<<
 * 
 *     for i1, i2 in zip(theoretical_list, experimental_list):
 */
  __pyx_v_top = 0.0;

  /* "biosaur2/cutils.pyx":240
 *     top = 0
 * 
 *     for i1, i2 in zip(theoretical_list, experimental_list):             # <<<<<<<<<<<<<<
 *         top += i1 * i2
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_theoretical_list);
  __Pyx_GIVEREF(__pyx_v_theoretical_list);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_theoretical_list);
  __Pyx_INCREF(__pyx_v_experimental_list);
  __Pyx_GIVEREF(__pyx_v_experimental_list);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_experimental_list);
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
    __pyx_t_1 = __pyx_t_6; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 240, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 240, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 240, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 240, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 240, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 240, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 240, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_8); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 240, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_12 = __pyx_PyFloat_AsFloat(__pyx_t_9); if (unlikely((__pyx_t_12 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 240, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_i1 = __pyx_t_2;
    __pyx_v_i2 = __pyx_t_12;

    /* "biosaur2/cutils.pyx":241
 * 
 *     for i1, i2 in zip(theoretical_list, experimental_list):
 *         top += i1 * i2             # <<<<<<<<<<<<<<
 * 
 *     if not top:
 */
    __pyx_v_top = (__pyx_v_top + (__pyx_v_i1 * __pyx_v_i2));

    /* "biosaur2/cutils.pyx":240
 *     top = 0
 * 
 *     for i1, i2 in zip(theoretical_list, experimental_list):             # <<<<<<<<<<<<<<
 *         top += i1 * i2
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":243
 *         top += i1 * i2
 * 
 *     if not top:             # <<<<<<<<<<<<<<
 *         return 0, 0
 *     else:
 */
  __pyx_t_13 = ((!(__pyx_v_top != 0)) != 0);
  if (__pyx_t_13) {

    /* "biosaur2/cutils.pyx":244
 * 
 *     if not top:
 *         return 0, 0             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_tuple__2);
    __pyx_r = __pyx_tuple__2;
    goto __pyx_L0;

    /* "biosaur2/cutils.pyx":243
 *         top += i1 * i2
 * 
 *     if not top:             # <<<<<<<<<<<<<<
 *         return 0, 0
 *     else:
 */
  }

  /* "biosaur2/cutils.pyx":247
 *     else:
 * 
 *         bottom = math.sqrt(sum([numb * numb for numb in theoretical_list])) * \             # <<<<<<<<<<<<<<
 *             math.sqrt(sum([numb * numb for numb in experimental_list]))
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_math); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __pyx_v_theoretical_list; __Pyx_INCREF(__pyx_t_8); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_8)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_10 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 247, __pyx_L1_error)
      #else
      __pyx_t_10 = PySequence_ITEM(__pyx_t_8, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __pyx_t_12 = __pyx_PyFloat_AsFloat(__pyx_t_10); if (unlikely((__pyx_t_12 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_numb = __pyx_t_12;
      __pyx_t_10 = PyFloat_FromDouble((__pyx_v_numb * __pyx_v_numb)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "biosaur2/cutils.pyx":248
 * 
 *         bottom = math.sqrt(sum([numb * numb for numb in theoretical_list])) * \
 *             math.sqrt(sum([numb * numb for numb in experimental_list]))             # <<<<<<<<<<<<<<
 * 
 *         averagineExplained = sum(theoretical_list) / theor_total_sum
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_math); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __pyx_v_experimental_list; __Pyx_INCREF(__pyx_t_10); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_10)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_14 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_5); __Pyx_INCREF(__pyx_t_14); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 248, __pyx_L1_error)
      #else
      __pyx_t_14 = PySequence_ITEM(__pyx_t_10, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      #endif
      __pyx_t_12 = __pyx_PyFloat_AsFloat(__pyx_t_14); if (unlikely((__pyx_t_12 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_v_numb = __pyx_t_12;
      __pyx_t_14 = PyFloat_FromDouble((__pyx_v_numb * __pyx_v_numb)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_14))) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_9 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":247
 *     else:
 * 
 *         bottom = math.sqrt(sum([numb * numb for numb in theoretical_list])) * \             # <<<<<<<<<<<<<<
 *             math.sqrt(sum([numb * numb for numb in experimental_list]))
 * 
 */
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_1, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_12 = __pyx_PyFloat_AsFloat(__pyx_t_6); if (unlikely((__pyx_t_12 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_bottom = __pyx_t_12;

    /* "biosaur2/cutils.pyx":250
 *             math.sqrt(sum([numb * numb for numb in experimental_list]))
 * 
 *         averagineExplained = sum(theoretical_list) / theor_total_sum             # <<<<<<<<<<<<<<
 * 
 *         return top / bottom, averagineExplained
 */
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_v_theoretical_list); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = PyFloat_FromDouble(__pyx_v_theor_total_sum); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_12 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_12 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_averagineExplained = __pyx_t_12;

    /* "biosaur2/cutils.pyx":252
 *         averagineExplained = sum(theoretical_list) / theor_total_sum
 * 
 *         return top / bottom, averagineExplained             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyFloat_FromDouble((__pyx_v_top / __pyx_v_bottom)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = PyFloat_FromDouble(__pyx_v_averagineExplained); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9);
    __pyx_t_1 = 0;
    __pyx_t_9 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "biosaur2/cutils.pyx":228
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation_new(list theoretical_list, list experimental_list):             # <<<<<<<<<<<<<<
 *     cdef float theor_total_sum, top, i1, i2, bottom, numb, averagineExplained
 *     cdef int suit_len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("biosaur2.cutils.cos_correlation_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_theoretical_list);
  __Pyx_XDECREF(__pyx_v_experimental_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":257
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def checking_cos_correlation_for_carbon(list theoretical_list, list experimental_list, float thresh):             # <<<<<<<<<<<<<<
 * 
 *     cdef float averagineCorrelation, averagineExplained, best_cor
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_11checking_cos_correlation_for_carbon(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_10checking_cos_correlation_for_carbon[] = "checking_cos_correlation_for_carbon(list theoretical_list, list experimental_list, float thresh)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_11checking_cos_correlation_for_carbon = {"checking_cos_correlation_for_carbon", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_11checking_cos_correlation_for_carbon, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_10checking_cos_correlation_for_carbon};
static PyObject *__pyx_pw_8biosaur2_6cutils_11checking_cos_correlation_for_carbon(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_theoretical_list = 0;
  PyObject *__pyx_v_experimental_list = 0;
  float __pyx_v_thresh;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checking_cos_correlation_for_carbon (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_theoretical_list,&__pyx_n_s_experimental_list,&__pyx_n_s_thresh,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theoretical_list)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_experimental_list)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checking_cos_correlation_for_carbon", 1, 3, 3, 1); __PYX_ERR(0, 257, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checking_cos_correlation_for_carbon", 1, 3, 3, 2); __PYX_ERR(0, 257, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checking_cos_correlation_for_carbon") < 0)) __PYX_ERR(0, 257, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_theoretical_list = ((PyObject*)values[0]);
    __pyx_v_experimental_list = ((PyObject*)values[1]);
    __pyx_v_thresh = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_thresh == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 257, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checking_cos_correlation_for_carbon", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 257, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.checking_cos_correlation_for_carbon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_theoretical_list), (&PyList_Type), 1, "theoretical_list", 1))) __PYX_ERR(0, 257, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_experimental_list), (&PyList_Type), 1, "experimental_list", 1))) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_10checking_cos_correlation_for_carbon(__pyx_self, __pyx_v_theoretical_list, __pyx_v_experimental_list, __pyx_v_thresh);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_10checking_cos_correlation_for_carbon(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_theoretical_list, PyObject *__pyx_v_experimental_list, float __pyx_v_thresh) {
  float __pyx_v_averagineCorrelation;
  float __pyx_v_averagineExplained;
  float __pyx_v_best_cor;
  int __pyx_v_best_pos;
  int __pyx_v_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  float __pyx_t_10;
  float __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("checking_cos_correlation_for_carbon", 0);

  /* "biosaur2/cutils.pyx":262
 *     cdef int best_pos, pos
 * 
 *     best_pos = 1             # <<<<<<<<<<<<<<
 *     best_cor = 0
 * 
 */
  __pyx_v_best_pos = 1;

  /* "biosaur2/cutils.pyx":263
 * 
 *     best_pos = 1
 *     best_cor = 0             # <<<<<<<<<<<<<<
 * 
 *     pos = len(experimental_list)
 */
  __pyx_v_best_cor = 0.0;

  /* "biosaur2/cutils.pyx":265
 *     best_cor = 0
 * 
 *     pos = len(experimental_list)             # <<<<<<<<<<<<<<
 * 
 *     while pos != 1:
 */
  if (unlikely(__pyx_v_experimental_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 265, __pyx_L1_error)
  }
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_experimental_list); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 265, __pyx_L1_error)
  __pyx_v_pos = __pyx_t_1;

  /* "biosaur2/cutils.pyx":267
 *     pos = len(experimental_list)
 * 
 *     while pos != 1:             # <<<<<<<<<<<<<<
 * 
 *         averagineCorrelation, averagineExplained = cos_correlation_new(theoretical_list, experimental_list[:pos])
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_pos != 1) != 0);
    if (!__pyx_t_2) break;

    /* "biosaur2/cutils.pyx":269
 *     while pos != 1:
 * 
 *         averagineCorrelation, averagineExplained = cos_correlation_new(theoretical_list, experimental_list[:pos])             # <<<<<<<<<<<<<<
 * 
 *         if averagineExplained >= 0.5 and averagineCorrelation >= thresh:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_cos_correlation_new); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_v_experimental_list == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 269, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyList_GetSlice(__pyx_v_experimental_list, 0, __pyx_v_pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_theoretical_list, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_theoretical_list, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_theoretical_list);
      __Pyx_GIVEREF(__pyx_v_theoretical_list);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_theoretical_list);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 269, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_9(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_8 = __pyx_t_9(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_5), 2) < 0) __PYX_ERR(0, 269, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 269, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_10 = __pyx_PyFloat_AsFloat(__pyx_t_4); if (unlikely((__pyx_t_10 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_11 = __pyx_PyFloat_AsFloat(__pyx_t_8); if (unlikely((__pyx_t_11 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_averagineCorrelation = __pyx_t_10;
    __pyx_v_averagineExplained = __pyx_t_11;

    /* "biosaur2/cutils.pyx":271
 *         averagineCorrelation, averagineExplained = cos_correlation_new(theoretical_list, experimental_list[:pos])
 * 
 *         if averagineExplained >= 0.5 and averagineCorrelation >= thresh:             # <<<<<<<<<<<<<<
 *             if averagineCorrelation > best_cor:
 *                 best_cor = averagineCorrelation
 */
    __pyx_t_12 = ((__pyx_v_averagineExplained >= 0.5) != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_2 = __pyx_t_12;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_12 = ((__pyx_v_averagineCorrelation >= __pyx_v_thresh) != 0);
    __pyx_t_2 = __pyx_t_12;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_2) {

      /* "biosaur2/cutils.pyx":272
 * 
 *         if averagineExplained >= 0.5 and averagineCorrelation >= thresh:
 *             if averagineCorrelation > best_cor:             # <<<<<<<<<<<<<<
 *                 best_cor = averagineCorrelation
 *                 best_pos = pos
 */
      __pyx_t_2 = ((__pyx_v_averagineCorrelation > __pyx_v_best_cor) != 0);
      if (__pyx_t_2) {

        /* "biosaur2/cutils.pyx":273
 *         if averagineExplained >= 0.5 and averagineCorrelation >= thresh:
 *             if averagineCorrelation > best_cor:
 *                 best_cor = averagineCorrelation             # <<<<<<<<<<<<<<
 *                 best_pos = pos
 * 
 */
        __pyx_v_best_cor = __pyx_v_averagineCorrelation;

        /* "biosaur2/cutils.pyx":274
 *             if averagineCorrelation > best_cor:
 *                 best_cor = averagineCorrelation
 *                 best_pos = pos             # <<<<<<<<<<<<<<
 * 
 *             break
 */
        __pyx_v_best_pos = __pyx_v_pos;

        /* "biosaur2/cutils.pyx":272
 * 
 *         if averagineExplained >= 0.5 and averagineCorrelation >= thresh:
 *             if averagineCorrelation > best_cor:             # <<<<<<<<<<<<<<
 *                 best_cor = averagineCorrelation
 *                 best_pos = pos
 */
      }

      /* "biosaur2/cutils.pyx":276
 *                 best_pos = pos
 * 
 *             break             # <<<<<<<<<<<<<<
 * 
 *         pos -= 1
 */
      goto __pyx_L4_break;

      /* "biosaur2/cutils.pyx":271
 *         averagineCorrelation, averagineExplained = cos_correlation_new(theoretical_list, experimental_list[:pos])
 * 
 *         if averagineExplained >= 0.5 and averagineCorrelation >= thresh:             # <<<<<<<<<<<<<<
 *             if averagineCorrelation > best_cor:
 *                 best_cor = averagineCorrelation
 */
    }

    /* "biosaur2/cutils.pyx":278
 *             break
 * 
 *         pos -= 1             # <<<<<<<<<<<<<<
 * 
 *     return best_cor, best_pos
 */
    __pyx_v_pos = (__pyx_v_pos - 1);
  }
  __pyx_L4_break:;

  /* "biosaur2/cutils.pyx":280
 *         pos -= 1
 * 
 *     return best_cor, best_pos             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_best_cor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_best_pos); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
  __pyx_t_3 = 0;
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":257
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def checking_cos_correlation_for_carbon(list theoretical_list, list experimental_list, float thresh):             # <<<<<<<<<<<<<<
 * 
 *     cdef float averagineCorrelation, averagineExplained, best_cor
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("biosaur2.cutils.checking_cos_correlation_for_carbon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":286
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_fast_dict(np.ndarray mz_sorted, float mz_step, list basic_id_sorted):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict fast_dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_13get_fast_dict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_12get_fast_dict[] = "get_fast_dict(ndarray mz_sorted, float mz_step, list basic_id_sorted)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_13get_fast_dict = {"get_fast_dict", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_13get_fast_dict, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_12get_fast_dict};
static PyObject *__pyx_pw_8biosaur2_6cutils_13get_fast_dict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_mz_sorted = 0;
  float __pyx_v_mz_step;
  PyObject *__pyx_v_basic_id_sorted = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fast_dict (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz_sorted,&__pyx_n_s_mz_step,&__pyx_n_s_basic_id_sorted,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz_sorted)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_fast_dict", 1, 3, 3, 1); __PYX_ERR(0, 286, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_basic_id_sorted)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_fast_dict", 1, 3, 3, 2); __PYX_ERR(0, 286, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_fast_dict") < 0)) __PYX_ERR(0, 286, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_mz_sorted = ((PyArrayObject *)values[0]);
    __pyx_v_mz_step = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_mz_step == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L3_error)
    __pyx_v_basic_id_sorted = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_fast_dict", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 286, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.get_fast_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mz_sorted), __pyx_ptype_5numpy_ndarray, 1, "mz_sorted", 0))) __PYX_ERR(0, 286, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_basic_id_sorted), (&PyList_Type), 1, "basic_id_sorted", 1))) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_12get_fast_dict(__pyx_self, __pyx_v_mz_sorted, __pyx_v_mz_step, __pyx_v_basic_id_sorted);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_12get_fast_dict(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mz_sorted, float __pyx_v_mz_step, PyObject *__pyx_v_basic_id_sorted) {
  PyObject *__pyx_v_fast_dict = 0;
  PyObject *__pyx_v_fast_array = 0;
  int __pyx_v_idx;
  int __pyx_v_fm;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fast_dict", 0);

  /* "biosaur2/cutils.pyx":292
 *     cdef int idx, fm
 * 
 *     fast_dict = dict()             # <<<<<<<<<<<<<<
 *     fast_array = list((mz_sorted/mz_step).astype(int))
 *     for idx, fm in zip(basic_id_sorted, fast_array):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fast_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":293
 * 
 *     fast_dict = dict()
 *     fast_array = list((mz_sorted/mz_step).astype(int))             # <<<<<<<<<<<<<<
 *     for idx, fm in zip(basic_id_sorted, fast_array):
 *         if fm not in fast_dict:
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_mz_step); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_mz_sorted), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)(&PyInt_Type))) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)(&PyInt_Type)));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fast_array = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":294
 *     fast_dict = dict()
 *     fast_array = list((mz_sorted/mz_step).astype(int))
 *     for idx, fm in zip(basic_id_sorted, fast_array):             # <<<<<<<<<<<<<<
 *         if fm not in fast_dict:
 *             fast_dict[fm] = [idx, ]
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_basic_id_sorted);
  __Pyx_GIVEREF(__pyx_v_basic_id_sorted);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_basic_id_sorted);
  __Pyx_INCREF(__pyx_v_fast_array);
  __Pyx_GIVEREF(__pyx_v_fast_array);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_fast_array);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 294, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 294, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 294, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_idx = __pyx_t_9;
    __pyx_v_fm = __pyx_t_10;

    /* "biosaur2/cutils.pyx":295
 *     fast_array = list((mz_sorted/mz_step).astype(int))
 *     for idx, fm in zip(basic_id_sorted, fast_array):
 *         if fm not in fast_dict:             # <<<<<<<<<<<<<<
 *             fast_dict[fm] = [idx, ]
 *         else:
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyDict_ContainsTF(__pyx_t_1, __pyx_v_fast_dict, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {

      /* "biosaur2/cutils.pyx":296
 *     for idx, fm in zip(basic_id_sorted, fast_array):
 *         if fm not in fast_dict:
 *             fast_dict[fm] = [idx, ]             # <<<<<<<<<<<<<<
 *         else:
 *             fast_dict[fm].append(idx)
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1);
      PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(PyDict_SetItem(__pyx_v_fast_dict, __pyx_t_1, __pyx_t_6) < 0)) __PYX_ERR(0, 296, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "biosaur2/cutils.pyx":295
 *     fast_array = list((mz_sorted/mz_step).astype(int))
 *     for idx, fm in zip(basic_id_sorted, fast_array):
 *         if fm not in fast_dict:             # <<<<<<<<<<<<<<
 *             fast_dict[fm] = [idx, ]
 *         else:
 */
      goto __pyx_L7;
    }

    /* "biosaur2/cutils.pyx":298
 *             fast_dict[fm] = [idx, ]
 *         else:
 *             fast_dict[fm].append(idx)             # <<<<<<<<<<<<<<
 *     return fast_dict, fast_array
 * 
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_fm); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_fast_dict, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_6); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_L7:;

    /* "biosaur2/cutils.pyx":294
 *     fast_dict = dict()
 *     fast_array = list((mz_sorted/mz_step).astype(int))
 *     for idx, fm in zip(basic_id_sorted, fast_array):             # <<<<<<<<<<<<<<
 *         if fm not in fast_dict:
 *             fast_dict[fm] = [idx, ]
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":299
 *         else:
 *             fast_dict[fm].append(idx)
 *     return fast_dict, fast_array             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_fast_dict);
  __Pyx_GIVEREF(__pyx_v_fast_dict);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_fast_dict);
  __Pyx_INCREF(__pyx_v_fast_array);
  __Pyx_GIVEREF(__pyx_v_fast_array);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_fast_array);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":286
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_fast_dict(np.ndarray mz_sorted, float mz_step, list basic_id_sorted):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict fast_dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("biosaur2.cutils.get_fast_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fast_dict);
  __Pyx_XDECREF(__pyx_v_fast_array);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":305
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def meanfilt(list data, int window_width):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray cumsum_vec, ma_vec_array
 *     cdef list ma_vec
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_15meanfilt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_14meanfilt[] = "meanfilt(list data, int window_width)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_15meanfilt = {"meanfilt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_15meanfilt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_14meanfilt};
static PyObject *__pyx_pw_8biosaur2_6cutils_15meanfilt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  int __pyx_v_window_width;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("meanfilt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,&__pyx_n_s_window_width,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_window_width)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("meanfilt", 1, 2, 2, 1); __PYX_ERR(0, 305, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "meanfilt") < 0)) __PYX_ERR(0, 305, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_data = ((PyObject*)values[0]);
    __pyx_v_window_width = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_window_width == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("meanfilt", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 305, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.meanfilt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyList_Type), 1, "data", 1))) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_14meanfilt(__pyx_self, __pyx_v_data, __pyx_v_window_width);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_14meanfilt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, int __pyx_v_window_width) {
  PyArrayObject *__pyx_v_cumsum_vec = 0;
  PyArrayObject *__pyx_v_ma_vec_array = 0;
  PyObject *__pyx_v_ma_vec = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("meanfilt", 0);

  /* "biosaur2/cutils.pyx":308
 *     cdef np.ndarray cumsum_vec, ma_vec_array
 *     cdef list ma_vec
 *     cumsum_vec = np.cumsum(np.insert(data, 0, 0))             # <<<<<<<<<<<<<<
 *     ma_vec_array = (cumsum_vec[window_width:] -
 *               cumsum_vec[:-window_width]) / window_width
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_cumsum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_insert); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_data, __pyx_int_0, __pyx_int_0};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_data, __pyx_int_0, __pyx_int_0};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_data);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_0);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_int_0);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_v_cumsum_vec = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":309
 *     cdef list ma_vec
 *     cumsum_vec = np.cumsum(np.insert(data, 0, 0))
 *     ma_vec_array = (cumsum_vec[window_width:] -             # <<<<<<<<<<<<<<
 *               cumsum_vec[:-window_width]) / window_width
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_cumsum_vec), __pyx_v_window_width, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "biosaur2/cutils.pyx":310
 *     cumsum_vec = np.cumsum(np.insert(data, 0, 0))
 *     ma_vec_array = (cumsum_vec[window_width:] -
 *               cumsum_vec[:-window_width]) / window_width             # <<<<<<<<<<<<<<
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]
 *     return ma_vec
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_cumsum_vec), 0, (-__pyx_v_window_width), NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "biosaur2/cutils.pyx":309
 *     cdef list ma_vec
 *     cumsum_vec = np.cumsum(np.insert(data, 0, 0))
 *     ma_vec_array = (cumsum_vec[window_width:] -             # <<<<<<<<<<<<<<
 *               cumsum_vec[:-window_width]) / window_width
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]
 */
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "biosaur2/cutils.pyx":310
 *     cumsum_vec = np.cumsum(np.insert(data, 0, 0))
 *     ma_vec_array = (cumsum_vec[window_width:] -
 *               cumsum_vec[:-window_width]) / window_width             # <<<<<<<<<<<<<<
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]
 *     return ma_vec
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_window_width); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 310, __pyx_L1_error)
  __pyx_v_ma_vec_array = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":311
 *     ma_vec_array = (cumsum_vec[window_width:] -
 *               cumsum_vec[:-window_width]) / window_width
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]             # <<<<<<<<<<<<<<
 *     return ma_vec
 * 
 */
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 311, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_data, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_ma_vec_array), 1, 0, NULL, NULL, &__pyx_slice__3, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 311, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GetSlice(__pyx_v_data, -1L, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ma_vec = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":312
 *               cumsum_vec[:-window_width]) / window_width
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]
 *     return ma_vec             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ma_vec);
  __pyx_r = __pyx_v_ma_vec;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":305
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def meanfilt(list data, int window_width):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray cumsum_vec, ma_vec_array
 *     cdef list ma_vec
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("biosaur2.cutils.meanfilt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cumsum_vec);
  __Pyx_XDECREF((PyObject *)__pyx_v_ma_vec_array);
  __Pyx_XDECREF(__pyx_v_ma_vec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":318
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks(dict hills_dict, list data_for_analyse_tmp, dict args, dict counter_hills_idx, list sorted_idx_child_process, np.ndarray sorted_idx_array_child_process, int nproc, int checked_id):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_17split_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_16split_peaks[] = "split_peaks(dict hills_dict, list data_for_analyse_tmp, dict args, dict counter_hills_idx, list sorted_idx_child_process, ndarray sorted_idx_array_child_process, int nproc, int checked_id)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_17split_peaks = {"split_peaks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_17split_peaks, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_16split_peaks};
static PyObject *__pyx_pw_8biosaur2_6cutils_17split_peaks(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hills_dict = 0;
  PyObject *__pyx_v_data_for_analyse_tmp = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_counter_hills_idx = 0;
  PyObject *__pyx_v_sorted_idx_child_process = 0;
  PyArrayObject *__pyx_v_sorted_idx_array_child_process = 0;
  CYTHON_UNUSED int __pyx_v_nproc;
  int __pyx_v_checked_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("split_peaks (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hills_dict,&__pyx_n_s_data_for_analyse_tmp,&__pyx_n_s_args,&__pyx_n_s_counter_hills_idx,&__pyx_n_s_sorted_idx_child_process,&__pyx_n_s_sorted_idx_array_child_process,&__pyx_n_s_nproc,&__pyx_n_s_checked_id,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hills_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_for_analyse_tmp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 1); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 2); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_counter_hills_idx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 3); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sorted_idx_child_process)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 4); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sorted_idx_array_child_process)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 5); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nproc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 6); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_checked_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, 7); __PYX_ERR(0, 318, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "split_peaks") < 0)) __PYX_ERR(0, 318, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_hills_dict = ((PyObject*)values[0]);
    __pyx_v_data_for_analyse_tmp = ((PyObject*)values[1]);
    __pyx_v_args = ((PyObject*)values[2]);
    __pyx_v_counter_hills_idx = ((PyObject*)values[3]);
    __pyx_v_sorted_idx_child_process = ((PyObject*)values[4]);
    __pyx_v_sorted_idx_array_child_process = ((PyArrayObject *)values[5]);
    __pyx_v_nproc = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_nproc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L3_error)
    __pyx_v_checked_id = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_checked_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("split_peaks", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 318, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.split_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hills_dict), (&PyDict_Type), 1, "hills_dict", 1))) __PYX_ERR(0, 318, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data_for_analyse_tmp), (&PyList_Type), 1, "data_for_analyse_tmp", 1))) __PYX_ERR(0, 318, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyDict_Type), 1, "args", 1))) __PYX_ERR(0, 318, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_counter_hills_idx), (&PyDict_Type), 1, "counter_hills_idx", 1))) __PYX_ERR(0, 318, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sorted_idx_child_process), (&PyList_Type), 1, "sorted_idx_child_process", 1))) __PYX_ERR(0, 318, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sorted_idx_array_child_process), __pyx_ptype_5numpy_ndarray, 1, "sorted_idx_array_child_process", 0))) __PYX_ERR(0, 318, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_16split_peaks(__pyx_self, __pyx_v_hills_dict, __pyx_v_data_for_analyse_tmp, __pyx_v_args, __pyx_v_counter_hills_idx, __pyx_v_sorted_idx_child_process, __pyx_v_sorted_idx_array_child_process, __pyx_v_nproc, __pyx_v_checked_id);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_16split_peaks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, PyObject *__pyx_v_data_for_analyse_tmp, PyObject *__pyx_v_args, PyObject *__pyx_v_counter_hills_idx, PyObject *__pyx_v_sorted_idx_child_process, PyArrayObject *__pyx_v_sorted_idx_array_child_process, CYTHON_UNUSED int __pyx_v_nproc, int __pyx_v_checked_id) {
  float __pyx_v_hillValleyFactor;
  float __pyx_v_min_val;
  float __pyx_v_mult_val;
  int __pyx_v_min_length_hill;
  int __pyx_v_idx_start;
  int __pyx_v_idx_end;
  int __pyx_v_cur_new_idx;
  int __pyx_v_hill_idx;
  int __pyx_v_hill_length;
  int __pyx_v_c_len;
  int __pyx_v_l_idx;
  PyArrayObject *__pyx_v_tmp_scans = 0;
  PyArrayObject *__pyx_v_tmp_orig_idx = 0;
  PyArrayObject *__pyx_v_new_index_list = 0;
  PyObject *__pyx_v_tmp_intensity = 0;
  PyObject *__pyx_v_smothed_intensity = 0;
  PyObject *__pyx_v_min_idx_list = 0;
  PyObject *__pyx_v_recheck_r_r = 0;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_l_r = NULL;
  PyObject *__pyx_v_r_r = NULL;
  PyObject *__pyx_v_include_factor = NULL;
  PyObject *__pyx_v_min_idx = NULL;
  PyObject *__pyx_v_end_idx = NULL;
  PyObject *__pyx_v_recheck_idx = NULL;
  PyObject *__pyx_v_orig_idx_val = NULL;
  PyObject *__pyx_v_scan_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  float __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("split_peaks", 0);

  /* "biosaur2/cutils.pyx":325
 *     cdef list tmp_intensity, smothed_intensity, min_idx_list, recheck_r_r
 * 
 *     new_index_list = copy(sorted_idx_array_child_process)             # <<<<<<<<<<<<<<
 * 
 *     hillValleyFactor = args['hvf']
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_sorted_idx_array_child_process)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_sorted_idx_array_child_process));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 325, __pyx_L1_error)
  __pyx_v_new_index_list = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":327
 *     new_index_list = copy(sorted_idx_array_child_process)
 * 
 *     hillValleyFactor = args['hvf']             # <<<<<<<<<<<<<<
 *     min_length_hill = args['minlh']
 *     min_length_hill = max(2, min_length_hill)
 */
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 327, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_hvf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hillValleyFactor = __pyx_t_4;

  /* "biosaur2/cutils.pyx":328
 * 
 *     hillValleyFactor = args['hvf']
 *     min_length_hill = args['minlh']             # <<<<<<<<<<<<<<
 *     min_length_hill = max(2, min_length_hill)
 * 
 */
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 328, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_minlh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_min_length_hill = __pyx_t_5;

  /* "biosaur2/cutils.pyx":329
 *     hillValleyFactor = args['hvf']
 *     min_length_hill = args['minlh']
 *     min_length_hill = max(2, min_length_hill)             # <<<<<<<<<<<<<<
 * 
 *     idx_start = 0
 */
  __pyx_t_5 = __pyx_v_min_length_hill;
  __pyx_t_6 = 2;
  if (((__pyx_t_5 > __pyx_t_6) != 0)) {
    __pyx_t_7 = __pyx_t_5;
  } else {
    __pyx_t_7 = __pyx_t_6;
  }
  __pyx_v_min_length_hill = __pyx_t_7;

  /* "biosaur2/cutils.pyx":331
 *     min_length_hill = max(2, min_length_hill)
 * 
 *     idx_start = 0             # <<<<<<<<<<<<<<
 *     idx_end = 0
 * 
 */
  __pyx_v_idx_start = 0;

  /* "biosaur2/cutils.pyx":332
 * 
 *     idx_start = 0
 *     idx_end = 0             # <<<<<<<<<<<<<<
 * 
 *     cur_new_idx = checked_id + max(sorted_idx_child_process) + 1
 */
  __pyx_v_idx_end = 0;

  /* "biosaur2/cutils.pyx":334
 *     idx_end = 0
 * 
 *     cur_new_idx = checked_id + max(sorted_idx_child_process) + 1             # <<<<<<<<<<<<<<
 * 
 *     for hill_idx in sorted_idx_child_process:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_checked_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_sorted_idx_child_process); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cur_new_idx = __pyx_t_5;

  /* "biosaur2/cutils.pyx":336
 *     cur_new_idx = checked_id + max(sorted_idx_child_process) + 1
 * 
 *     for hill_idx in sorted_idx_child_process:             # <<<<<<<<<<<<<<
 * 
 *         hill_length = counter_hills_idx[hill_idx]
 */
  if (unlikely(__pyx_v_sorted_idx_child_process == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 336, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_sorted_idx_child_process; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
  for (;;) {
    if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_hill_idx = __pyx_t_5;

    /* "biosaur2/cutils.pyx":338
 *     for hill_idx in sorted_idx_child_process:
 * 
 *         hill_length = counter_hills_idx[hill_idx]             # <<<<<<<<<<<<<<
 *         idx_end = idx_start + hill_length
 * 
 */
    if (unlikely(__pyx_v_counter_hills_idx == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 338, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_hill_idx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_counter_hills_idx, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_hill_length = __pyx_t_5;

    /* "biosaur2/cutils.pyx":339
 * 
 *         hill_length = counter_hills_idx[hill_idx]
 *         idx_end = idx_start + hill_length             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_idx_end = (__pyx_v_idx_start + __pyx_v_hill_length);

    /* "biosaur2/cutils.pyx":342
 * 
 * 
 *         if hill_length >= min_length_hill * 2:             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_9 = ((__pyx_v_hill_length >= (__pyx_v_min_length_hill * 2)) != 0);
    if (__pyx_t_9) {

      /* "biosaur2/cutils.pyx":345
 * 
 * 
 *             tmp_scans = hills_dict['scan_idx_array'][checked_id+idx_start:checked_id+idx_end]             # <<<<<<<<<<<<<<
 *             tmp_orig_idx = hills_dict['orig_idx_array'][checked_id+idx_start:checked_id+idx_end]
 *             tmp_intensity = [data_for_analyse_tmp[scan_val][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 345, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, (__pyx_v_checked_id + __pyx_v_idx_start), (__pyx_v_checked_id + __pyx_v_idx_end), NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_tmp_scans, ((PyArrayObject *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "biosaur2/cutils.pyx":346
 * 
 *             tmp_scans = hills_dict['scan_idx_array'][checked_id+idx_start:checked_id+idx_end]
 *             tmp_orig_idx = hills_dict['orig_idx_array'][checked_id+idx_start:checked_id+idx_end]             # <<<<<<<<<<<<<<
 *             tmp_intensity = [data_for_analyse_tmp[scan_val][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 * 
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 346, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_3, (__pyx_v_checked_id + __pyx_v_idx_start), (__pyx_v_checked_id + __pyx_v_idx_end), NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_tmp_orig_idx, ((PyArrayObject *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":347
 *             tmp_scans = hills_dict['scan_idx_array'][checked_id+idx_start:checked_id+idx_end]
 *             tmp_orig_idx = hills_dict['orig_idx_array'][checked_id+idx_start:checked_id+idx_end]
 *             tmp_intensity = [data_for_analyse_tmp[scan_val][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]             # <<<<<<<<<<<<<<
 * 
 *             smothed_intensity = meanfilt(tmp_intensity, 2)
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(((PyObject *)__pyx_v_tmp_orig_idx));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_tmp_orig_idx));
      PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_tmp_orig_idx));
      __Pyx_INCREF(((PyObject *)__pyx_v_tmp_scans));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_tmp_scans));
      PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_tmp_scans));
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
        __pyx_t_3 = __pyx_t_10; __Pyx_INCREF(__pyx_t_3); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_12 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 347, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_11); __Pyx_INCREF(__pyx_t_10); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 347, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_11); __Pyx_INCREF(__pyx_t_10); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 347, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_12(__pyx_t_3);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 347, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
          PyObject* sequence = __pyx_t_10;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 347, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_13 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_14 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_13 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_14 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          #else
          __pyx_t_13 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 347, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 347, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_15 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 347, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
          index = 0; __pyx_t_13 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_13)) goto __pyx_L8_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_13);
          index = 1; __pyx_t_14 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_14)) goto __pyx_L8_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_14);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 2) < 0) __PYX_ERR(0, 347, __pyx_L1_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L9_unpacking_done;
          __pyx_L8_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 347, __pyx_L1_error)
          __pyx_L9_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_orig_idx_val, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_XDECREF_SET(__pyx_v_scan_val, __pyx_t_14);
        __pyx_t_14 = 0;
        if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 347, __pyx_L1_error)
        }
        __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_data_for_analyse_tmp, __pyx_v_scan_val); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_v_orig_idx_val); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_14))) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tmp_intensity, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":349
 *             tmp_intensity = [data_for_analyse_tmp[scan_val][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 * 
 *             smothed_intensity = meanfilt(tmp_intensity, 2)             # <<<<<<<<<<<<<<
 *             c_len = hill_length - min_length_hill
 *             idx = int(min_length_hill) - 1
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_meanfilt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_tmp_intensity, __pyx_int_2};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_tmp_intensity, __pyx_int_2};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_14) {
          __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_14); __pyx_t_14 = NULL;
        }
        __Pyx_INCREF(__pyx_v_tmp_intensity);
        __Pyx_GIVEREF(__pyx_v_tmp_intensity);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_5, __pyx_v_tmp_intensity);
        __Pyx_INCREF(__pyx_int_2);
        __Pyx_GIVEREF(__pyx_int_2);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_5, __pyx_int_2);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 349, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_smothed_intensity, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":350
 * 
 *             smothed_intensity = meanfilt(tmp_intensity, 2)
 *             c_len = hill_length - min_length_hill             # <<<<<<<<<<<<<<
 *             idx = int(min_length_hill) - 1
 *             min_idx_list = []
 */
      __pyx_v_c_len = (__pyx_v_hill_length - __pyx_v_min_length_hill);

      /* "biosaur2/cutils.pyx":351
 *             smothed_intensity = meanfilt(tmp_intensity, 2)
 *             c_len = hill_length - min_length_hill
 *             idx = int(min_length_hill) - 1             # <<<<<<<<<<<<<<
 *             min_idx_list = []
 *             min_val = 0
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":352
 *             c_len = hill_length - min_length_hill
 *             idx = int(min_length_hill) - 1
 *             min_idx_list = []             # <<<<<<<<<<<<<<
 *             min_val = 0
 *             l_idx = 0
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_min_idx_list, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":353
 *             idx = int(min_length_hill) - 1
 *             min_idx_list = []
 *             min_val = 0             # <<<<<<<<<<<<<<
 *             l_idx = 0
 *             recheck_r_r = []
 */
      __pyx_v_min_val = 0.0;

      /* "biosaur2/cutils.pyx":354
 *             min_idx_list = []
 *             min_val = 0
 *             l_idx = 0             # <<<<<<<<<<<<<<
 *             recheck_r_r = []
 * 
 */
      __pyx_v_l_idx = 0;

      /* "biosaur2/cutils.pyx":355
 *             min_val = 0
 *             l_idx = 0
 *             recheck_r_r = []             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_recheck_r_r, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":358
 * 
 * 
 *             while idx <= c_len:             # <<<<<<<<<<<<<<
 * 
 *                 if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:
 */
      while (1) {
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_c_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!__pyx_t_9) break;

        /* "biosaur2/cutils.pyx":360
 *             while idx <= c_len:
 * 
 *                 if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                     l_idx = min_idx_list[-1]
 * 
 */
        __pyx_t_11 = PyList_GET_SIZE(__pyx_v_min_idx_list); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 360, __pyx_L1_error)
        __pyx_t_17 = (__pyx_t_11 != 0);
        if (__pyx_t_17) {
        } else {
          __pyx_t_9 = __pyx_t_17;
          goto __pyx_L13_bool_binop_done;
        }
        __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_min_idx_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_10, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = __pyx_t_17;
        __pyx_L13_bool_binop_done:;
        if (__pyx_t_9) {

          /* "biosaur2/cutils.pyx":361
 * 
 *                 if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:
 *                     l_idx = min_idx_list[-1]             # <<<<<<<<<<<<<<
 * 
 *                 l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 */
          __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_min_idx_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 361, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_l_idx = __pyx_t_5;

          /* "biosaur2/cutils.pyx":360
 *             while idx <= c_len:
 * 
 *                 if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                     l_idx = min_idx_list[-1]
 * 
 */
        }

        /* "biosaur2/cutils.pyx":363
 *                     l_idx = min_idx_list[-1]
 * 
 *                 l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])             # <<<<<<<<<<<<<<
 *                 if l_r >= hillValleyFactor:
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 */
        if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 363, __pyx_L1_error)
        }
        __Pyx_INCREF(__pyx_v_idx);
        __pyx_t_1 = __pyx_v_idx;
        __pyx_t_9 = (__pyx_t_1 == Py_None);
        if (__pyx_t_9) {
          __pyx_t_11 = PY_SSIZE_T_MAX;
        } else {
          __pyx_t_18 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_18 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 363, __pyx_L1_error)
          __pyx_t_11 = __pyx_t_18;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_smothed_intensity, __pyx_v_l_idx, __pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 363, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_smothed_intensity, __pyx_v_idx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_l_r, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "biosaur2/cutils.pyx":364
 * 
 *                 l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 *                 if l_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                     if r_r >= hillValleyFactor:
 */
        __pyx_t_1 = PyFloat_FromDouble(__pyx_v_hillValleyFactor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PyObject_RichCompare(__pyx_v_l_r, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_9) {

          /* "biosaur2/cutils.pyx":365
 *                 l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 *                 if l_r >= hillValleyFactor:
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])             # <<<<<<<<<<<<<<
 *                     if r_r >= hillValleyFactor:
 *                         mult_val = l_r * r_r
 */
          if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 365, __pyx_L1_error)
          }
          __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_idx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_9 = (__pyx_t_3 == Py_None);
          if (__pyx_t_9) {
            __pyx_t_11 = 0;
          } else {
            __pyx_t_18 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_18 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 365, __pyx_L1_error)
            __pyx_t_11 = __pyx_t_18;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyList_GetSlice(__pyx_v_smothed_intensity, __pyx_t_11, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 365, __pyx_L1_error)
          }
          __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_smothed_intensity, __pyx_v_idx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = __Pyx_PyNumber_Float(__pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF_SET(__pyx_v_r_r, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "biosaur2/cutils.pyx":366
 *                 if l_r >= hillValleyFactor:
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                     if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                         mult_val = l_r * r_r
 *                         include_factor = (1 if l_r > r_r else 0)
 */
          __pyx_t_3 = PyFloat_FromDouble(__pyx_v_hillValleyFactor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = PyObject_RichCompare(__pyx_v_r_r, __pyx_t_3, Py_GE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 366, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_9) {

            /* "biosaur2/cutils.pyx":367
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                     if r_r >= hillValleyFactor:
 *                         mult_val = l_r * r_r             # <<<<<<<<<<<<<<
 *                         include_factor = (1 if l_r > r_r else 0)
 *                         if (min_length_hill <= idx + include_factor <= c_len):
 */
            __pyx_t_10 = PyNumber_Multiply(__pyx_v_l_r, __pyx_v_r_r); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 367, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_10); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_v_mult_val = __pyx_t_4;

            /* "biosaur2/cutils.pyx":368
 *                     if r_r >= hillValleyFactor:
 *                         mult_val = l_r * r_r
 *                         include_factor = (1 if l_r > r_r else 0)             # <<<<<<<<<<<<<<
 *                         if (min_length_hill <= idx + include_factor <= c_len):
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 */
            __pyx_t_3 = PyObject_RichCompare(__pyx_v_l_r, __pyx_v_r_r, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
            __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 368, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (__pyx_t_9) {
              __Pyx_INCREF(__pyx_int_1);
              __pyx_t_10 = __pyx_int_1;
            } else {
              __Pyx_INCREF(__pyx_int_0);
              __pyx_t_10 = __pyx_int_0;
            }
            __Pyx_XDECREF_SET(__pyx_v_include_factor, __pyx_t_10);
            __pyx_t_10 = 0;

            /* "biosaur2/cutils.pyx":369
 *                         mult_val = l_r * r_r
 *                         include_factor = (1 if l_r > r_r else 0)
 *                         if (min_length_hill <= idx + include_factor <= c_len):             # <<<<<<<<<<<<<<
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                 min_idx_list.append(idx + include_factor)
 */
            __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 369, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_3 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 369, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = PyObject_RichCompare(__pyx_t_10, __pyx_t_3, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
            if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
              __Pyx_DECREF(__pyx_t_1);
              __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_c_len); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 369, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_14, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 369, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_9) {

              /* "biosaur2/cutils.pyx":370
 *                         include_factor = (1 if l_r > r_r else 0)
 *                         if (min_length_hill <= idx + include_factor <= c_len):
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                                 min_idx_list.append(idx + include_factor)
 *                                 recheck_r_r.append(idx)
 */
              __pyx_t_11 = PyList_GET_SIZE(__pyx_v_min_idx_list); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 370, __pyx_L1_error)
              __pyx_t_17 = ((!(__pyx_t_11 != 0)) != 0);
              if (!__pyx_t_17) {
              } else {
                __pyx_t_9 = __pyx_t_17;
                goto __pyx_L19_bool_binop_done;
              }
              __pyx_t_1 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_min_idx_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 370, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_14 = PyNumber_Add(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 370, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_10 = PyObject_RichCompare(__pyx_t_1, __pyx_t_14, Py_GE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 370, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 370, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_9 = __pyx_t_17;
              __pyx_L19_bool_binop_done:;
              if (__pyx_t_9) {

                /* "biosaur2/cutils.pyx":371
 *                         if (min_length_hill <= idx + include_factor <= c_len):
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                 min_idx_list.append(idx + include_factor)             # <<<<<<<<<<<<<<
 *                                 recheck_r_r.append(idx)
 *                                 min_val = mult_val
 */
                __pyx_t_10 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 371, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_min_idx_list, __pyx_t_10); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 371, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "biosaur2/cutils.pyx":372
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                 min_idx_list.append(idx + include_factor)
 *                                 recheck_r_r.append(idx)             # <<<<<<<<<<<<<<
 *                                 min_val = mult_val
 *                             elif mult_val > min_val:
 */
                __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_recheck_r_r, __pyx_v_idx); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 372, __pyx_L1_error)

                /* "biosaur2/cutils.pyx":373
 *                                 min_idx_list.append(idx + include_factor)
 *                                 recheck_r_r.append(idx)
 *                                 min_val = mult_val             # <<<<<<<<<<<<<<
 *                             elif mult_val > min_val:
 *                                 min_idx_list[-1] = idx + include_factor
 */
                __pyx_v_min_val = __pyx_v_mult_val;

                /* "biosaur2/cutils.pyx":370
 *                         include_factor = (1 if l_r > r_r else 0)
 *                         if (min_length_hill <= idx + include_factor <= c_len):
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                                 min_idx_list.append(idx + include_factor)
 *                                 recheck_r_r.append(idx)
 */
                goto __pyx_L18;
              }

              /* "biosaur2/cutils.pyx":374
 *                                 recheck_r_r.append(idx)
 *                                 min_val = mult_val
 *                             elif mult_val > min_val:             # <<<<<<<<<<<<<<
 *                                 min_idx_list[-1] = idx + include_factor
 *                                 recheck_r_r[-1] = idx
 */
              __pyx_t_9 = ((__pyx_v_mult_val > __pyx_v_min_val) != 0);
              if (__pyx_t_9) {

                /* "biosaur2/cutils.pyx":375
 *                                 min_val = mult_val
 *                             elif mult_val > min_val:
 *                                 min_idx_list[-1] = idx + include_factor             # <<<<<<<<<<<<<<
 *                                 recheck_r_r[-1] = idx
 *                                 min_val = mult_val
 */
                __pyx_t_10 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 375, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_10);
                if (unlikely(__Pyx_SetItemInt(__pyx_v_min_idx_list, -1L, __pyx_t_10, long, 1, __Pyx_PyInt_From_long, 1, 1, 0) < 0)) __PYX_ERR(0, 375, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

                /* "biosaur2/cutils.pyx":376
 *                             elif mult_val > min_val:
 *                                 min_idx_list[-1] = idx + include_factor
 *                                 recheck_r_r[-1] = idx             # <<<<<<<<<<<<<<
 *                                 min_val = mult_val
 *                 idx += 1
 */
                if (unlikely(__Pyx_SetItemInt(__pyx_v_recheck_r_r, -1L, __pyx_v_idx, long, 1, __Pyx_PyInt_From_long, 1, 1, 0) < 0)) __PYX_ERR(0, 376, __pyx_L1_error)

                /* "biosaur2/cutils.pyx":377
 *                                 min_idx_list[-1] = idx + include_factor
 *                                 recheck_r_r[-1] = idx
 *                                 min_val = mult_val             # <<<<<<<<<<<<<<
 *                 idx += 1
 *             if len(min_idx_list):
 */
                __pyx_v_min_val = __pyx_v_mult_val;

                /* "biosaur2/cutils.pyx":374
 *                                 recheck_r_r.append(idx)
 *                                 min_val = mult_val
 *                             elif mult_val > min_val:             # <<<<<<<<<<<<<<
 *                                 min_idx_list[-1] = idx + include_factor
 *                                 recheck_r_r[-1] = idx
 */
              }
              __pyx_L18:;

              /* "biosaur2/cutils.pyx":369
 *                         mult_val = l_r * r_r
 *                         include_factor = (1 if l_r > r_r else 0)
 *                         if (min_length_hill <= idx + include_factor <= c_len):             # <<<<<<<<<<<<<<
 *                             if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                 min_idx_list.append(idx + include_factor)
 */
            }

            /* "biosaur2/cutils.pyx":366
 *                 if l_r >= hillValleyFactor:
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                     if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                         mult_val = l_r * r_r
 *                         include_factor = (1 if l_r > r_r else 0)
 */
          }

          /* "biosaur2/cutils.pyx":364
 * 
 *                 l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 *                 if l_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                     r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                     if r_r >= hillValleyFactor:
 */
        }

        /* "biosaur2/cutils.pyx":378
 *                                 recheck_r_r[-1] = idx
 *                                 min_val = mult_val
 *                 idx += 1             # <<<<<<<<<<<<<<
 *             if len(min_idx_list):
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 */
        __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_v_idx, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 378, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF_SET(__pyx_v_idx, __pyx_t_10);
        __pyx_t_10 = 0;
      }

      /* "biosaur2/cutils.pyx":379
 *                                 min_val = mult_val
 *                 idx += 1
 *             if len(min_idx_list):             # <<<<<<<<<<<<<<
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 */
      __pyx_t_11 = PyList_GET_SIZE(__pyx_v_min_idx_list); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 379, __pyx_L1_error)
      __pyx_t_9 = (__pyx_t_11 != 0);
      if (__pyx_t_9) {

        /* "biosaur2/cutils.pyx":380
 *                 idx += 1
 *             if len(min_idx_list):
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):             # <<<<<<<<<<<<<<
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                     if r_r >= hillValleyFactor:
 */
        __pyx_t_10 = __Pyx_PyList_GetSlice(__pyx_v_min_idx_list, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_14 = __Pyx_PyInt_From_int((__pyx_v_idx_start + __pyx_v_hill_length)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_14);
        PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_14);
        __pyx_t_14 = 0;
        __pyx_t_14 = PyNumber_Add(__pyx_t_10, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_min_idx_list);
        __Pyx_GIVEREF(__pyx_v_min_idx_list);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_min_idx_list);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_14);
        __Pyx_INCREF(__pyx_v_recheck_r_r);
        __Pyx_GIVEREF(__pyx_v_recheck_r_r);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_recheck_r_r);
        __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
          __pyx_t_1 = __pyx_t_14; __Pyx_INCREF(__pyx_t_1); __pyx_t_11 = 0;
          __pyx_t_12 = NULL;
        } else {
          __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 380, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        for (;;) {
          if (likely(!__pyx_t_12)) {
            if (likely(PyList_CheckExact(__pyx_t_1))) {
              if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_1)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_14 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_14); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 380, __pyx_L1_error)
              #else
              __pyx_t_14 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 380, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              #endif
            } else {
              if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_14); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 380, __pyx_L1_error)
              #else
              __pyx_t_14 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 380, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              #endif
            }
          } else {
            __pyx_t_14 = __pyx_t_12(__pyx_t_1);
            if (unlikely(!__pyx_t_14)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 380, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_14);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
            PyObject* sequence = __pyx_t_14;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 3)) {
              if (size > 3) __Pyx_RaiseTooManyValuesError(3);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 380, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
              __pyx_t_13 = PyTuple_GET_ITEM(sequence, 2); 
            } else {
              __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
              __pyx_t_13 = PyList_GET_ITEM(sequence, 2); 
            }
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_13);
            #else
            __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 380, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_13 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 380, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            #endif
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_15 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 380, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
            index = 0; __pyx_t_10 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_10)) goto __pyx_L24_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_10);
            index = 1; __pyx_t_3 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_3)) goto __pyx_L24_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            index = 2; __pyx_t_13 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_13)) goto __pyx_L24_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_13);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 3) < 0) __PYX_ERR(0, 380, __pyx_L1_error)
            __pyx_t_16 = NULL;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L25_unpacking_done;
            __pyx_L24_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_16 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 380, __pyx_L1_error)
            __pyx_L25_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_min_idx, __pyx_t_10);
          __pyx_t_10 = 0;
          __Pyx_XDECREF_SET(__pyx_v_end_idx, __pyx_t_3);
          __pyx_t_3 = 0;
          __Pyx_XDECREF_SET(__pyx_v_recheck_idx, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "biosaur2/cutils.pyx":381
 *             if len(min_idx_list):
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])             # <<<<<<<<<<<<<<
 *                     if r_r >= hillValleyFactor:
 *                         new_index_list[idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 */
          if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 381, __pyx_L1_error)
          }
          __pyx_t_14 = __Pyx_PyInt_AddObjC(__pyx_v_recheck_idx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_9 = (__pyx_t_14 == Py_None);
          if (__pyx_t_9) {
            __pyx_t_18 = 0;
          } else {
            __pyx_t_20 = __Pyx_PyIndex_AsSsize_t(__pyx_t_14); if (unlikely((__pyx_t_20 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L1_error)
            __pyx_t_18 = __pyx_t_20;
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_INCREF(__pyx_v_end_idx);
          __pyx_t_14 = __pyx_v_end_idx;
          __pyx_t_9 = (__pyx_t_14 == Py_None);
          if (__pyx_t_9) {
            __pyx_t_20 = PY_SSIZE_T_MAX;
          } else {
            __pyx_t_21 = __Pyx_PyIndex_AsSsize_t(__pyx_t_14); if (unlikely((__pyx_t_21 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L1_error)
            __pyx_t_20 = __pyx_t_21;
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyList_GetSlice(__pyx_v_smothed_intensity, __pyx_t_18, __pyx_t_20); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_13 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 381, __pyx_L1_error)
          }
          __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_smothed_intensity, __pyx_v_recheck_idx); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyNumber_Divide(__pyx_t_13, __pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 381, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF_SET(__pyx_v_r_r, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "biosaur2/cutils.pyx":382
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                     if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                         new_index_list[idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 *                         cur_new_idx += 1
 */
          __pyx_t_14 = PyFloat_FromDouble(__pyx_v_hillValleyFactor); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_3 = PyObject_RichCompare(__pyx_v_r_r, __pyx_t_14, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_9) {

            /* "biosaur2/cutils.pyx":383
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                     if r_r >= hillValleyFactor:
 *                         new_index_list[idx_start+min_idx:idx_start+hill_length] = cur_new_idx             # <<<<<<<<<<<<<<
 *                         cur_new_idx += 1
 * 
 */
            __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cur_new_idx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 383, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_idx_start); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 383, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_13 = PyNumber_Add(__pyx_t_14, __pyx_v_min_idx); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 383, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (__Pyx_PyObject_SetSlice(((PyObject *)__pyx_v_new_index_list), __pyx_t_3, 0, (__pyx_v_idx_start + __pyx_v_hill_length), &__pyx_t_13, NULL, NULL, 0, 1, 1) < 0) __PYX_ERR(0, 383, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "biosaur2/cutils.pyx":384
 *                     if r_r >= hillValleyFactor:
 *                         new_index_list[idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 *                         cur_new_idx += 1             # <<<<<<<<<<<<<<
 * 
 *         idx_start = idx_end
 */
            __pyx_v_cur_new_idx = (__pyx_v_cur_new_idx + 1);

            /* "biosaur2/cutils.pyx":382
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                     if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                         new_index_list[idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 *                         cur_new_idx += 1
 */
          }

          /* "biosaur2/cutils.pyx":380
 *                 idx += 1
 *             if len(min_idx_list):
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):             # <<<<<<<<<<<<<<
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                     if r_r >= hillValleyFactor:
 */
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "biosaur2/cutils.pyx":379
 *                                 min_val = mult_val
 *                 idx += 1
 *             if len(min_idx_list):             # <<<<<<<<<<<<<<
 *                 for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                     r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 */
      }

      /* "biosaur2/cutils.pyx":342
 * 
 * 
 *         if hill_length >= min_length_hill * 2:             # <<<<<<<<<<<<<<
 * 
 * 
 */
    }

    /* "biosaur2/cutils.pyx":386
 *                         cur_new_idx += 1
 * 
 *         idx_start = idx_end             # <<<<<<<<<<<<<<
 * 
 *     return new_index_list
 */
    __pyx_v_idx_start = __pyx_v_idx_end;

    /* "biosaur2/cutils.pyx":336
 *     cur_new_idx = checked_id + max(sorted_idx_child_process) + 1
 * 
 *     for hill_idx in sorted_idx_child_process:             # <<<<<<<<<<<<<<
 * 
 *         hill_length = counter_hills_idx[hill_idx]
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":388
 *         idx_start = idx_end
 * 
 *     return new_index_list             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_new_index_list));
  __pyx_r = ((PyObject *)__pyx_v_new_index_list);
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":318
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks(dict hills_dict, list data_for_analyse_tmp, dict args, dict counter_hills_idx, list sorted_idx_child_process, np.ndarray sorted_idx_array_child_process, int nproc, int checked_id):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("biosaur2.cutils.split_peaks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp_scans);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp_orig_idx);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_index_list);
  __Pyx_XDECREF(__pyx_v_tmp_intensity);
  __Pyx_XDECREF(__pyx_v_smothed_intensity);
  __Pyx_XDECREF(__pyx_v_min_idx_list);
  __Pyx_XDECREF(__pyx_v_recheck_r_r);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_l_r);
  __Pyx_XDECREF(__pyx_v_r_r);
  __Pyx_XDECREF(__pyx_v_include_factor);
  __Pyx_XDECREF(__pyx_v_min_idx);
  __Pyx_XDECREF(__pyx_v_end_idx);
  __Pyx_XDECREF(__pyx_v_recheck_idx);
  __Pyx_XDECREF(__pyx_v_orig_idx_val);
  __Pyx_XDECREF(__pyx_v_scan_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":396
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks_old(dict hills_dict, list data_for_analyse_tmp, dict args):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_19split_peaks_old(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_18split_peaks_old[] = "split_peaks_old(dict hills_dict, list data_for_analyse_tmp, dict args)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_19split_peaks_old = {"split_peaks_old", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_19split_peaks_old, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_18split_peaks_old};
static PyObject *__pyx_pw_8biosaur2_6cutils_19split_peaks_old(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hills_dict = 0;
  PyObject *__pyx_v_data_for_analyse_tmp = 0;
  PyObject *__pyx_v_args = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("split_peaks_old (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hills_dict,&__pyx_n_s_data_for_analyse_tmp,&__pyx_n_s_args,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hills_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_for_analyse_tmp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks_old", 1, 3, 3, 1); __PYX_ERR(0, 396, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split_peaks_old", 1, 3, 3, 2); __PYX_ERR(0, 396, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "split_peaks_old") < 0)) __PYX_ERR(0, 396, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_hills_dict = ((PyObject*)values[0]);
    __pyx_v_data_for_analyse_tmp = ((PyObject*)values[1]);
    __pyx_v_args = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("split_peaks_old", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 396, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.split_peaks_old", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hills_dict), (&PyDict_Type), 1, "hills_dict", 1))) __PYX_ERR(0, 396, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data_for_analyse_tmp), (&PyList_Type), 1, "data_for_analyse_tmp", 1))) __PYX_ERR(0, 396, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyDict_Type), 1, "args", 1))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_18split_peaks_old(__pyx_self, __pyx_v_hills_dict, __pyx_v_data_for_analyse_tmp, __pyx_v_args);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_18split_peaks_old(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, PyObject *__pyx_v_data_for_analyse_tmp, PyObject *__pyx_v_args) {
  float __pyx_v_hillValleyFactor;
  float __pyx_v_min_val;
  float __pyx_v_mult_val;
  int __pyx_v_min_length_hill;
  int __pyx_v_idx_start;
  int __pyx_v_idx_end;
  int __pyx_v_cur_new_idx;
  CYTHON_UNUSED int __pyx_v_idx_1;
  int __pyx_v_hill_idx;
  int __pyx_v_hill_length;
  int __pyx_v_c_len;
  int __pyx_v_l_idx;
  PyArrayObject *__pyx_v_tmp_hill_length = 0;
  PyArrayObject *__pyx_v_idx_minl = 0;
  PyArrayObject *__pyx_v_idx_sort = 0;
  PyArrayObject *__pyx_v_tmp_scans = 0;
  PyArrayObject *__pyx_v_tmp_orig_idx = 0;
  PyObject *__pyx_v_tmp_intensity = 0;
  PyObject *__pyx_v_smothed_intensity = 0;
  PyObject *__pyx_v_min_idx_list = 0;
  PyObject *__pyx_v_recheck_r_r = 0;
  PyObject *__pyx_v_counter_hills_idx = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_l_r = NULL;
  PyObject *__pyx_v_r_r = NULL;
  PyObject *__pyx_v_include_factor = NULL;
  PyObject *__pyx_v_min_idx = NULL;
  PyObject *__pyx_v_end_idx = NULL;
  PyObject *__pyx_v_recheck_idx = NULL;
  PyObject *__pyx_v_orig_idx_val = NULL;
  PyObject *__pyx_v_scan_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  float __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("split_peaks_old", 0);

  /* "biosaur2/cutils.pyx":403
 *     cdef list tmp_intensity, smothed_intensity, min_idx_list, recheck_r_r
 * 
 *     hillValleyFactor = args['hvf']             # <<<<<<<<<<<<<<
 *     min_length_hill = args['minlh']
 *     min_length_hill = max(2, min_length_hill)
 */
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 403, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_hvf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hillValleyFactor = __pyx_t_2;

  /* "biosaur2/cutils.pyx":404
 * 
 *     hillValleyFactor = args['hvf']
 *     min_length_hill = args['minlh']             # <<<<<<<<<<<<<<
 *     min_length_hill = max(2, min_length_hill)
 * 
 */
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 404, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_minlh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_min_length_hill = __pyx_t_3;

  /* "biosaur2/cutils.pyx":405
 *     hillValleyFactor = args['hvf']
 *     min_length_hill = args['minlh']
 *     min_length_hill = max(2, min_length_hill)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __pyx_v_min_length_hill;
  __pyx_t_4 = 2;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_min_length_hill = __pyx_t_5;

  /* "biosaur2/cutils.pyx":408
 * 
 * 
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])             # <<<<<<<<<<<<<<
 *     hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 408, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 408, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":409
 * 
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
 *     hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])             # <<<<<<<<<<<<<<
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 409, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 409, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":410
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
 *     hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])             # <<<<<<<<<<<<<<
 * 
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 410, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 410, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":412
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])
 * 
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])             # <<<<<<<<<<<<<<
 * 
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Counter); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 412, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_counter_hills_idx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":414
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])
 * 
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])             # <<<<<<<<<<<<<<
 *     idx_minl = tmp_hill_length >= min_length_hill
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 414, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
    __pyx_t_9 = __pyx_t_8; __Pyx_INCREF(__pyx_t_9); __pyx_t_10 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 414, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_9))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 414, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 414, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_11(__pyx_t_9);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 414, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_hill_idx = __pyx_t_3;
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_counter_hills_idx, __pyx_v_hill_idx, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 414, __pyx_L1_error)
  __pyx_v_tmp_hill_length = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":415
 * 
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
 *     idx_minl = tmp_hill_length >= min_length_hill             # <<<<<<<<<<<<<<
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyObject_RichCompare(((PyObject *)__pyx_v_tmp_hill_length), __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 415, __pyx_L1_error)
  __pyx_v_idx_minl = ((PyArrayObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "biosaur2/cutils.pyx":416
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
 *     idx_minl = tmp_hill_length >= min_length_hill
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]             # <<<<<<<<<<<<<<
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 416, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_6, ((PyObject *)__pyx_v_idx_minl)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 416, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":417
 *     idx_minl = tmp_hill_length >= min_length_hill
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]             # <<<<<<<<<<<<<<
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 * 
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 417, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_1, ((PyObject *)__pyx_v_idx_minl)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 417, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_6) < 0)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "biosaur2/cutils.pyx":418
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]             # <<<<<<<<<<<<<<
 * 
 *     if len(hills_dict['orig_idx_array']):
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 418, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_6, ((PyObject *)__pyx_v_idx_minl)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 418, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":420
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 * 
 *     if len(hills_dict['orig_idx_array']):             # <<<<<<<<<<<<<<
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 420, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = (__pyx_t_10 != 0);
  if (__pyx_t_12) {

    /* "biosaur2/cutils.pyx":422
 *     if len(hills_dict['orig_idx_array']):
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))             # <<<<<<<<<<<<<<
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_argsort); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 422, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 422, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 422, __pyx_L1_error)
    }
    __pyx_t_13 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_max); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_14);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_14, function);
      }
    }
    __pyx_t_8 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_14);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyInt_AddObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyNumber_Add(__pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 422, __pyx_L1_error)
    __pyx_v_idx_sort = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":423
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]             # <<<<<<<<<<<<<<
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 423, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_1, ((PyObject *)__pyx_v_idx_sort)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 423, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_7) < 0)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "biosaur2/cutils.pyx":424
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]             # <<<<<<<<<<<<<<
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 424, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_7, ((PyObject *)__pyx_v_idx_sort)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 424, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":425
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 425, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_1, ((PyObject *)__pyx_v_idx_sort)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 425, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_7) < 0)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "biosaur2/cutils.pyx":427
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]
 * 
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))             # <<<<<<<<<<<<<<
 * 
 *         idx_start = 0
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 427, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PySequence_List(__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = ((PyObject*)__pyx_t_14);
    __pyx_t_14 = 0;
    __pyx_t_15 = PyList_Sort(__pyx_t_7); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 427, __pyx_L1_error)
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 427, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique, __pyx_t_7) < 0)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "biosaur2/cutils.pyx":429
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))
 * 
 *         idx_start = 0             # <<<<<<<<<<<<<<
 *         idx_end = 0
 * 
 */
    __pyx_v_idx_start = 0;

    /* "biosaur2/cutils.pyx":430
 * 
 *         idx_start = 0
 *         idx_end = 0             # <<<<<<<<<<<<<<
 * 
 *         cur_new_idx = max(hills_dict['hills_idx_array_unique']) + 1
 */
    __pyx_v_idx_end = 0;

    /* "biosaur2/cutils.pyx":432
 *         idx_end = 0
 * 
 *         cur_new_idx = max(hills_dict['hills_idx_array_unique']) + 1             # <<<<<<<<<<<<<<
 * 
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 432, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_cur_new_idx = __pyx_t_3;

    /* "biosaur2/cutils.pyx":434
 *         cur_new_idx = max(hills_dict['hills_idx_array_unique']) + 1
 * 
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):             # <<<<<<<<<<<<<<
 * 
 *             hill_length = counter_hills_idx[hill_idx]
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 434, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_14 = __pyx_t_7; __Pyx_INCREF(__pyx_t_14); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_11 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 434, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_14))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_14)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_14, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 434, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_14, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 434, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_11(__pyx_t_14);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 434, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_hill_idx = __pyx_t_16;
      __pyx_v_idx_1 = __pyx_t_3;
      __pyx_t_3 = (__pyx_t_3 + 1);

      /* "biosaur2/cutils.pyx":436
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):
 * 
 *             hill_length = counter_hills_idx[hill_idx]             # <<<<<<<<<<<<<<
 *             idx_end = idx_start + hill_length
 * 
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_counter_hills_idx, __pyx_v_hill_idx, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_hill_length = __pyx_t_16;

      /* "biosaur2/cutils.pyx":437
 * 
 *             hill_length = counter_hills_idx[hill_idx]
 *             idx_end = idx_start + hill_length             # <<<<<<<<<<<<<<
 * 
 *             if hill_length >= min_length_hill * 2:
 */
      __pyx_v_idx_end = (__pyx_v_idx_start + __pyx_v_hill_length);

      /* "biosaur2/cutils.pyx":439
 *             idx_end = idx_start + hill_length
 * 
 *             if hill_length >= min_length_hill * 2:             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_12 = ((__pyx_v_hill_length >= (__pyx_v_min_length_hill * 2)) != 0);
      if (__pyx_t_12) {

        /* "biosaur2/cutils.pyx":442
 * 
 * 
 *                 tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]             # <<<<<<<<<<<<<<
 *                 tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]
 *                 tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 442, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_7, __pyx_v_idx_start, __pyx_v_idx_end, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 442, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_tmp_scans, ((PyArrayObject *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "biosaur2/cutils.pyx":443
 * 
 *                 tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]
 *                 tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]             # <<<<<<<<<<<<<<
 *                 tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 * 
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 443, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_t_1, __pyx_v_idx_start, __pyx_v_idx_end, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 443, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 443, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_tmp_orig_idx, ((PyArrayObject *)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "biosaur2/cutils.pyx":444
 *                 tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]
 *                 tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]
 *                 tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]             # <<<<<<<<<<<<<<
 * 
 *                 smothed_intensity = meanfilt(tmp_intensity, 2)
 */
        __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(((PyObject *)__pyx_v_tmp_orig_idx));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_tmp_orig_idx));
        PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_tmp_orig_idx));
        __Pyx_INCREF(((PyObject *)__pyx_v_tmp_scans));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_tmp_scans));
        PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_tmp_scans));
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 444, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
          __pyx_t_1 = __pyx_t_8; __Pyx_INCREF(__pyx_t_1); __pyx_t_17 = 0;
          __pyx_t_18 = NULL;
        } else {
          __pyx_t_17 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_18 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 444, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        for (;;) {
          if (likely(!__pyx_t_18)) {
            if (likely(PyList_CheckExact(__pyx_t_1))) {
              if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_1)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_17); __Pyx_INCREF(__pyx_t_8); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 444, __pyx_L1_error)
              #else
              __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 444, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              #endif
            } else {
              if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_17); __Pyx_INCREF(__pyx_t_8); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 444, __pyx_L1_error)
              #else
              __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 444, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              #endif
            }
          } else {
            __pyx_t_8 = __pyx_t_18(__pyx_t_1);
            if (unlikely(!__pyx_t_8)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 444, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_8);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
            PyObject* sequence = __pyx_t_8;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 444, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_9);
            #else
            __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 444, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 444, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_13 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 444, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_19 = Py_TYPE(__pyx_t_13)->tp_iternext;
            index = 0; __pyx_t_6 = __pyx_t_19(__pyx_t_13); if (unlikely(!__pyx_t_6)) goto __pyx_L11_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_6);
            index = 1; __pyx_t_9 = __pyx_t_19(__pyx_t_13); if (unlikely(!__pyx_t_9)) goto __pyx_L11_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_9);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_13), 2) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
            __pyx_t_19 = NULL;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L12_unpacking_done;
            __pyx_L11_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_19 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 444, __pyx_L1_error)
            __pyx_L12_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_orig_idx_val, __pyx_t_6);
          __pyx_t_6 = 0;
          __Pyx_XDECREF_SET(__pyx_v_scan_val, __pyx_t_9);
          __pyx_t_9 = 0;
          if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 444, __pyx_L1_error)
          }
          __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_data_for_analyse_tmp, __pyx_v_scan_val); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_8, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_v_orig_idx_val); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 444, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_tmp_intensity, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "biosaur2/cutils.pyx":446
 *                 tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 * 
 *                 smothed_intensity = meanfilt(tmp_intensity, 2)             # <<<<<<<<<<<<<<
 *                 c_len = hill_length - min_length_hill
 *                 idx = int(min_length_hill) - 1
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_meanfilt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = NULL;
        __pyx_t_16 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_16 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_tmp_intensity, __pyx_int_2};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 446, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_tmp_intensity, __pyx_int_2};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 446, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_9 = PyTuple_New(2+__pyx_t_16); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 446, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_v_tmp_intensity);
          __Pyx_GIVEREF(__pyx_v_tmp_intensity);
          PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_16, __pyx_v_tmp_intensity);
          __Pyx_INCREF(__pyx_int_2);
          __Pyx_GIVEREF(__pyx_int_2);
          PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_16, __pyx_int_2);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 446, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 446, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_smothed_intensity, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "biosaur2/cutils.pyx":447
 * 
 *                 smothed_intensity = meanfilt(tmp_intensity, 2)
 *                 c_len = hill_length - min_length_hill             # <<<<<<<<<<<<<<
 *                 idx = int(min_length_hill) - 1
 *                 min_idx_list = []
 */
        __pyx_v_c_len = (__pyx_v_hill_length - __pyx_v_min_length_hill);

        /* "biosaur2/cutils.pyx":448
 *                 smothed_intensity = meanfilt(tmp_intensity, 2)
 *                 c_len = hill_length - min_length_hill
 *                 idx = int(min_length_hill) - 1             # <<<<<<<<<<<<<<
 *                 min_idx_list = []
 *                 min_val = 0
 */
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 448, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 448, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "biosaur2/cutils.pyx":449
 *                 c_len = hill_length - min_length_hill
 *                 idx = int(min_length_hill) - 1
 *                 min_idx_list = []             # <<<<<<<<<<<<<<
 *                 min_val = 0
 *                 l_idx = 0
 */
        __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 449, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_XDECREF_SET(__pyx_v_min_idx_list, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "biosaur2/cutils.pyx":450
 *                 idx = int(min_length_hill) - 1
 *                 min_idx_list = []
 *                 min_val = 0             # <<<<<<<<<<<<<<
 *                 l_idx = 0
 *                 recheck_r_r = []
 */
        __pyx_v_min_val = 0.0;

        /* "biosaur2/cutils.pyx":451
 *                 min_idx_list = []
 *                 min_val = 0
 *                 l_idx = 0             # <<<<<<<<<<<<<<
 *                 recheck_r_r = []
 * 
 */
        __pyx_v_l_idx = 0;

        /* "biosaur2/cutils.pyx":452
 *                 min_val = 0
 *                 l_idx = 0
 *                 recheck_r_r = []             # <<<<<<<<<<<<<<
 * 
 *                 while idx <= c_len:
 */
        __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_XDECREF_SET(__pyx_v_recheck_r_r, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "biosaur2/cutils.pyx":454
 *                 recheck_r_r = []
 * 
 *                 while idx <= c_len:             # <<<<<<<<<<<<<<
 * 
 *                     if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:
 */
        while (1) {
          __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_c_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_7, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_12) break;

          /* "biosaur2/cutils.pyx":456
 *                 while idx <= c_len:
 * 
 *                     if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                         l_idx = min_idx_list[-1]
 * 
 */
          __pyx_t_17 = PyList_GET_SIZE(__pyx_v_min_idx_list); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 456, __pyx_L1_error)
          __pyx_t_20 = (__pyx_t_17 != 0);
          if (__pyx_t_20) {
          } else {
            __pyx_t_12 = __pyx_t_20;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_min_idx_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = PyNumber_Add(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_9, Py_GE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 456, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __pyx_t_20;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_12) {

            /* "biosaur2/cutils.pyx":457
 * 
 *                     if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:
 *                         l_idx = min_idx_list[-1]             # <<<<<<<<<<<<<<
 * 
 *                     l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 */
            __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_min_idx_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 457, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_v_l_idx = __pyx_t_16;

            /* "biosaur2/cutils.pyx":456
 *                 while idx <= c_len:
 * 
 *                     if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                         l_idx = min_idx_list[-1]
 * 
 */
          }

          /* "biosaur2/cutils.pyx":459
 *                         l_idx = min_idx_list[-1]
 * 
 *                     l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])             # <<<<<<<<<<<<<<
 *                     if l_r >= hillValleyFactor:
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 */
          if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 459, __pyx_L1_error)
          }
          __Pyx_INCREF(__pyx_v_idx);
          __pyx_t_7 = __pyx_v_idx;
          __pyx_t_12 = (__pyx_t_7 == Py_None);
          if (__pyx_t_12) {
            __pyx_t_17 = PY_SSIZE_T_MAX;
          } else {
            __pyx_t_21 = __Pyx_PyIndex_AsSsize_t(__pyx_t_7); if (unlikely((__pyx_t_21 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 459, __pyx_L1_error)
            __pyx_t_17 = __pyx_t_21;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyList_GetSlice(__pyx_v_smothed_intensity, __pyx_v_l_idx, __pyx_t_17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 459, __pyx_L1_error)
          }
          __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_smothed_intensity, __pyx_v_idx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF_SET(__pyx_v_l_r, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "biosaur2/cutils.pyx":460
 * 
 *                     l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 *                     if l_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                         if r_r >= hillValleyFactor:
 */
          __pyx_t_7 = PyFloat_FromDouble(__pyx_v_hillValleyFactor); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = PyObject_RichCompare(__pyx_v_l_r, __pyx_t_7, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_12) {

            /* "biosaur2/cutils.pyx":461
 *                     l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 *                     if l_r >= hillValleyFactor:
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])             # <<<<<<<<<<<<<<
 *                         if r_r >= hillValleyFactor:
 *                             mult_val = l_r * r_r
 */
            if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 461, __pyx_L1_error)
            }
            __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_idx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_12 = (__pyx_t_1 == Py_None);
            if (__pyx_t_12) {
              __pyx_t_17 = 0;
            } else {
              __pyx_t_21 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_21 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 461, __pyx_L1_error)
              __pyx_t_17 = __pyx_t_21;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_smothed_intensity, __pyx_t_17, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 461, __pyx_L1_error)
            }
            __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_smothed_intensity, __pyx_v_idx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_9 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_7, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_XDECREF_SET(__pyx_v_r_r, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "biosaur2/cutils.pyx":462
 *                     if l_r >= hillValleyFactor:
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                         if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                             mult_val = l_r * r_r
 *                             include_factor = (1 if l_r > r_r else 0)
 */
            __pyx_t_1 = PyFloat_FromDouble(__pyx_v_hillValleyFactor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_9 = PyObject_RichCompare(__pyx_v_r_r, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 462, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 462, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (__pyx_t_12) {

              /* "biosaur2/cutils.pyx":463
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                         if r_r >= hillValleyFactor:
 *                             mult_val = l_r * r_r             # <<<<<<<<<<<<<<
 *                             include_factor = (1 if l_r > r_r else 0)
 *                             if (min_length_hill <= idx + include_factor <= c_len):
 */
              __pyx_t_9 = PyNumber_Multiply(__pyx_v_l_r, __pyx_v_r_r); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 463, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_9); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __pyx_v_mult_val = __pyx_t_2;

              /* "biosaur2/cutils.pyx":464
 *                         if r_r >= hillValleyFactor:
 *                             mult_val = l_r * r_r
 *                             include_factor = (1 if l_r > r_r else 0)             # <<<<<<<<<<<<<<
 *                             if (min_length_hill <= idx + include_factor <= c_len):
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 */
              __pyx_t_1 = PyObject_RichCompare(__pyx_v_l_r, __pyx_v_r_r, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
              __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (__pyx_t_12) {
                __Pyx_INCREF(__pyx_int_1);
                __pyx_t_9 = __pyx_int_1;
              } else {
                __Pyx_INCREF(__pyx_int_0);
                __pyx_t_9 = __pyx_int_0;
              }
              __Pyx_XDECREF_SET(__pyx_v_include_factor, __pyx_t_9);
              __pyx_t_9 = 0;

              /* "biosaur2/cutils.pyx":465
 *                             mult_val = l_r * r_r
 *                             include_factor = (1 if l_r > r_r else 0)
 *                             if (min_length_hill <= idx + include_factor <= c_len):             # <<<<<<<<<<<<<<
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                     min_idx_list.append(idx + include_factor)
 */
              __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 465, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_1 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_7 = PyObject_RichCompare(__pyx_t_9, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 465, __pyx_L1_error)
              if (__Pyx_PyObject_IsTrue(__pyx_t_7)) {
                __Pyx_DECREF(__pyx_t_7);
                __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_c_len); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 465, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_8, Py_LE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 465, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              }
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (__pyx_t_12) {

                /* "biosaur2/cutils.pyx":466
 *                             include_factor = (1 if l_r > r_r else 0)
 *                             if (min_length_hill <= idx + include_factor <= c_len):
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                                     min_idx_list.append(idx + include_factor)
 *                                     recheck_r_r.append(idx)
 */
                __pyx_t_17 = PyList_GET_SIZE(__pyx_v_min_idx_list); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 466, __pyx_L1_error)
                __pyx_t_20 = ((!(__pyx_t_17 != 0)) != 0);
                if (!__pyx_t_20) {
                } else {
                  __pyx_t_12 = __pyx_t_20;
                  goto __pyx_L22_bool_binop_done;
                }
                __pyx_t_7 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 466, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_min_idx_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 466, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_8 = PyNumber_Add(__pyx_t_1, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 466, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_t_8, Py_GE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 466, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                __pyx_t_12 = __pyx_t_20;
                __pyx_L22_bool_binop_done:;
                if (__pyx_t_12) {

                  /* "biosaur2/cutils.pyx":467
 *                             if (min_length_hill <= idx + include_factor <= c_len):
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                     min_idx_list.append(idx + include_factor)             # <<<<<<<<<<<<<<
 *                                     recheck_r_r.append(idx)
 *                                     min_val = mult_val
 */
                  __pyx_t_9 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 467, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_min_idx_list, __pyx_t_9); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 467, __pyx_L1_error)
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

                  /* "biosaur2/cutils.pyx":468
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                     min_idx_list.append(idx + include_factor)
 *                                     recheck_r_r.append(idx)             # <<<<<<<<<<<<<<
 *                                     min_val = mult_val
 *                                 elif mult_val > min_val:
 */
                  __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_recheck_r_r, __pyx_v_idx); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 468, __pyx_L1_error)

                  /* "biosaur2/cutils.pyx":469
 *                                     min_idx_list.append(idx + include_factor)
 *                                     recheck_r_r.append(idx)
 *                                     min_val = mult_val             # <<<<<<<<<<<<<<
 *                                 elif mult_val > min_val:
 *                                     min_idx_list[-1] = idx + include_factor
 */
                  __pyx_v_min_val = __pyx_v_mult_val;

                  /* "biosaur2/cutils.pyx":466
 *                             include_factor = (1 if l_r > r_r else 0)
 *                             if (min_length_hill <= idx + include_factor <= c_len):
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:             # <<<<<<<<<<<<<<
 *                                     min_idx_list.append(idx + include_factor)
 *                                     recheck_r_r.append(idx)
 */
                  goto __pyx_L21;
                }

                /* "biosaur2/cutils.pyx":470
 *                                     recheck_r_r.append(idx)
 *                                     min_val = mult_val
 *                                 elif mult_val > min_val:             # <<<<<<<<<<<<<<
 *                                     min_idx_list[-1] = idx + include_factor
 *                                     recheck_r_r[-1] = idx
 */
                __pyx_t_12 = ((__pyx_v_mult_val > __pyx_v_min_val) != 0);
                if (__pyx_t_12) {

                  /* "biosaur2/cutils.pyx":471
 *                                     min_val = mult_val
 *                                 elif mult_val > min_val:
 *                                     min_idx_list[-1] = idx + include_factor             # <<<<<<<<<<<<<<
 *                                     recheck_r_r[-1] = idx
 *                                     min_val = mult_val
 */
                  __pyx_t_9 = PyNumber_Add(__pyx_v_idx, __pyx_v_include_factor); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 471, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_9);
                  if (unlikely(__Pyx_SetItemInt(__pyx_v_min_idx_list, -1L, __pyx_t_9, long, 1, __Pyx_PyInt_From_long, 1, 1, 0) < 0)) __PYX_ERR(0, 471, __pyx_L1_error)
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

                  /* "biosaur2/cutils.pyx":472
 *                                 elif mult_val > min_val:
 *                                     min_idx_list[-1] = idx + include_factor
 *                                     recheck_r_r[-1] = idx             # <<<<<<<<<<<<<<
 *                                     min_val = mult_val
 *                     idx += 1
 */
                  if (unlikely(__Pyx_SetItemInt(__pyx_v_recheck_r_r, -1L, __pyx_v_idx, long, 1, __Pyx_PyInt_From_long, 1, 1, 0) < 0)) __PYX_ERR(0, 472, __pyx_L1_error)

                  /* "biosaur2/cutils.pyx":473
 *                                     min_idx_list[-1] = idx + include_factor
 *                                     recheck_r_r[-1] = idx
 *                                     min_val = mult_val             # <<<<<<<<<<<<<<
 *                     idx += 1
 *                 if len(min_idx_list):
 */
                  __pyx_v_min_val = __pyx_v_mult_val;

                  /* "biosaur2/cutils.pyx":470
 *                                     recheck_r_r.append(idx)
 *                                     min_val = mult_val
 *                                 elif mult_val > min_val:             # <<<<<<<<<<<<<<
 *                                     min_idx_list[-1] = idx + include_factor
 *                                     recheck_r_r[-1] = idx
 */
                }
                __pyx_L21:;

                /* "biosaur2/cutils.pyx":465
 *                             mult_val = l_r * r_r
 *                             include_factor = (1 if l_r > r_r else 0)
 *                             if (min_length_hill <= idx + include_factor <= c_len):             # <<<<<<<<<<<<<<
 *                                 if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
 *                                     min_idx_list.append(idx + include_factor)
 */
              }

              /* "biosaur2/cutils.pyx":462
 *                     if l_r >= hillValleyFactor:
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                         if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                             mult_val = l_r * r_r
 *                             include_factor = (1 if l_r > r_r else 0)
 */
            }

            /* "biosaur2/cutils.pyx":460
 * 
 *                     l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
 *                     if l_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                         r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
 *                         if r_r >= hillValleyFactor:
 */
          }

          /* "biosaur2/cutils.pyx":474
 *                                     recheck_r_r[-1] = idx
 *                                     min_val = mult_val
 *                     idx += 1             # <<<<<<<<<<<<<<
 *                 if len(min_idx_list):
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 */
          __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_v_idx, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 474, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF_SET(__pyx_v_idx, __pyx_t_9);
          __pyx_t_9 = 0;
        }

        /* "biosaur2/cutils.pyx":475
 *                                     min_val = mult_val
 *                     idx += 1
 *                 if len(min_idx_list):             # <<<<<<<<<<<<<<
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 */
        __pyx_t_17 = PyList_GET_SIZE(__pyx_v_min_idx_list); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 475, __pyx_L1_error)
        __pyx_t_12 = (__pyx_t_17 != 0);
        if (__pyx_t_12) {

          /* "biosaur2/cutils.pyx":476
 *                     idx += 1
 *                 if len(min_idx_list):
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):             # <<<<<<<<<<<<<<
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                         if r_r >= hillValleyFactor:
 */
          __pyx_t_9 = __Pyx_PyList_GetSlice(__pyx_v_min_idx_list, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_8 = __Pyx_PyInt_From_int((__pyx_v_idx_start + __pyx_v_hill_length)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
          __pyx_t_8 = 0;
          __pyx_t_8 = PyNumber_Add(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_v_min_idx_list);
          __Pyx_GIVEREF(__pyx_v_min_idx_list);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_min_idx_list);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
          __Pyx_INCREF(__pyx_v_recheck_r_r);
          __Pyx_GIVEREF(__pyx_v_recheck_r_r);
          PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_recheck_r_r);
          __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_7, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 476, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
            __pyx_t_7 = __pyx_t_8; __Pyx_INCREF(__pyx_t_7); __pyx_t_17 = 0;
            __pyx_t_18 = NULL;
          } else {
            __pyx_t_17 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 476, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_18 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 476, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          for (;;) {
            if (likely(!__pyx_t_18)) {
              if (likely(PyList_CheckExact(__pyx_t_7))) {
                if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_7)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_17); __Pyx_INCREF(__pyx_t_8); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_7, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 476, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_17); __Pyx_INCREF(__pyx_t_8); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
                #else
                __pyx_t_8 = PySequence_ITEM(__pyx_t_7, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 476, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              }
            } else {
              __pyx_t_8 = __pyx_t_18(__pyx_t_7);
              if (unlikely(!__pyx_t_8)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 476, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_8);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
              PyObject* sequence = __pyx_t_8;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 3)) {
                if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 476, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
                __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
              } else {
                __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
                __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
              }
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_6);
              #else
              __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 476, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 476, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_6);
              #endif
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_13 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 476, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_19 = Py_TYPE(__pyx_t_13)->tp_iternext;
              index = 0; __pyx_t_9 = __pyx_t_19(__pyx_t_13); if (unlikely(!__pyx_t_9)) goto __pyx_L27_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_9);
              index = 1; __pyx_t_1 = __pyx_t_19(__pyx_t_13); if (unlikely(!__pyx_t_1)) goto __pyx_L27_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_1);
              index = 2; __pyx_t_6 = __pyx_t_19(__pyx_t_13); if (unlikely(!__pyx_t_6)) goto __pyx_L27_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_6);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_13), 3) < 0) __PYX_ERR(0, 476, __pyx_L1_error)
              __pyx_t_19 = NULL;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              goto __pyx_L28_unpacking_done;
              __pyx_L27_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_t_19 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 476, __pyx_L1_error)
              __pyx_L28_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_v_min_idx, __pyx_t_9);
            __pyx_t_9 = 0;
            __Pyx_XDECREF_SET(__pyx_v_end_idx, __pyx_t_1);
            __pyx_t_1 = 0;
            __Pyx_XDECREF_SET(__pyx_v_recheck_idx, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "biosaur2/cutils.pyx":477
 *                 if len(min_idx_list):
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])             # <<<<<<<<<<<<<<
 *                         if r_r >= hillValleyFactor:
 *                             hills_dict['hills_idx_array'][idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 */
            if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 477, __pyx_L1_error)
            }
            __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_v_recheck_idx, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 477, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_12 = (__pyx_t_8 == Py_None);
            if (__pyx_t_12) {
              __pyx_t_21 = 0;
            } else {
              __pyx_t_22 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_22 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
              __pyx_t_21 = __pyx_t_22;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_INCREF(__pyx_v_end_idx);
            __pyx_t_8 = __pyx_v_end_idx;
            __pyx_t_12 = (__pyx_t_8 == Py_None);
            if (__pyx_t_12) {
              __pyx_t_22 = PY_SSIZE_T_MAX;
            } else {
              __pyx_t_23 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_23 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
              __pyx_t_22 = __pyx_t_23;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = __Pyx_PyList_GetSlice(__pyx_v_smothed_intensity, __pyx_t_21, __pyx_t_22); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 477, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 477, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(__pyx_v_smothed_intensity == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 477, __pyx_L1_error)
            }
            __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_smothed_intensity, __pyx_v_recheck_idx); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 477, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 477, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF_SET(__pyx_v_r_r, __pyx_t_8);
            __pyx_t_8 = 0;

            /* "biosaur2/cutils.pyx":478
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                         if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                             hills_dict['hills_idx_array'][idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 *                             cur_new_idx += 1
 */
            __pyx_t_8 = PyFloat_FromDouble(__pyx_v_hillValleyFactor); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 478, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_1 = PyObject_RichCompare(__pyx_v_r_r, __pyx_t_8, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 478, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_12) {

              /* "biosaur2/cutils.pyx":479
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                         if r_r >= hillValleyFactor:
 *                             hills_dict['hills_idx_array'][idx_start+min_idx:idx_start+hill_length] = cur_new_idx             # <<<<<<<<<<<<<<
 *                             cur_new_idx += 1
 * 
 */
              __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_cur_new_idx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              if (unlikely(__pyx_v_hills_dict == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 479, __pyx_L1_error)
              }
              __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 479, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_idx_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 479, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_9 = PyNumber_Add(__pyx_t_6, __pyx_v_min_idx); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 479, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (__Pyx_PyObject_SetSlice(__pyx_t_8, __pyx_t_1, 0, (__pyx_v_idx_start + __pyx_v_hill_length), &__pyx_t_9, NULL, NULL, 0, 1, 1) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "biosaur2/cutils.pyx":480
 *                         if r_r >= hillValleyFactor:
 *                             hills_dict['hills_idx_array'][idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 *                             cur_new_idx += 1             # <<<<<<<<<<<<<<
 * 
 *             idx_start = idx_end
 */
              __pyx_v_cur_new_idx = (__pyx_v_cur_new_idx + 1);

              /* "biosaur2/cutils.pyx":478
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                         if r_r >= hillValleyFactor:             # <<<<<<<<<<<<<<
 *                             hills_dict['hills_idx_array'][idx_start+min_idx:idx_start+hill_length] = cur_new_idx
 *                             cur_new_idx += 1
 */
            }

            /* "biosaur2/cutils.pyx":476
 *                     idx += 1
 *                 if len(min_idx_list):
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):             # <<<<<<<<<<<<<<
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 *                         if r_r >= hillValleyFactor:
 */
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "biosaur2/cutils.pyx":475
 *                                     min_val = mult_val
 *                     idx += 1
 *                 if len(min_idx_list):             # <<<<<<<<<<<<<<
 *                     for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
 *                         r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
 */
        }

        /* "biosaur2/cutils.pyx":439
 *             idx_end = idx_start + hill_length
 * 
 *             if hill_length >= min_length_hill * 2:             # <<<<<<<<<<<<<<
 * 
 * 
 */
      }

      /* "biosaur2/cutils.pyx":482
 *                             cur_new_idx += 1
 * 
 *             idx_start = idx_end             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])
 */
      __pyx_v_idx_start = __pyx_v_idx_end;

      /* "biosaur2/cutils.pyx":434
 *         cur_new_idx = max(hills_dict['hills_idx_array_unique']) + 1
 * 
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):             # <<<<<<<<<<<<<<
 * 
 *             hill_length = counter_hills_idx[hill_idx]
 */
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "biosaur2/cutils.pyx":484
 *             idx_start = idx_end
 * 
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])             # <<<<<<<<<<<<<<
 *         del hills_dict['hills_idx_array_unique']
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 484, __pyx_L1_error)
    }
    __pyx_t_14 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_7 = PySequence_List(__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 484, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_7) < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "biosaur2/cutils.pyx":485
 * 
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])
 *         del hills_dict['hills_idx_array_unique']             # <<<<<<<<<<<<<<
 * 
 *     return hills_dict
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 485, __pyx_L1_error)
    }
    if (unlikely(PyDict_DelItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique) < 0)) __PYX_ERR(0, 485, __pyx_L1_error)

    /* "biosaur2/cutils.pyx":420
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 * 
 *     if len(hills_dict['orig_idx_array']):             # <<<<<<<<<<<<<<
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 */
  }

  /* "biosaur2/cutils.pyx":487
 *         del hills_dict['hills_idx_array_unique']
 * 
 *     return hills_dict             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_hills_dict);
  __pyx_r = __pyx_v_hills_dict;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":396
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks_old(dict hills_dict, list data_for_analyse_tmp, dict args):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("biosaur2.cutils.split_peaks_old", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp_hill_length);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_minl);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_sort);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp_scans);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp_orig_idx);
  __Pyx_XDECREF(__pyx_v_tmp_intensity);
  __Pyx_XDECREF(__pyx_v_smothed_intensity);
  __Pyx_XDECREF(__pyx_v_min_idx_list);
  __Pyx_XDECREF(__pyx_v_recheck_r_r);
  __Pyx_XDECREF(__pyx_v_counter_hills_idx);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_l_r);
  __Pyx_XDECREF(__pyx_v_r_r);
  __Pyx_XDECREF(__pyx_v_include_factor);
  __Pyx_XDECREF(__pyx_v_min_idx);
  __Pyx_XDECREF(__pyx_v_end_idx);
  __Pyx_XDECREF(__pyx_v_recheck_idx);
  __Pyx_XDECREF(__pyx_v_orig_idx_val);
  __Pyx_XDECREF(__pyx_v_scan_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":494
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def detect_hills(list data_for_analyse_tmp, dict args, float mz_step, bint paseftol, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict hills_dict, prev_fast_dict, z
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_21detect_hills(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_20detect_hills[] = "detect_hills(list data_for_analyse_tmp, dict args, float mz_step, bool paseftol, bool dia=False)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_21detect_hills = {"detect_hills", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_21detect_hills, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_20detect_hills};
static PyObject *__pyx_pw_8biosaur2_6cutils_21detect_hills(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data_for_analyse_tmp = 0;
  PyObject *__pyx_v_args = 0;
  float __pyx_v_mz_step;
  int __pyx_v_paseftol;
  int __pyx_v_dia;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("detect_hills (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data_for_analyse_tmp,&__pyx_n_s_args,&__pyx_n_s_mz_step,&__pyx_n_s_paseftol,&__pyx_n_s_dia,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_for_analyse_tmp)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("detect_hills", 0, 4, 5, 1); __PYX_ERR(0, 494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("detect_hills", 0, 4, 5, 2); __PYX_ERR(0, 494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_paseftol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("detect_hills", 0, 4, 5, 3); __PYX_ERR(0, 494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dia);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "detect_hills") < 0)) __PYX_ERR(0, 494, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_data_for_analyse_tmp = ((PyObject*)values[0]);
    __pyx_v_args = ((PyObject*)values[1]);
    __pyx_v_mz_step = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_mz_step == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L3_error)
    __pyx_v_paseftol = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_paseftol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L3_error)
    if (values[4]) {
      __pyx_v_dia = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_dia == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 494, __pyx_L3_error)
    } else {
      __pyx_v_dia = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("detect_hills", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 494, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.detect_hills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data_for_analyse_tmp), (&PyList_Type), 1, "data_for_analyse_tmp", 1))) __PYX_ERR(0, 494, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyDict_Type), 1, "args", 1))) __PYX_ERR(0, 494, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_20detect_hills(__pyx_self, __pyx_v_data_for_analyse_tmp, __pyx_v_args, __pyx_v_mz_step, __pyx_v_paseftol, __pyx_v_dia);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_20detect_hills(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data_for_analyse_tmp, PyObject *__pyx_v_args, float __pyx_v_mz_step, int __pyx_v_paseftol, int __pyx_v_dia) {
  PyObject *__pyx_v_hills_dict = 0;
  PyObject *__pyx_v_prev_fast_dict = 0;
  PyObject *__pyx_v_z = 0;
  CYTHON_UNUSED PyObject *__pyx_v_total_num_hills = 0;
  CYTHON_UNUSED PyObject *__pyx_v_total_mass_diff = 0;
  PyObject *__pyx_v_basic_id_sorted = 0;
  PyObject *__pyx_v_all_idx = 0;
  PyObject *__pyx_v_all_prevs = 0;
  PyObject *__pyx_v_banned_prev_idx_set = 0;
  float __pyx_v_hill_mass_accuracy;
  float __pyx_v_best_intensity;
  float __pyx_v_mz_cur;
  float __pyx_v_cur_intensity;
  float __pyx_v_cur_mass_diff_with_sign;
  float __pyx_v_cur_mass_diff;
  CYTHON_UNUSED float __pyx_v_best_mass_diff;
  int __pyx_v_last_idx;
  int __pyx_v_prev_idx;
  int __pyx_v_spec_idx;
  int __pyx_v_len_mz;
  int __pyx_v_idx;
  int __pyx_v_fm;
  int __pyx_v_fi;
  int __pyx_v_best_idx_prev;
  int __pyx_v_idx_prev;
  PyArrayObject *__pyx_v_idx_for_sort = 0;
  PyArrayObject *__pyx_v_mz_sorted = 0;
  PyArrayObject *__pyx_v_im_sorted = 0;
  int __pyx_v_flag1;
  int __pyx_v_flag2;
  int __pyx_v_flag3;
  PyObject *__pyx_v_prev_fast_dict_im = NULL;
  PyObject *__pyx_v_fast_dict_im = NULL;
  PyObject *__pyx_v_fast_array_im = NULL;
  PyObject *__pyx_v_fast_dict = NULL;
  PyObject *__pyx_v_fast_array = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  float __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_t_20;
  long __pyx_t_21;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("detect_hills", 0);

  /* "biosaur2/cutils.pyx":504
 *     cdef bint flag1, flag2, flag3
 * 
 *     hills_dict = {}             # <<<<<<<<<<<<<<
 * 
 *     if dia is False:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_hills_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":506
 *     hills_dict = {}
 * 
 *     if dia is False:             # <<<<<<<<<<<<<<
 *         hill_mass_accuracy = args['htol']
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_dia == 0) != 0);
  if (__pyx_t_2) {

    /* "biosaur2/cutils.pyx":507
 * 
 *     if dia is False:
 *         hill_mass_accuracy = args['htol']             # <<<<<<<<<<<<<<
 *     else:
 *         hill_mass_accuracy = args['diahtol']
 */
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 507, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_htol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_3 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_hill_mass_accuracy = __pyx_t_3;

    /* "biosaur2/cutils.pyx":506
 *     hills_dict = {}
 * 
 *     if dia is False:             # <<<<<<<<<<<<<<
 *         hill_mass_accuracy = args['htol']
 *     else:
 */
    goto __pyx_L3;
  }

  /* "biosaur2/cutils.pyx":509
 *         hill_mass_accuracy = args['htol']
 *     else:
 *         hill_mass_accuracy = args['diahtol']             # <<<<<<<<<<<<<<
 * 
 *     hills_dict['hills_idx_array'] = []
 */
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 509, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_diahtol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_3 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_hill_mass_accuracy = __pyx_t_3;
  }
  __pyx_L3:;

  /* "biosaur2/cutils.pyx":511
 *         hill_mass_accuracy = args['diahtol']
 * 
 *     hills_dict['hills_idx_array'] = []             # <<<<<<<<<<<<<<
 *     hills_dict['orig_idx_array'] = []
 *     hills_dict['scan_idx_array'] = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":512
 * 
 *     hills_dict['hills_idx_array'] = []
 *     hills_dict['orig_idx_array'] = []             # <<<<<<<<<<<<<<
 *     hills_dict['scan_idx_array'] = []
 *     hills_dict['mzs_array'] = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":513
 *     hills_dict['hills_idx_array'] = []
 *     hills_dict['orig_idx_array'] = []
 *     hills_dict['scan_idx_array'] = []             # <<<<<<<<<<<<<<
 *     hills_dict['mzs_array'] = []
 *     hills_dict['intensity_array'] = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":514
 *     hills_dict['orig_idx_array'] = []
 *     hills_dict['scan_idx_array'] = []
 *     hills_dict['mzs_array'] = []             # <<<<<<<<<<<<<<
 *     hills_dict['intensity_array'] = []
 *     if paseftol is not False:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_mzs_array, __pyx_t_1) < 0)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":515
 *     hills_dict['scan_idx_array'] = []
 *     hills_dict['mzs_array'] = []
 *     hills_dict['intensity_array'] = []             # <<<<<<<<<<<<<<
 *     if paseftol is not False:
 *         hills_dict['im_array'] = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_intensity_array_2, __pyx_t_1) < 0)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":516
 *     hills_dict['mzs_array'] = []
 *     hills_dict['intensity_array'] = []
 *     if paseftol is not False:             # <<<<<<<<<<<<<<
 *         hills_dict['im_array'] = []
 *         prev_fast_dict_im = dict()
 */
  __pyx_t_2 = ((__pyx_v_paseftol != 0) != 0);
  if (__pyx_t_2) {

    /* "biosaur2/cutils.pyx":517
 *     hills_dict['intensity_array'] = []
 *     if paseftol is not False:
 *         hills_dict['im_array'] = []             # <<<<<<<<<<<<<<
 *         prev_fast_dict_im = dict()
 * 
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_im_array, __pyx_t_1) < 0)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":518
 *     if paseftol is not False:
 *         hills_dict['im_array'] = []
 *         prev_fast_dict_im = dict()             # <<<<<<<<<<<<<<
 * 
 *     last_idx = -1
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_prev_fast_dict_im = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":516
 *     hills_dict['mzs_array'] = []
 *     hills_dict['intensity_array'] = []
 *     if paseftol is not False:             # <<<<<<<<<<<<<<
 *         hills_dict['im_array'] = []
 *         prev_fast_dict_im = dict()
 */
  }

  /* "biosaur2/cutils.pyx":520
 *         prev_fast_dict_im = dict()
 * 
 *     last_idx = -1             # <<<<<<<<<<<<<<
 *     prev_idx = -1
 *     prev_fast_dict = dict()
 */
  __pyx_v_last_idx = -1;

  /* "biosaur2/cutils.pyx":521
 * 
 *     last_idx = -1
 *     prev_idx = -1             # <<<<<<<<<<<<<<
 *     prev_fast_dict = dict()
 * 
 */
  __pyx_v_prev_idx = -1;

  /* "biosaur2/cutils.pyx":522
 *     last_idx = -1
 *     prev_idx = -1
 *     prev_fast_dict = dict()             # <<<<<<<<<<<<<<
 * 
 *     total_num_hills = []
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_prev_fast_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":524
 *     prev_fast_dict = dict()
 * 
 *     total_num_hills = []             # <<<<<<<<<<<<<<
 *     total_mass_diff = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_total_num_hills = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":525
 * 
 *     total_num_hills = []
 *     total_mass_diff = []             # <<<<<<<<<<<<<<
 * 
 *     for spec_idx, z in enumerate(data_for_analyse_tmp):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_total_mass_diff = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":527
 *     total_mass_diff = []
 * 
 *     for spec_idx, z in enumerate(data_for_analyse_tmp):             # <<<<<<<<<<<<<<
 * 
 *         len_mz = len(z['m/z array'])
 */
  __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_v_data_for_analyse_tmp; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)
    #else
    __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_z, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;
    __pyx_v_spec_idx = __pyx_t_4;
    __pyx_t_4 = (__pyx_t_4 + 1);

    /* "biosaur2/cutils.pyx":529
 *     for spec_idx, z in enumerate(data_for_analyse_tmp):
 * 
 *         len_mz = len(z['m/z array'])             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['hills_idx_array'].extend(list(range(last_idx+1, last_idx+1+len_mz, 1)))
 */
    if (unlikely(__pyx_v_z == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 529, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_len_mz = __pyx_t_7;

    /* "biosaur2/cutils.pyx":531
 *         len_mz = len(z['m/z array'])
 * 
 *         hills_dict['hills_idx_array'].extend(list(range(last_idx+1, last_idx+1+len_mz, 1)))             # <<<<<<<<<<<<<<
 *         hills_dict['orig_idx_array'].extend(range(len_mz))
 *         hills_dict['scan_idx_array'].extend([spec_idx] * len_mz)
 */
    __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_extend); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_From_long((__pyx_v_last_idx + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __Pyx_PyInt_From_long(((__pyx_v_last_idx + 1) + __pyx_v_len_mz)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_10);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_int_1);
    __pyx_t_8 = 0;
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_11, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PySequence_List(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":532
 * 
 *         hills_dict['hills_idx_array'].extend(list(range(last_idx+1, last_idx+1+len_mz, 1)))
 *         hills_dict['orig_idx_array'].extend(range(len_mz))             # <<<<<<<<<<<<<<
 *         hills_dict['scan_idx_array'].extend([spec_idx] * len_mz)
 * 
 */
    __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_extend); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_len_mz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":533
 *         hills_dict['hills_idx_array'].extend(list(range(last_idx+1, last_idx+1+len_mz, 1)))
 *         hills_dict['orig_idx_array'].extend(range(len_mz))
 *         hills_dict['scan_idx_array'].extend([spec_idx] * len_mz)             # <<<<<<<<<<<<<<
 * 
 *         idx_for_sort = np.argsort(z['intensity array'])[::-1]
 */
    __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_extend); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_spec_idx); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_9 = PyList_New(1 * ((__pyx_v_len_mz<0) ? 0:__pyx_v_len_mz)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_len_mz; __pyx_temp++) {
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_11);
        PyList_SET_ITEM(__pyx_t_9, __pyx_temp, __pyx_t_11);
      }
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":535
 *         hills_dict['scan_idx_array'].extend([spec_idx] * len_mz)
 * 
 *         idx_for_sort = np.argsort(z['intensity array'])[::-1]             # <<<<<<<<<<<<<<
 * 
 *         mz_sorted = z['m/z array'][idx_for_sort]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_argsort); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v_z == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 535, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_slice_); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_idx_for_sort, ((PyArrayObject *)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "biosaur2/cutils.pyx":537
 *         idx_for_sort = np.argsort(z['intensity array'])[::-1]
 * 
 *         mz_sorted = z['m/z array'][idx_for_sort]             # <<<<<<<<<<<<<<
 *         basic_id_sorted = list(np.array(range(len_mz))[idx_for_sort])
 * 
 */
    if (unlikely(__pyx_v_z == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 537, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_9, ((PyObject *)__pyx_v_idx_for_sort)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_mz_sorted, ((PyArrayObject *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":538
 * 
 *         mz_sorted = z['m/z array'][idx_for_sort]
 *         basic_id_sorted = list(np.array(range(len_mz))[idx_for_sort])             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['mzs_array'].extend(z['m/z array'])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_len_mz); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_9, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_6, ((PyObject *)__pyx_v_idx_for_sort)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PySequence_List(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_basic_id_sorted, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":540
 *         basic_id_sorted = list(np.array(range(len_mz))[idx_for_sort])
 * 
 *         hills_dict['mzs_array'].extend(z['m/z array'])             # <<<<<<<<<<<<<<
 *         hills_dict['intensity_array'].extend(z['intensity array'])
 * 
 */
    __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_mzs_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_extend); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v_z == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 540, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":541
 * 
 *         hills_dict['mzs_array'].extend(z['m/z array'])
 *         hills_dict['intensity_array'].extend(z['intensity array'])             # <<<<<<<<<<<<<<
 * 
 *         if paseftol is not False:
 */
    __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_intensity_array_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_extend); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(__pyx_v_z == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 541, __pyx_L1_error)
    }
    __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_9, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "biosaur2/cutils.pyx":543
 *         hills_dict['intensity_array'].extend(z['intensity array'])
 * 
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             im_sorted = z['mean inverse reduced ion mobility array'][idx_for_sort]
 *             hills_dict['im_array'].extend(z['mean inverse reduced ion mobility array'])
 */
    __pyx_t_2 = ((__pyx_v_paseftol != 0) != 0);
    if (__pyx_t_2) {

      /* "biosaur2/cutils.pyx":544
 * 
 *         if paseftol is not False:
 *             im_sorted = z['mean inverse reduced ion mobility array'][idx_for_sort]             # <<<<<<<<<<<<<<
 *             hills_dict['im_array'].extend(z['mean inverse reduced ion mobility array'])
 * 
 */
      if (unlikely(__pyx_v_z == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 544, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_mean_inverse_reduced_ion_mobilit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_6, ((PyObject *)__pyx_v_idx_for_sort)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_im_sorted, ((PyArrayObject *)__pyx_t_10));
      __pyx_t_10 = 0;

      /* "biosaur2/cutils.pyx":545
 *         if paseftol is not False:
 *             im_sorted = z['mean inverse reduced ion mobility array'][idx_for_sort]
 *             hills_dict['im_array'].extend(z['mean inverse reduced ion mobility array'])             # <<<<<<<<<<<<<<
 * 
 *             fast_dict_im, fast_array_im = get_fast_dict(im_sorted, paseftol, basic_id_sorted)
 */
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_im_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_extend); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__pyx_v_z == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 545, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_mean_inverse_reduced_ion_mobilit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
        }
      }
      __pyx_t_10 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_9, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "biosaur2/cutils.pyx":547
 *             hills_dict['im_array'].extend(z['mean inverse reduced ion mobility array'])
 * 
 *             fast_dict_im, fast_array_im = get_fast_dict(im_sorted, paseftol, basic_id_sorted)             # <<<<<<<<<<<<<<
 * 
 *         fast_dict, fast_array = get_fast_dict(mz_sorted, mz_step, basic_id_sorted)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_get_fast_dict); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_paseftol); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_11)) {
        PyObject *__pyx_temp[4] = {__pyx_t_9, ((PyObject *)__pyx_v_im_sorted), __pyx_t_6, __pyx_v_basic_id_sorted};
        __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
        PyObject *__pyx_temp[4] = {__pyx_t_9, ((PyObject *)__pyx_v_im_sorted), __pyx_t_6, __pyx_v_basic_id_sorted};
        __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_im_sorted));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_im_sorted));
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_12, ((PyObject *)__pyx_v_im_sorted));
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_12, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_basic_id_sorted);
        __Pyx_GIVEREF(__pyx_v_basic_id_sorted);
        PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_12, __pyx_v_basic_id_sorted);
        __pyx_t_6 = 0;
        __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_8, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
        PyObject* sequence = __pyx_t_10;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 547, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_11 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_11 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_13 = Py_TYPE(__pyx_t_6)->tp_iternext;
        index = 0; __pyx_t_11 = __pyx_t_13(__pyx_t_6); if (unlikely(!__pyx_t_11)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_11);
        index = 1; __pyx_t_8 = __pyx_t_13(__pyx_t_6); if (unlikely(!__pyx_t_8)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_6), 2) < 0) __PYX_ERR(0, 547, __pyx_L1_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 547, __pyx_L1_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_fast_dict_im, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_fast_array_im, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "biosaur2/cutils.pyx":543
 *         hills_dict['intensity_array'].extend(z['intensity array'])
 * 
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             im_sorted = z['mean inverse reduced ion mobility array'][idx_for_sort]
 *             hills_dict['im_array'].extend(z['mean inverse reduced ion mobility array'])
 */
    }

    /* "biosaur2/cutils.pyx":549
 *             fast_dict_im, fast_array_im = get_fast_dict(im_sorted, paseftol, basic_id_sorted)
 * 
 *         fast_dict, fast_array = get_fast_dict(mz_sorted, mz_step, basic_id_sorted)             # <<<<<<<<<<<<<<
 * 
 *         banned_prev_idx_set = set()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_get_fast_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_11 = PyFloat_FromDouble(__pyx_v_mz_step); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_6 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[4] = {__pyx_t_6, ((PyObject *)__pyx_v_mz_sorted), __pyx_t_11, __pyx_v_basic_id_sorted};
      __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[4] = {__pyx_t_6, ((PyObject *)__pyx_v_mz_sorted), __pyx_t_11, __pyx_v_basic_id_sorted};
      __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_mz_sorted));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_mz_sorted));
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_12, ((PyObject *)__pyx_v_mz_sorted));
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_12, __pyx_t_11);
      __Pyx_INCREF(__pyx_v_basic_id_sorted);
      __Pyx_GIVEREF(__pyx_v_basic_id_sorted);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_12, __pyx_v_basic_id_sorted);
      __pyx_t_11 = 0;
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
      PyObject* sequence = __pyx_t_10;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 549, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_11 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_13 = Py_TYPE(__pyx_t_11)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_13(__pyx_t_11); if (unlikely(!__pyx_t_8)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_13(__pyx_t_11); if (unlikely(!__pyx_t_9)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_11), 2) < 0) __PYX_ERR(0, 549, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 549, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_fast_dict, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fast_array, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "biosaur2/cutils.pyx":551
 *         fast_dict, fast_array = get_fast_dict(mz_sorted, mz_step, basic_id_sorted)
 * 
 *         banned_prev_idx_set = set()             # <<<<<<<<<<<<<<
 * 
 *         for idx, fm, fi in zip(basic_id_sorted, fast_array, (fast_array if paseftol is False else fast_array_im)):
 */
    __pyx_t_10 = PySet_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_XDECREF_SET(__pyx_v_banned_prev_idx_set, ((PyObject*)__pyx_t_10));
    __pyx_t_10 = 0;

    /* "biosaur2/cutils.pyx":553
 *         banned_prev_idx_set = set()
 * 
 *         for idx, fm, fi in zip(basic_id_sorted, fast_array, (fast_array if paseftol is False else fast_array_im)):             # <<<<<<<<<<<<<<
 * 
 *             flag1 = fm in prev_fast_dict
 */
    if (((__pyx_v_paseftol == 0) != 0)) {
      __Pyx_INCREF(__pyx_v_fast_array);
      __pyx_t_10 = __pyx_v_fast_array;
    } else {
      if (unlikely(!__pyx_v_fast_array_im)) { __Pyx_RaiseUnboundLocalError("fast_array_im"); __PYX_ERR(0, 553, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_v_fast_array_im);
      __pyx_t_10 = __pyx_v_fast_array_im;
    }
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_basic_id_sorted);
    __Pyx_GIVEREF(__pyx_v_basic_id_sorted);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_basic_id_sorted);
    __Pyx_INCREF(__pyx_v_fast_array);
    __Pyx_GIVEREF(__pyx_v_fast_array);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_fast_array);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
      __pyx_t_9 = __pyx_t_10; __Pyx_INCREF(__pyx_t_9); __pyx_t_7 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_14 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 553, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_7); __Pyx_INCREF(__pyx_t_10); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_9, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 553, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_7); __Pyx_INCREF(__pyx_t_10); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_9, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 553, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        }
      } else {
        __pyx_t_10 = __pyx_t_14(__pyx_t_9);
        if (unlikely(!__pyx_t_10)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 553, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_10);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
        PyObject* sequence = __pyx_t_10;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 3)) {
          if (size > 3) __Pyx_RaiseTooManyValuesError(3);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 553, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
        } else {
          __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_11 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
        }
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_15 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_13 = Py_TYPE(__pyx_t_15)->tp_iternext;
        index = 0; __pyx_t_8 = __pyx_t_13(__pyx_t_15); if (unlikely(!__pyx_t_8)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        index = 1; __pyx_t_11 = __pyx_t_13(__pyx_t_15); if (unlikely(!__pyx_t_11)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_11);
        index = 2; __pyx_t_6 = __pyx_t_13(__pyx_t_15); if (unlikely(!__pyx_t_6)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_15), 3) < 0) __PYX_ERR(0, 553, __pyx_L1_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        goto __pyx_L15_unpacking_done;
        __pyx_L14_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 553, __pyx_L1_error)
        __pyx_L15_unpacking_done:;
      }
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 553, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_idx = __pyx_t_12;
      __pyx_v_fm = __pyx_t_16;
      __pyx_v_fi = __pyx_t_17;

      /* "biosaur2/cutils.pyx":555
 *         for idx, fm, fi in zip(basic_id_sorted, fast_array, (fast_array if paseftol is False else fast_array_im)):
 * 
 *             flag1 = fm in prev_fast_dict             # <<<<<<<<<<<<<<
 *             flag2 = fm-1 in prev_fast_dict
 *             flag3 = fm+1 in prev_fast_dict
 */
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_fm); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 555, __pyx_L1_error)
      }
      __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_t_10, __pyx_v_prev_fast_dict, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 555, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_flag1 = __pyx_t_2;

      /* "biosaur2/cutils.pyx":556
 * 
 *             flag1 = fm in prev_fast_dict
 *             flag2 = fm-1 in prev_fast_dict             # <<<<<<<<<<<<<<
 *             flag3 = fm+1 in prev_fast_dict
 * 
 */
      __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_fm - 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 556, __pyx_L1_error)
      }
      __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_t_10, __pyx_v_prev_fast_dict, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_flag2 = __pyx_t_2;

      /* "biosaur2/cutils.pyx":557
 *             flag1 = fm in prev_fast_dict
 *             flag2 = fm-1 in prev_fast_dict
 *             flag3 = fm+1 in prev_fast_dict             # <<<<<<<<<<<<<<
 * 
 *             if flag1 or flag2 or flag3:
 */
      __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_fm + 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 557, __pyx_L1_error)
      }
      __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_t_10, __pyx_v_prev_fast_dict, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 557, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_flag3 = __pyx_t_2;

      /* "biosaur2/cutils.pyx":559
 *             flag3 = fm+1 in prev_fast_dict
 * 
 *             if flag1 or flag2 or flag3:             # <<<<<<<<<<<<<<
 * 
 *                 if flag1:
 */
      __pyx_t_18 = (__pyx_v_flag1 != 0);
      if (!__pyx_t_18) {
      } else {
        __pyx_t_2 = __pyx_t_18;
        goto __pyx_L17_bool_binop_done;
      }
      __pyx_t_18 = (__pyx_v_flag2 != 0);
      if (!__pyx_t_18) {
      } else {
        __pyx_t_2 = __pyx_t_18;
        goto __pyx_L17_bool_binop_done;
      }
      __pyx_t_18 = (__pyx_v_flag3 != 0);
      __pyx_t_2 = __pyx_t_18;
      __pyx_L17_bool_binop_done:;
      if (__pyx_t_2) {

        /* "biosaur2/cutils.pyx":561
 *             if flag1 or flag2 or flag3:
 * 
 *                 if flag1:             # <<<<<<<<<<<<<<
 *                     all_idx = prev_fast_dict[fm]
 *                     if flag2:
 */
        __pyx_t_2 = (__pyx_v_flag1 != 0);
        if (__pyx_t_2) {

          /* "biosaur2/cutils.pyx":562
 * 
 *                 if flag1:
 *                     all_idx = prev_fast_dict[fm]             # <<<<<<<<<<<<<<
 *                     if flag2:
 *                         all_idx += prev_fast_dict[fm-1]
 */
          if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 562, __pyx_L1_error)
          }
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_fm); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 562, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_prev_fast_dict, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 562, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 562, __pyx_L1_error)
          __Pyx_XDECREF_SET(__pyx_v_all_idx, ((PyObject*)__pyx_t_6));
          __pyx_t_6 = 0;

          /* "biosaur2/cutils.pyx":563
 *                 if flag1:
 *                     all_idx = prev_fast_dict[fm]
 *                     if flag2:             # <<<<<<<<<<<<<<
 *                         all_idx += prev_fast_dict[fm-1]
 *                     if flag3:
 */
          __pyx_t_2 = (__pyx_v_flag2 != 0);
          if (__pyx_t_2) {

            /* "biosaur2/cutils.pyx":564
 *                     all_idx = prev_fast_dict[fm]
 *                     if flag2:
 *                         all_idx += prev_fast_dict[fm-1]             # <<<<<<<<<<<<<<
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]
 */
            if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 564, __pyx_L1_error)
            }
            __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_fm - 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_prev_fast_dict, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 564, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_all_idx, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 564, __pyx_L1_error)
            __Pyx_DECREF_SET(__pyx_v_all_idx, ((PyObject*)__pyx_t_6));
            __pyx_t_6 = 0;

            /* "biosaur2/cutils.pyx":563
 *                 if flag1:
 *                     all_idx = prev_fast_dict[fm]
 *                     if flag2:             # <<<<<<<<<<<<<<
 *                         all_idx += prev_fast_dict[fm-1]
 *                     if flag3:
 */
          }

          /* "biosaur2/cutils.pyx":565
 *                     if flag2:
 *                         all_idx += prev_fast_dict[fm-1]
 *                     if flag3:             # <<<<<<<<<<<<<<
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag2:
 */
          __pyx_t_2 = (__pyx_v_flag3 != 0);
          if (__pyx_t_2) {

            /* "biosaur2/cutils.pyx":566
 *                         all_idx += prev_fast_dict[fm-1]
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]             # <<<<<<<<<<<<<<
 *                 elif flag2:
 *                     all_idx = prev_fast_dict[fm-1]
 */
            if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 566, __pyx_L1_error)
            }
            __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_fm + 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_prev_fast_dict, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 566, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_all_idx, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 566, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 566, __pyx_L1_error)
            __Pyx_DECREF_SET(__pyx_v_all_idx, ((PyObject*)__pyx_t_6));
            __pyx_t_6 = 0;

            /* "biosaur2/cutils.pyx":565
 *                     if flag2:
 *                         all_idx += prev_fast_dict[fm-1]
 *                     if flag3:             # <<<<<<<<<<<<<<
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag2:
 */
          }

          /* "biosaur2/cutils.pyx":561
 *             if flag1 or flag2 or flag3:
 * 
 *                 if flag1:             # <<<<<<<<<<<<<<
 *                     all_idx = prev_fast_dict[fm]
 *                     if flag2:
 */
          goto __pyx_L20;
        }

        /* "biosaur2/cutils.pyx":567
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag2:             # <<<<<<<<<<<<<<
 *                     all_idx = prev_fast_dict[fm-1]
 *                     if flag3:
 */
        __pyx_t_2 = (__pyx_v_flag2 != 0);
        if (__pyx_t_2) {

          /* "biosaur2/cutils.pyx":568
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag2:
 *                     all_idx = prev_fast_dict[fm-1]             # <<<<<<<<<<<<<<
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]
 */
          if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 568, __pyx_L1_error)
          }
          __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_fm - 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 568, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_prev_fast_dict, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 568, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (!(likely(PyList_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 568, __pyx_L1_error)
          __Pyx_XDECREF_SET(__pyx_v_all_idx, ((PyObject*)__pyx_t_10));
          __pyx_t_10 = 0;

          /* "biosaur2/cutils.pyx":569
 *                 elif flag2:
 *                     all_idx = prev_fast_dict[fm-1]
 *                     if flag3:             # <<<<<<<<<<<<<<
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag3:
 */
          __pyx_t_2 = (__pyx_v_flag3 != 0);
          if (__pyx_t_2) {

            /* "biosaur2/cutils.pyx":570
 *                     all_idx = prev_fast_dict[fm-1]
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]             # <<<<<<<<<<<<<<
 *                 elif flag3:
 *                     all_idx = prev_fast_dict[fm+1]
 */
            if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 570, __pyx_L1_error)
            }
            __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_fm + 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 570, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_prev_fast_dict, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 570, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_all_idx, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 570, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (!(likely(PyList_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 570, __pyx_L1_error)
            __Pyx_DECREF_SET(__pyx_v_all_idx, ((PyObject*)__pyx_t_10));
            __pyx_t_10 = 0;

            /* "biosaur2/cutils.pyx":569
 *                 elif flag2:
 *                     all_idx = prev_fast_dict[fm-1]
 *                     if flag3:             # <<<<<<<<<<<<<<
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag3:
 */
          }

          /* "biosaur2/cutils.pyx":567
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag2:             # <<<<<<<<<<<<<<
 *                     all_idx = prev_fast_dict[fm-1]
 *                     if flag3:
 */
          goto __pyx_L20;
        }

        /* "biosaur2/cutils.pyx":571
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag3:             # <<<<<<<<<<<<<<
 *                     all_idx = prev_fast_dict[fm+1]
 * 
 */
        __pyx_t_2 = (__pyx_v_flag3 != 0);
        if (__pyx_t_2) {

          /* "biosaur2/cutils.pyx":572
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag3:
 *                     all_idx = prev_fast_dict[fm+1]             # <<<<<<<<<<<<<<
 * 
 *                 best_intensity = 0
 */
          if (unlikely(__pyx_v_prev_fast_dict == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 572, __pyx_L1_error)
          }
          __pyx_t_10 = __Pyx_PyInt_From_long((__pyx_v_fm + 1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 572, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_prev_fast_dict, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 572, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 572, __pyx_L1_error)
          __Pyx_XDECREF_SET(__pyx_v_all_idx, ((PyObject*)__pyx_t_6));
          __pyx_t_6 = 0;

          /* "biosaur2/cutils.pyx":571
 *                     if flag3:
 *                         all_idx += prev_fast_dict[fm+1]
 *                 elif flag3:             # <<<<<<<<<<<<<<
 *                     all_idx = prev_fast_dict[fm+1]
 * 
 */
        }
        __pyx_L20:;

        /* "biosaur2/cutils.pyx":574
 *                     all_idx = prev_fast_dict[fm+1]
 * 
 *                 best_intensity = 0             # <<<<<<<<<<<<<<
 *                 best_idx_prev = 0
 *                 mz_cur = z['m/z array'][idx]
 */
        __pyx_v_best_intensity = 0.0;

        /* "biosaur2/cutils.pyx":575
 * 
 *                 best_intensity = 0
 *                 best_idx_prev = 0             # <<<<<<<<<<<<<<
 *                 mz_cur = z['m/z array'][idx]
 * 
 */
        __pyx_v_best_idx_prev = 0;

        /* "biosaur2/cutils.pyx":576
 *                 best_intensity = 0
 *                 best_idx_prev = 0
 *                 mz_cur = z['m/z array'][idx]             # <<<<<<<<<<<<<<
 * 
 * 
 */
        if (unlikely(__pyx_v_z == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 576, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_z, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_6, __pyx_v_idx, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_3 = __pyx_PyFloat_AsFloat(__pyx_t_10); if (unlikely((__pyx_t_3 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_v_mz_cur = __pyx_t_3;

        /* "biosaur2/cutils.pyx":579
 * 
 * 
 *                 all_prevs = [[idx_prev, data_for_analyse_tmp[spec_idx-1]['intensity array'][idx_prev]] for idx_prev in all_idx if idx_prev not in banned_prev_idx_set and (paseftol is False or idx_prev in prev_fast_dict_im[fi])]             # <<<<<<<<<<<<<<
 *                 for idx_prev, cur_intensity in all_prevs:
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
 */
        __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 579, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (unlikely(!__pyx_v_all_idx)) { __Pyx_RaiseUnboundLocalError("all_idx"); __PYX_ERR(0, 579, __pyx_L1_error) }
        if (unlikely(__pyx_v_all_idx == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 579, __pyx_L1_error)
        }
        __pyx_t_6 = __pyx_v_all_idx; __Pyx_INCREF(__pyx_t_6); __pyx_t_19 = 0;
        for (;;) {
          if (__pyx_t_19 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_19); __Pyx_INCREF(__pyx_t_11); __pyx_t_19++; if (unlikely(0 < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
          #else
          __pyx_t_11 = PySequence_ITEM(__pyx_t_6, __pyx_t_19); __pyx_t_19++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
          __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_idx_prev = __pyx_t_17;
          __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_idx_prev); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_18 = (__Pyx_PySet_ContainsTF(__pyx_t_11, __pyx_v_banned_prev_idx_set, Py_NE)); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_20 = (__pyx_t_18 != 0);
          if (__pyx_t_20) {
          } else {
            __pyx_t_2 = __pyx_t_20;
            goto __pyx_L27_bool_binop_done;
          }
          __pyx_t_20 = ((__pyx_v_paseftol == 0) != 0);
          if (!__pyx_t_20) {
          } else {
            __pyx_t_2 = __pyx_t_20;
            goto __pyx_L27_bool_binop_done;
          }
          __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_idx_prev); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (unlikely(!__pyx_v_prev_fast_dict_im)) { __Pyx_RaiseUnboundLocalError("prev_fast_dict_im"); __PYX_ERR(0, 579, __pyx_L1_error) }
          __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_prev_fast_dict_im, __pyx_v_fi, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_20 = (__Pyx_PySequence_ContainsTF(__pyx_t_11, __pyx_t_8, Py_EQ)); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_18 = (__pyx_t_20 != 0);
          __pyx_t_2 = __pyx_t_18;
          __pyx_L27_bool_binop_done:;
          if (__pyx_t_2) {
            __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_idx_prev); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 579, __pyx_L1_error)
            }
            __pyx_t_21 = (__pyx_v_spec_idx - 1);
            __pyx_t_11 = __Pyx_GetItemInt_List(__pyx_v_data_for_analyse_tmp, __pyx_t_21, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_15 = __Pyx_PyObject_Dict_GetItem(__pyx_t_11, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_15, __pyx_v_idx_prev, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_15 = PyList_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_15, 0, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_15, 1, __pyx_t_11);
            __pyx_t_8 = 0;
            __pyx_t_11 = 0;
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_10, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_all_prevs, ((PyObject*)__pyx_t_10));
        __pyx_t_10 = 0;

        /* "biosaur2/cutils.pyx":580
 * 
 *                 all_prevs = [[idx_prev, data_for_analyse_tmp[spec_idx-1]['intensity array'][idx_prev]] for idx_prev in all_idx if idx_prev not in banned_prev_idx_set and (paseftol is False or idx_prev in prev_fast_dict_im[fi])]
 *                 for idx_prev, cur_intensity in all_prevs:             # <<<<<<<<<<<<<<
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)
 */
        __pyx_t_10 = __pyx_v_all_prevs; __Pyx_INCREF(__pyx_t_10); __pyx_t_19 = 0;
        for (;;) {
          if (__pyx_t_19 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_19); __Pyx_INCREF(__pyx_t_6); __pyx_t_19++; if (unlikely(0 < 0)) __PYX_ERR(0, 580, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_10, __pyx_t_19); __pyx_t_19++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 580, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
          if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
            PyObject* sequence = __pyx_t_6;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 580, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_15 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_15 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_11 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(__pyx_t_11);
            #else
            __pyx_t_15 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 580, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_11 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 580, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_8 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 580, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_13 = Py_TYPE(__pyx_t_8)->tp_iternext;
            index = 0; __pyx_t_15 = __pyx_t_13(__pyx_t_8); if (unlikely(!__pyx_t_15)) goto __pyx_L32_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_15);
            index = 1; __pyx_t_11 = __pyx_t_13(__pyx_t_8); if (unlikely(!__pyx_t_11)) goto __pyx_L32_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_11);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_8), 2) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
            __pyx_t_13 = NULL;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L33_unpacking_done;
            __pyx_L32_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_13 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 580, __pyx_L1_error)
            __pyx_L33_unpacking_done:;
          }
          __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_15); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_3 = __pyx_PyFloat_AsFloat(__pyx_t_11); if (unlikely((__pyx_t_3 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_idx_prev = __pyx_t_17;
          __pyx_v_cur_intensity = __pyx_t_3;

          /* "biosaur2/cutils.pyx":581
 *                 all_prevs = [[idx_prev, data_for_analyse_tmp[spec_idx-1]['intensity array'][idx_prev]] for idx_prev in all_idx if idx_prev not in banned_prev_idx_set and (paseftol is False or idx_prev in prev_fast_dict_im[fi])]
 *                 for idx_prev, cur_intensity in all_prevs:
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6             # <<<<<<<<<<<<<<
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)
 *                     if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:
 */
          __pyx_t_6 = PyFloat_FromDouble(__pyx_v_mz_cur); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 581, __pyx_L1_error)
          }
          __pyx_t_21 = (__pyx_v_spec_idx - 1);
          __pyx_t_11 = __Pyx_GetItemInt_List(__pyx_v_data_for_analyse_tmp, __pyx_t_21, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_15 = __Pyx_PyObject_Dict_GetItem(__pyx_t_11, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_15, __pyx_v_idx_prev, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_15 = PyNumber_Subtract(__pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = PyFloat_FromDouble(__pyx_v_mz_cur); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_15, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = PyNumber_Multiply(__pyx_t_6, __pyx_float_1e6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_3 = __pyx_PyFloat_AsFloat(__pyx_t_11); if (unlikely((__pyx_t_3 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_cur_mass_diff_with_sign = __pyx_t_3;

          /* "biosaur2/cutils.pyx":582
 *                 for idx_prev, cur_intensity in all_prevs:
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)             # <<<<<<<<<<<<<<
 *                     if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:
 *                         best_mass_diff = cur_mass_diff
 */
          __pyx_v_cur_mass_diff = fabsf(__pyx_v_cur_mass_diff_with_sign);

          /* "biosaur2/cutils.pyx":583
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)
 *                     if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:             # <<<<<<<<<<<<<<
 *                         best_mass_diff = cur_mass_diff
 *                         best_intensity = cur_intensity
 */
          __pyx_t_18 = ((__pyx_v_cur_mass_diff <= __pyx_v_hill_mass_accuracy) != 0);
          if (__pyx_t_18) {
          } else {
            __pyx_t_2 = __pyx_t_18;
            goto __pyx_L35_bool_binop_done;
          }
          __pyx_t_18 = ((__pyx_v_cur_intensity >= __pyx_v_best_intensity) != 0);
          __pyx_t_2 = __pyx_t_18;
          __pyx_L35_bool_binop_done:;
          if (__pyx_t_2) {

            /* "biosaur2/cutils.pyx":584
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)
 *                     if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:
 *                         best_mass_diff = cur_mass_diff             # <<<<<<<<<<<<<<
 *                         best_intensity = cur_intensity
 *                         best_idx_prev = idx_prev
 */
            __pyx_v_best_mass_diff = __pyx_v_cur_mass_diff;

            /* "biosaur2/cutils.pyx":585
 *                     if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:
 *                         best_mass_diff = cur_mass_diff
 *                         best_intensity = cur_intensity             # <<<<<<<<<<<<<<
 *                         best_idx_prev = idx_prev
 *                         hills_dict['hills_idx_array'][last_idx+1+idx] = hills_dict['hills_idx_array'][prev_idx+1+idx_prev]
 */
            __pyx_v_best_intensity = __pyx_v_cur_intensity;

            /* "biosaur2/cutils.pyx":586
 *                         best_mass_diff = cur_mass_diff
 *                         best_intensity = cur_intensity
 *                         best_idx_prev = idx_prev             # <<<<<<<<<<<<<<
 *                         hills_dict['hills_idx_array'][last_idx+1+idx] = hills_dict['hills_idx_array'][prev_idx+1+idx_prev]
 * 
 */
            __pyx_v_best_idx_prev = __pyx_v_idx_prev;

            /* "biosaur2/cutils.pyx":587
 *                         best_intensity = cur_intensity
 *                         best_idx_prev = idx_prev
 *                         hills_dict['hills_idx_array'][last_idx+1+idx] = hills_dict['hills_idx_array'][prev_idx+1+idx_prev]             # <<<<<<<<<<<<<<
 * 
 *                 if best_idx_prev != 0:
 */
            __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 587, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_21 = ((__pyx_v_prev_idx + 1) + __pyx_v_idx_prev);
            __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_11, __pyx_t_21, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 587, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 587, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_21 = ((__pyx_v_last_idx + 1) + __pyx_v_idx);
            if (unlikely(__Pyx_SetItemInt(__pyx_t_11, __pyx_t_21, __pyx_t_6, long, 1, __Pyx_PyInt_From_long, 0, 1, 0) < 0)) __PYX_ERR(0, 587, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "biosaur2/cutils.pyx":583
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)
 *                     if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:             # <<<<<<<<<<<<<<
 *                         best_mass_diff = cur_mass_diff
 *                         best_intensity = cur_intensity
 */
          }

          /* "biosaur2/cutils.pyx":580
 * 
 *                 all_prevs = [[idx_prev, data_for_analyse_tmp[spec_idx-1]['intensity array'][idx_prev]] for idx_prev in all_idx if idx_prev not in banned_prev_idx_set and (paseftol is False or idx_prev in prev_fast_dict_im[fi])]
 *                 for idx_prev, cur_intensity in all_prevs:             # <<<<<<<<<<<<<<
 *                     cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
 *                     cur_mass_diff = abs(cur_mass_diff_with_sign)
 */
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "biosaur2/cutils.pyx":589
 *                         hills_dict['hills_idx_array'][last_idx+1+idx] = hills_dict['hills_idx_array'][prev_idx+1+idx_prev]
 * 
 *                 if best_idx_prev != 0:             # <<<<<<<<<<<<<<
 *                     banned_prev_idx_set.add(best_idx_prev)
 * 
 */
        __pyx_t_2 = ((__pyx_v_best_idx_prev != 0) != 0);
        if (__pyx_t_2) {

          /* "biosaur2/cutils.pyx":590
 * 
 *                 if best_idx_prev != 0:
 *                     banned_prev_idx_set.add(best_idx_prev)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_best_idx_prev); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 590, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_22 = PySet_Add(__pyx_v_banned_prev_idx_set, __pyx_t_10); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 590, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "biosaur2/cutils.pyx":589
 *                         hills_dict['hills_idx_array'][last_idx+1+idx] = hills_dict['hills_idx_array'][prev_idx+1+idx_prev]
 * 
 *                 if best_idx_prev != 0:             # <<<<<<<<<<<<<<
 *                     banned_prev_idx_set.add(best_idx_prev)
 * 
 */
        }

        /* "biosaur2/cutils.pyx":559
 *             flag3 = fm+1 in prev_fast_dict
 * 
 *             if flag1 or flag2 or flag3:             # <<<<<<<<<<<<<<
 * 
 *                 if flag1:
 */
      }

      /* "biosaur2/cutils.pyx":553
 *         banned_prev_idx_set = set()
 * 
 *         for idx, fm, fi in zip(basic_id_sorted, fast_array, (fast_array if paseftol is False else fast_array_im)):             # <<<<<<<<<<<<<<
 * 
 *             flag1 = fm in prev_fast_dict
 */
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "biosaur2/cutils.pyx":593
 * 
 * 
 *         prev_fast_dict = fast_dict             # <<<<<<<<<<<<<<
 *         if paseftol is not False:
 *             prev_fast_dict_im = fast_dict_im
 */
    if (!(likely(PyDict_CheckExact(__pyx_v_fast_dict))||((__pyx_v_fast_dict) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_fast_dict)->tp_name), 0))) __PYX_ERR(0, 593, __pyx_L1_error)
    __pyx_t_9 = __pyx_v_fast_dict;
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_DECREF_SET(__pyx_v_prev_fast_dict, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "biosaur2/cutils.pyx":594
 * 
 *         prev_fast_dict = fast_dict
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             prev_fast_dict_im = fast_dict_im
 *         prev_idx = last_idx
 */
    __pyx_t_2 = ((__pyx_v_paseftol != 0) != 0);
    if (__pyx_t_2) {

      /* "biosaur2/cutils.pyx":595
 *         prev_fast_dict = fast_dict
 *         if paseftol is not False:
 *             prev_fast_dict_im = fast_dict_im             # <<<<<<<<<<<<<<
 *         prev_idx = last_idx
 *         last_idx = last_idx+len_mz
 */
      if (unlikely(!__pyx_v_fast_dict_im)) { __Pyx_RaiseUnboundLocalError("fast_dict_im"); __PYX_ERR(0, 595, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_v_fast_dict_im);
      __Pyx_XDECREF_SET(__pyx_v_prev_fast_dict_im, __pyx_v_fast_dict_im);

      /* "biosaur2/cutils.pyx":594
 * 
 *         prev_fast_dict = fast_dict
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             prev_fast_dict_im = fast_dict_im
 *         prev_idx = last_idx
 */
    }

    /* "biosaur2/cutils.pyx":596
 *         if paseftol is not False:
 *             prev_fast_dict_im = fast_dict_im
 *         prev_idx = last_idx             # <<<<<<<<<<<<<<
 *         last_idx = last_idx+len_mz
 * 
 */
    __pyx_v_prev_idx = __pyx_v_last_idx;

    /* "biosaur2/cutils.pyx":597
 *             prev_fast_dict_im = fast_dict_im
 *         prev_idx = last_idx
 *         last_idx = last_idx+len_mz             # <<<<<<<<<<<<<<
 * 
 *     return hills_dict
 */
    __pyx_v_last_idx = (__pyx_v_last_idx + __pyx_v_len_mz);

    /* "biosaur2/cutils.pyx":527
 *     total_mass_diff = []
 * 
 *     for spec_idx, z in enumerate(data_for_analyse_tmp):             # <<<<<<<<<<<<<<
 * 
 *         len_mz = len(z['m/z array'])
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":599
 *         last_idx = last_idx+len_mz
 * 
 *     return hills_dict             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_hills_dict);
  __pyx_r = __pyx_v_hills_dict;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":494
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def detect_hills(list data_for_analyse_tmp, dict args, float mz_step, bint paseftol, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict hills_dict, prev_fast_dict, z
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("biosaur2.cutils.detect_hills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hills_dict);
  __Pyx_XDECREF(__pyx_v_prev_fast_dict);
  __Pyx_XDECREF(__pyx_v_z);
  __Pyx_XDECREF(__pyx_v_total_num_hills);
  __Pyx_XDECREF(__pyx_v_total_mass_diff);
  __Pyx_XDECREF(__pyx_v_basic_id_sorted);
  __Pyx_XDECREF(__pyx_v_all_idx);
  __Pyx_XDECREF(__pyx_v_all_prevs);
  __Pyx_XDECREF(__pyx_v_banned_prev_idx_set);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_for_sort);
  __Pyx_XDECREF((PyObject *)__pyx_v_mz_sorted);
  __Pyx_XDECREF((PyObject *)__pyx_v_im_sorted);
  __Pyx_XDECREF(__pyx_v_prev_fast_dict_im);
  __Pyx_XDECREF(__pyx_v_fast_dict_im);
  __Pyx_XDECREF(__pyx_v_fast_array_im);
  __Pyx_XDECREF(__pyx_v_fast_dict);
  __Pyx_XDECREF(__pyx_v_fast_array);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "biosaur2/cutils.pyx":605
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def process_hills(dict hills_dict, list data_for_analyse_tmp, float mz_step, bint paseftol, dict args, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8biosaur2_6cutils_23process_hills(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8biosaur2_6cutils_22process_hills[] = "process_hills(dict hills_dict, list data_for_analyse_tmp, float mz_step, bool paseftol, dict args, bool dia=False)";
static PyMethodDef __pyx_mdef_8biosaur2_6cutils_23process_hills = {"process_hills", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8biosaur2_6cutils_23process_hills, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8biosaur2_6cutils_22process_hills};
static PyObject *__pyx_pw_8biosaur2_6cutils_23process_hills(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hills_dict = 0;
  PyObject *__pyx_v_data_for_analyse_tmp = 0;
  float __pyx_v_mz_step;
  int __pyx_v_paseftol;
  PyObject *__pyx_v_args = 0;
  int __pyx_v_dia;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_hills (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hills_dict,&__pyx_n_s_data_for_analyse_tmp,&__pyx_n_s_mz_step,&__pyx_n_s_paseftol,&__pyx_n_s_args,&__pyx_n_s_dia,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hills_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_for_analyse_tmp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("process_hills", 0, 5, 6, 1); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("process_hills", 0, 5, 6, 2); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_paseftol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("process_hills", 0, 5, 6, 3); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("process_hills", 0, 5, 6, 4); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dia);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "process_hills") < 0)) __PYX_ERR(0, 605, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_hills_dict = ((PyObject*)values[0]);
    __pyx_v_data_for_analyse_tmp = ((PyObject*)values[1]);
    __pyx_v_mz_step = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_mz_step == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L3_error)
    __pyx_v_paseftol = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_paseftol == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L3_error)
    __pyx_v_args = ((PyObject*)values[4]);
    if (values[5]) {
      __pyx_v_dia = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_dia == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L3_error)
    } else {
      __pyx_v_dia = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_hills", 0, 5, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 605, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("biosaur2.cutils.process_hills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hills_dict), (&PyDict_Type), 1, "hills_dict", 1))) __PYX_ERR(0, 605, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data_for_analyse_tmp), (&PyList_Type), 1, "data_for_analyse_tmp", 1))) __PYX_ERR(0, 605, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyDict_Type), 1, "args", 1))) __PYX_ERR(0, 605, __pyx_L1_error)
  __pyx_r = __pyx_pf_8biosaur2_6cutils_22process_hills(__pyx_self, __pyx_v_hills_dict, __pyx_v_data_for_analyse_tmp, __pyx_v_mz_step, __pyx_v_paseftol, __pyx_v_args, __pyx_v_dia);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8biosaur2_6cutils_22process_hills(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_hills_dict, PyObject *__pyx_v_data_for_analyse_tmp, float __pyx_v_mz_step, int __pyx_v_paseftol, PyObject *__pyx_v_args, int __pyx_v_dia) {
  float __pyx_v_mz_median;
  float __pyx_v_i_sum_tmp;
  float __pyx_v_mz_val_tmp;
  float __pyx_v_i_val_tmp;
  float __pyx_v_im_median;
  PyObject *__pyx_v_tmp_intensity = 0;
  PyObject *__pyx_v_tmp_mz_array = 0;
  PyObject *__pyx_v_tmp_im_array = 0;
  PyObject *__pyx_v_tmp_scans_list = 0;
  PyObject *__pyx_v_tmp_scans = 0;
  PyObject *__pyx_v_tmp_orig_idx = 0;
  PyObject *__pyx_v_tmp_scans_set = 0;
  PyObject *__pyx_v_tmp_val = 0;
  int __pyx_v_min_length_hill;
  int __pyx_v_idx_start;
  int __pyx_v_idx_end;
  int __pyx_v_idx_1;
  int __pyx_v_hill_idx;
  int __pyx_v_hill_length;
  int __pyx_v_mz_median_int;
  int __pyx_v_im_median_int;
  PyArrayObject *__pyx_v_tmp_hill_length = 0;
  PyArrayObject *__pyx_v_idx_minl = 0;
  PyObject *__pyx_v_counter_hills_idx = NULL;
  PyObject *__pyx_v_idx_sort = NULL;
  PyObject *__pyx_v_orig_idx_val = NULL;
  PyObject *__pyx_v_scan_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  PyObject *(*__pyx_t_16)(PyObject *);
  float __pyx_t_17;
  float __pyx_t_18;
  long __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("process_hills", 0);

  /* "biosaur2/cutils.pyx":616
 * 
 * 
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])             # <<<<<<<<<<<<<<
 *     if dia is False:
 *         min_length_hill = args['minlh']
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Counter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 616, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_counter_hills_idx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":617
 * 
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])
 *     if dia is False:             # <<<<<<<<<<<<<<
 *         min_length_hill = args['minlh']
 *     else:
 */
  __pyx_t_5 = ((__pyx_v_dia == 0) != 0);
  if (__pyx_t_5) {

    /* "biosaur2/cutils.pyx":618
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])
 *     if dia is False:
 *         min_length_hill = args['minlh']             # <<<<<<<<<<<<<<
 *     else:
 *         min_length_hill = args['diaminlh']
 */
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 618, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_minlh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_min_length_hill = __pyx_t_6;

    /* "biosaur2/cutils.pyx":617
 * 
 *     counter_hills_idx = Counter(hills_dict['hills_idx_array'])
 *     if dia is False:             # <<<<<<<<<<<<<<
 *         min_length_hill = args['minlh']
 *     else:
 */
    goto __pyx_L3;
  }

  /* "biosaur2/cutils.pyx":620
 *         min_length_hill = args['minlh']
 *     else:
 *         min_length_hill = args['diaminlh']             # <<<<<<<<<<<<<<
 * 
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
 */
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 620, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_args, __pyx_n_s_diaminlh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_min_length_hill = __pyx_t_6;
  }
  __pyx_L3:;

  /* "biosaur2/cutils.pyx":622
 *         min_length_hill = args['diaminlh']
 * 
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])             # <<<<<<<<<<<<<<
 *     hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 622, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 622, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":623
 * 
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
 *     hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])             # <<<<<<<<<<<<<<
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 623, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 623, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":624
 *     hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
 *     hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
 *     hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 624, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 624, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":627
 * 
 * 
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])             # <<<<<<<<<<<<<<
 *     idx_minl = tmp_hill_length >= min_length_hill
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 627, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_7 = __pyx_t_4; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 627, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 627, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_9(__pyx_t_7);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 627, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_hill_idx = __pyx_t_6;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_counter_hills_idx, __pyx_v_hill_idx, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 627, __pyx_L1_error)
  __pyx_v_tmp_hill_length = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":628
 * 
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
 *     idx_minl = tmp_hill_length >= min_length_hill             # <<<<<<<<<<<<<<
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_min_length_hill); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(((PyObject *)__pyx_v_tmp_hill_length), __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 628, __pyx_L1_error)
  __pyx_v_idx_minl = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":629
 *     tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
 *     idx_minl = tmp_hill_length >= min_length_hill
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]             # <<<<<<<<<<<<<<
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 629, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, ((PyObject *)__pyx_v_idx_minl)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 629, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":630
 *     idx_minl = tmp_hill_length >= min_length_hill
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]             # <<<<<<<<<<<<<<
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 * 
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 630, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, ((PyObject *)__pyx_v_idx_minl)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 630, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_2) < 0)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":631
 *     hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
 *     hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]             # <<<<<<<<<<<<<<
 * 
 *     if len(hills_dict['hills_idx_array']):
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 631, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, ((PyObject *)__pyx_v_idx_minl)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 631, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":633
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 * 
 *     if len(hills_dict['hills_idx_array']):             # <<<<<<<<<<<<<<
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 */
  if (unlikely(__pyx_v_hills_dict == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 633, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_8 != 0);
  if (__pyx_t_5) {

    /* "biosaur2/cutils.pyx":635
 *     if len(hills_dict['hills_idx_array']):
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))             # <<<<<<<<<<<<<<
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_argsort); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 635, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 635, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 635, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_max); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_4 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyNumber_Divide(__pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyNumber_Add(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_idx_sort = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":636
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]             # <<<<<<<<<<<<<<
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 636, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_idx_sort); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 636, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_3) < 0)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":637
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]             # <<<<<<<<<<<<<<
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 637, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_idx_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 637, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_1) < 0)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "biosaur2/cutils.pyx":638
 *         hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
 *         hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 638, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_idx_sort); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 638, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_3) < 0)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":640
 *         hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]
 * 
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))             # <<<<<<<<<<<<<<
 *         hills_dict['hills_mz_median'] = []
 *         hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 640, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PySequence_List(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = ((PyObject*)__pyx_t_11);
    __pyx_t_11 = 0;
    __pyx_t_12 = PyList_Sort(__pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 640, __pyx_L1_error)
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 640, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique, __pyx_t_3) < 0)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":641
 * 
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))
 *         hills_dict['hills_mz_median'] = []             # <<<<<<<<<<<<<<
 *         hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)
 *         if paseftol is not False:
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 641, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median, __pyx_t_3) < 0)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":642
 *         hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))
 *         hills_dict['hills_mz_median'] = []
 *         hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)             # <<<<<<<<<<<<<<
 *         if paseftol is not False:
 *             hills_dict['hills_im_median'] = []
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_1, ((PyObject *)(&PyList_Type))) : __Pyx_PyObject_CallOneArg(__pyx_t_11, ((PyObject *)(&PyList_Type)));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 642, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median_fast_dict, __pyx_t_3) < 0)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":643
 *         hills_dict['hills_mz_median'] = []
 *         hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             hills_dict['hills_im_median'] = []
 *             hills_dict['hills_im_median_fast_dict'] = defaultdict(set)
 */
    __pyx_t_5 = ((__pyx_v_paseftol != 0) != 0);
    if (__pyx_t_5) {

      /* "biosaur2/cutils.pyx":644
 *         hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)
 *         if paseftol is not False:
 *             hills_dict['hills_im_median'] = []             # <<<<<<<<<<<<<<
 *             hills_dict['hills_im_median_fast_dict'] = defaultdict(set)
 * 
 */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 644, __pyx_L1_error)
      }
      if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median, __pyx_t_3) < 0)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "biosaur2/cutils.pyx":645
 *         if paseftol is not False:
 *             hills_dict['hills_im_median'] = []
 *             hills_dict['hills_im_median_fast_dict'] = defaultdict(set)             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['hills_intensity_array'] = []
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
        }
      }
      __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_1, ((PyObject *)(&PySet_Type))) : __Pyx_PyObject_CallOneArg(__pyx_t_11, ((PyObject *)(&PySet_Type)));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 645, __pyx_L1_error)
      }
      if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median_fast_dict, __pyx_t_3) < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "biosaur2/cutils.pyx":643
 *         hills_dict['hills_mz_median'] = []
 *         hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)
 *         if paseftol is not False:             # <<<<<<<<<<<<<<
 *             hills_dict['hills_im_median'] = []
 *             hills_dict['hills_im_median_fast_dict'] = defaultdict(set)
 */
    }

    /* "biosaur2/cutils.pyx":647
 *             hills_dict['hills_im_median_fast_dict'] = defaultdict(set)
 * 
 *         hills_dict['hills_intensity_array'] = []             # <<<<<<<<<<<<<<
 *         hills_dict['hills_scan_sets'] = []
 *         hills_dict['hills_scan_lists'] = []
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 647, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_array, __pyx_t_3) < 0)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":648
 * 
 *         hills_dict['hills_intensity_array'] = []
 *         hills_dict['hills_scan_sets'] = []             # <<<<<<<<<<<<<<
 *         hills_dict['hills_scan_lists'] = []
 *         hills_dict['hills_lengths'] = []
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 648, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_sets, __pyx_t_3) < 0)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":649
 *         hills_dict['hills_intensity_array'] = []
 *         hills_dict['hills_scan_sets'] = []
 *         hills_dict['hills_scan_lists'] = []             # <<<<<<<<<<<<<<
 *         hills_dict['hills_lengths'] = []
 *         hills_dict['tmp_mz_array'] = []
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 649, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_lists, __pyx_t_3) < 0)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":650
 *         hills_dict['hills_scan_sets'] = []
 *         hills_dict['hills_scan_lists'] = []
 *         hills_dict['hills_lengths'] = []             # <<<<<<<<<<<<<<
 *         hills_dict['tmp_mz_array'] = []
 * 
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 650, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_lengths, __pyx_t_3) < 0)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":651
 *         hills_dict['hills_scan_lists'] = []
 *         hills_dict['hills_lengths'] = []
 *         hills_dict['tmp_mz_array'] = []             # <<<<<<<<<<<<<<
 * 
 *         hills_dict['scan_idx_array'] = list(hills_dict['scan_idx_array'])
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 651, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_tmp_mz_array, __pyx_t_3) < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":653
 *         hills_dict['tmp_mz_array'] = []
 * 
 *         hills_dict['scan_idx_array'] = list(hills_dict['scan_idx_array'])             # <<<<<<<<<<<<<<
 *         hills_dict['orig_idx_array'] = list(hills_dict['orig_idx_array'])
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 653, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 653, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array, __pyx_t_11) < 0)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "biosaur2/cutils.pyx":654
 * 
 *         hills_dict['scan_idx_array'] = list(hills_dict['scan_idx_array'])
 *         hills_dict['orig_idx_array'] = list(hills_dict['orig_idx_array'])             # <<<<<<<<<<<<<<
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 654, __pyx_L1_error)
    }
    __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_3 = PySequence_List(__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 654, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array, __pyx_t_3) < 0)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":655
 *         hills_dict['scan_idx_array'] = list(hills_dict['scan_idx_array'])
 *         hills_dict['orig_idx_array'] = list(hills_dict['orig_idx_array'])
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])             # <<<<<<<<<<<<<<
 * 
 *         idx_start = 0
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 655, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 655, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array, __pyx_t_11) < 0)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "biosaur2/cutils.pyx":657
 *         hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])
 * 
 *         idx_start = 0             # <<<<<<<<<<<<<<
 *         idx_end = 0
 * 
 */
    __pyx_v_idx_start = 0;

    /* "biosaur2/cutils.pyx":658
 * 
 *         idx_start = 0
 *         idx_end = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_idx_end = 0;

    /* "biosaur2/cutils.pyx":661
 * 
 * 
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):             # <<<<<<<<<<<<<<
 * 
 *             hill_length = counter_hills_idx[hill_idx]
 */
    __pyx_t_6 = 0;
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 661, __pyx_L1_error)
    }
    __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
      __pyx_t_3 = __pyx_t_11; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 661, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 661, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_11); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 661, __pyx_L1_error)
          #else
          __pyx_t_11 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 661, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_11); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 661, __pyx_L1_error)
          #else
          __pyx_t_11 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 661, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
        }
      } else {
        __pyx_t_11 = __pyx_t_9(__pyx_t_3);
        if (unlikely(!__pyx_t_11)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 661, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_11);
      }
      __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 661, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_v_hill_idx = __pyx_t_13;
      __pyx_v_idx_1 = __pyx_t_6;
      __pyx_t_6 = (__pyx_t_6 + 1);

      /* "biosaur2/cutils.pyx":663
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):
 * 
 *             hill_length = counter_hills_idx[hill_idx]             # <<<<<<<<<<<<<<
 *             idx_end = idx_start + hill_length
 * 
 */
      __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_counter_hills_idx, __pyx_v_hill_idx, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_v_hill_length = __pyx_t_13;

      /* "biosaur2/cutils.pyx":664
 * 
 *             hill_length = counter_hills_idx[hill_idx]
 *             idx_end = idx_start + hill_length             # <<<<<<<<<<<<<<
 * 
 *             tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]
 */
      __pyx_v_idx_end = (__pyx_v_idx_start + __pyx_v_hill_length);

      /* "biosaur2/cutils.pyx":666
 *             idx_end = idx_start + hill_length
 * 
 *             tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]             # <<<<<<<<<<<<<<
 *             tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]
 * 
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 666, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_scan_idx_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 666, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_11, __pyx_v_idx_start, __pyx_v_idx_end, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 666, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 666, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_tmp_scans, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "biosaur2/cutils.pyx":667
 * 
 *             tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]
 *             tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]             # <<<<<<<<<<<<<<
 * 
 *             tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 667, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_orig_idx_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_t_1, __pyx_v_idx_start, __pyx_v_idx_end, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_11))||((__pyx_t_11) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_11)->tp_name), 0))) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_tmp_orig_idx, ((PyObject*)__pyx_t_11));
      __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":669
 *             tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]
 * 
 *             tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]             # <<<<<<<<<<<<<<
 *             tmp_mz_array = [data_for_analyse_tmp[scan_val]['m/z array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *             mz_median = 0
 */
      __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_tmp_orig_idx);
      __Pyx_GIVEREF(__pyx_v_tmp_orig_idx);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_tmp_orig_idx);
      __Pyx_INCREF(__pyx_v_tmp_scans);
      __Pyx_GIVEREF(__pyx_v_tmp_scans);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_tmp_scans);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_15 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 669, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_15(__pyx_t_1);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 669, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
          PyObject* sequence = __pyx_t_4;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 669, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_16 = Py_TYPE(__pyx_t_10)->tp_iternext;
          index = 0; __pyx_t_2 = __pyx_t_16(__pyx_t_10); if (unlikely(!__pyx_t_2)) goto __pyx_L12_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_7 = __pyx_t_16(__pyx_t_10); if (unlikely(!__pyx_t_7)) goto __pyx_L12_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_10), 2) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L13_unpacking_done;
          __pyx_L12_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 669, __pyx_L1_error)
          __pyx_L13_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_orig_idx_val, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_scan_val, __pyx_t_7);
        __pyx_t_7 = 0;
        if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 669, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_data_for_analyse_tmp, __pyx_v_scan_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_kp_s_intensity_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_7, __pyx_v_orig_idx_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_11, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tmp_intensity, ((PyObject*)__pyx_t_11));
      __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":670
 * 
 *             tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *             tmp_mz_array = [data_for_analyse_tmp[scan_val]['m/z array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]             # <<<<<<<<<<<<<<
 *             mz_median = 0
 *             i_sum_tmp = 0
 */
      __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_tmp_orig_idx);
      __Pyx_GIVEREF(__pyx_v_tmp_orig_idx);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_tmp_orig_idx);
      __Pyx_INCREF(__pyx_v_tmp_scans);
      __Pyx_GIVEREF(__pyx_v_tmp_scans);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_tmp_scans);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_15 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 670, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_15(__pyx_t_1);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 670, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
          PyObject* sequence = __pyx_t_4;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 670, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_2);
          #else
          __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 670, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 670, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_16 = Py_TYPE(__pyx_t_10)->tp_iternext;
          index = 0; __pyx_t_7 = __pyx_t_16(__pyx_t_10); if (unlikely(!__pyx_t_7)) goto __pyx_L16_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          index = 1; __pyx_t_2 = __pyx_t_16(__pyx_t_10); if (unlikely(!__pyx_t_2)) goto __pyx_L16_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_10), 2) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L17_unpacking_done;
          __pyx_L16_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 670, __pyx_L1_error)
          __pyx_L17_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_orig_idx_val, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_scan_val, __pyx_t_2);
        __pyx_t_2 = 0;
        if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 670, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_data_for_analyse_tmp, __pyx_v_scan_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_kp_s_m_z_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_orig_idx_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_11, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tmp_mz_array, ((PyObject*)__pyx_t_11));
      __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":671
 *             tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *             tmp_mz_array = [data_for_analyse_tmp[scan_val]['m/z array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *             mz_median = 0             # <<<<<<<<<<<<<<
 *             i_sum_tmp = 0
 *             for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):
 */
      __pyx_v_mz_median = 0.0;

      /* "biosaur2/cutils.pyx":672
 *             tmp_mz_array = [data_for_analyse_tmp[scan_val]['m/z array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *             mz_median = 0
 *             i_sum_tmp = 0             # <<<<<<<<<<<<<<
 *             for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):
 *                 mz_median += mz_val_tmp * i_val_tmp
 */
      __pyx_v_i_sum_tmp = 0.0;

      /* "biosaur2/cutils.pyx":673
 *             mz_median = 0
 *             i_sum_tmp = 0
 *             for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):             # <<<<<<<<<<<<<<
 *                 mz_median += mz_val_tmp * i_val_tmp
 *                 i_sum_tmp += i_val_tmp
 */
      __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 673, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_v_tmp_mz_array);
      __Pyx_GIVEREF(__pyx_v_tmp_mz_array);
      PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_tmp_mz_array);
      __Pyx_INCREF(__pyx_v_tmp_intensity);
      __Pyx_GIVEREF(__pyx_v_tmp_intensity);
      PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_tmp_intensity);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_11 = __pyx_t_1; __Pyx_INCREF(__pyx_t_11); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_15 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 673, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_11))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_11)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 673, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_11, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 673, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_11, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_15(__pyx_t_11);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 673, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 673, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          #else
          __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 673, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_16 = Py_TYPE(__pyx_t_7)->tp_iternext;
          index = 0; __pyx_t_4 = __pyx_t_16(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L20_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          index = 1; __pyx_t_2 = __pyx_t_16(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L20_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_7), 2) < 0) __PYX_ERR(0, 673, __pyx_L1_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L21_unpacking_done;
          __pyx_L20_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 673, __pyx_L1_error)
          __pyx_L21_unpacking_done:;
        }
        __pyx_t_17 = __pyx_PyFloat_AsFloat(__pyx_t_4); if (unlikely((__pyx_t_17 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_18 = __pyx_PyFloat_AsFloat(__pyx_t_2); if (unlikely((__pyx_t_18 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 673, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_mz_val_tmp = __pyx_t_17;
        __pyx_v_i_val_tmp = __pyx_t_18;

        /* "biosaur2/cutils.pyx":674
 *             i_sum_tmp = 0
 *             for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):
 *                 mz_median += mz_val_tmp * i_val_tmp             # <<<<<<<<<<<<<<
 *                 i_sum_tmp += i_val_tmp
 *             mz_median = mz_median / i_sum_tmp
 */
        __pyx_v_mz_median = (__pyx_v_mz_median + (__pyx_v_mz_val_tmp * __pyx_v_i_val_tmp));

        /* "biosaur2/cutils.pyx":675
 *             for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):
 *                 mz_median += mz_val_tmp * i_val_tmp
 *                 i_sum_tmp += i_val_tmp             # <<<<<<<<<<<<<<
 *             mz_median = mz_median / i_sum_tmp
 *             if paseftol is not False:
 */
        __pyx_v_i_sum_tmp = (__pyx_v_i_sum_tmp + __pyx_v_i_val_tmp);

        /* "biosaur2/cutils.pyx":673
 *             mz_median = 0
 *             i_sum_tmp = 0
 *             for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):             # <<<<<<<<<<<<<<
 *                 mz_median += mz_val_tmp * i_val_tmp
 *                 i_sum_tmp += i_val_tmp
 */
      }
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":676
 *                 mz_median += mz_val_tmp * i_val_tmp
 *                 i_sum_tmp += i_val_tmp
 *             mz_median = mz_median / i_sum_tmp             # <<<<<<<<<<<<<<
 *             if paseftol is not False:
 *                 tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 */
      __pyx_v_mz_median = (__pyx_v_mz_median / __pyx_v_i_sum_tmp);

      /* "biosaur2/cutils.pyx":677
 *                 i_sum_tmp += i_val_tmp
 *             mz_median = mz_median / i_sum_tmp
 *             if paseftol is not False:             # <<<<<<<<<<<<<<
 *                 tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *                 im_median = np.average(tmp_im_array, weights=tmp_intensity)
 */
      __pyx_t_5 = ((__pyx_v_paseftol != 0) != 0);
      if (__pyx_t_5) {

        /* "biosaur2/cutils.pyx":678
 *             mz_median = mz_median / i_sum_tmp
 *             if paseftol is not False:
 *                 tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]             # <<<<<<<<<<<<<<
 *                 im_median = np.average(tmp_im_array, weights=tmp_intensity)
 *             tmp_scans_list = tmp_scans
 */
        __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_tmp_orig_idx);
        __Pyx_GIVEREF(__pyx_v_tmp_orig_idx);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_tmp_orig_idx);
        __Pyx_INCREF(__pyx_v_tmp_scans);
        __Pyx_GIVEREF(__pyx_v_tmp_scans);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_tmp_scans);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_14 = 0;
          __pyx_t_15 = NULL;
        } else {
          __pyx_t_14 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_15 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 678, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_15)) {
            if (likely(PyList_CheckExact(__pyx_t_1))) {
              if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_1)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 678, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 678, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_15(__pyx_t_1);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 678, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
            PyObject* sequence = __pyx_t_2;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 678, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            #else
            __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 678, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 678, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_10 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 678, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_16 = Py_TYPE(__pyx_t_10)->tp_iternext;
            index = 0; __pyx_t_4 = __pyx_t_16(__pyx_t_10); if (unlikely(!__pyx_t_4)) goto __pyx_L25_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            index = 1; __pyx_t_7 = __pyx_t_16(__pyx_t_10); if (unlikely(!__pyx_t_7)) goto __pyx_L25_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_7);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_10), 2) < 0) __PYX_ERR(0, 678, __pyx_L1_error)
            __pyx_t_16 = NULL;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            goto __pyx_L26_unpacking_done;
            __pyx_L25_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_16 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 678, __pyx_L1_error)
            __pyx_L26_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_orig_idx_val, __pyx_t_4);
          __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_scan_val, __pyx_t_7);
          __pyx_t_7 = 0;
          if (unlikely(__pyx_v_data_for_analyse_tmp == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 678, __pyx_L1_error)
          }
          __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_data_for_analyse_tmp, __pyx_v_scan_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_kp_s_mean_inverse_reduced_ion_mobilit); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_7, __pyx_v_orig_idx_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_11, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 678, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_tmp_im_array, ((PyObject*)__pyx_t_11));
        __pyx_t_11 = 0;

        /* "biosaur2/cutils.pyx":679
 *             if paseftol is not False:
 *                 tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *                 im_median = np.average(tmp_im_array, weights=tmp_intensity)             # <<<<<<<<<<<<<<
 *             tmp_scans_list = tmp_scans
 *             tmp_scans_set = set(tmp_scans)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_average); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_v_tmp_im_array);
        __Pyx_GIVEREF(__pyx_v_tmp_im_array);
        PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_tmp_im_array);
        __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_weights, __pyx_v_tmp_intensity) < 0) __PYX_ERR(0, 679, __pyx_L1_error)
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_18 = __pyx_PyFloat_AsFloat(__pyx_t_7); if (unlikely((__pyx_t_18 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_im_median = __pyx_t_18;

        /* "biosaur2/cutils.pyx":677
 *                 i_sum_tmp += i_val_tmp
 *             mz_median = mz_median / i_sum_tmp
 *             if paseftol is not False:             # <<<<<<<<<<<<<<
 *                 tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *                 im_median = np.average(tmp_im_array, weights=tmp_intensity)
 */
      }

      /* "biosaur2/cutils.pyx":680
 *                 tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
 *                 im_median = np.average(tmp_im_array, weights=tmp_intensity)
 *             tmp_scans_list = tmp_scans             # <<<<<<<<<<<<<<
 *             tmp_scans_set = set(tmp_scans)
 * 
 */
      __Pyx_INCREF(__pyx_v_tmp_scans);
      __Pyx_XDECREF_SET(__pyx_v_tmp_scans_list, __pyx_v_tmp_scans);

      /* "biosaur2/cutils.pyx":681
 *                 im_median = np.average(tmp_im_array, weights=tmp_intensity)
 *             tmp_scans_list = tmp_scans
 *             tmp_scans_set = set(tmp_scans)             # <<<<<<<<<<<<<<
 * 
 *             idx_start = idx_end
 */
      __pyx_t_7 = PySet_New(__pyx_v_tmp_scans); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 681, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_tmp_scans_set, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "biosaur2/cutils.pyx":683
 *             tmp_scans_set = set(tmp_scans)
 * 
 *             idx_start = idx_end             # <<<<<<<<<<<<<<
 * 
 *             hills_dict['hills_mz_median'].append(mz_median)
 */
      __pyx_v_idx_start = __pyx_v_idx_end;

      /* "biosaur2/cutils.pyx":685
 *             idx_start = idx_end
 * 
 *             hills_dict['hills_mz_median'].append(mz_median)             # <<<<<<<<<<<<<<
 * 
 *             mz_median_int = int(mz_median/mz_step)
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 685, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = PyFloat_FromDouble(__pyx_v_mz_median); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_t_2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 685, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "biosaur2/cutils.pyx":687
 *             hills_dict['hills_mz_median'].append(mz_median)
 * 
 *             mz_median_int = int(mz_median/mz_step)             # <<<<<<<<<<<<<<
 *             tmp_val = (idx_1, tmp_scans_list[0], tmp_scans_list[-1])
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int-1].append(tmp_val)
 */
      __pyx_v_mz_median_int = ((int)(__pyx_v_mz_median / __pyx_v_mz_step));

      /* "biosaur2/cutils.pyx":688
 * 
 *             mz_median_int = int(mz_median/mz_step)
 *             tmp_val = (idx_1, tmp_scans_list[0], tmp_scans_list[-1])             # <<<<<<<<<<<<<<
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int-1].append(tmp_val)
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int].append(tmp_val)
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_v_tmp_scans_list == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 688, __pyx_L1_error)
      }
      if (unlikely(__pyx_v_tmp_scans_list == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 688, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_tmp_scans_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_2);
      __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_tmp_scans_list, 0));
      __Pyx_GIVEREF(PyList_GET_ITEM(__pyx_v_tmp_scans_list, 0));
      PyTuple_SET_ITEM(__pyx_t_11, 1, PyList_GET_ITEM(__pyx_v_tmp_scans_list, 0));
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_t_7);
      __pyx_t_2 = 0;
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tmp_val, ((PyObject*)__pyx_t_11));
      __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":689
 *             mz_median_int = int(mz_median/mz_step)
 *             tmp_val = (idx_1, tmp_scans_list[0], tmp_scans_list[-1])
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int-1].append(tmp_val)             # <<<<<<<<<<<<<<
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int].append(tmp_val)
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int+1].append(tmp_val)
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 689, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median_fast_dict); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_19 = (__pyx_v_mz_median_int - 1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_11, __pyx_t_19, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_v_tmp_val); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "biosaur2/cutils.pyx":690
 *             tmp_val = (idx_1, tmp_scans_list[0], tmp_scans_list[-1])
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int-1].append(tmp_val)
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int].append(tmp_val)             # <<<<<<<<<<<<<<
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int+1].append(tmp_val)
 * 
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 690, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median_fast_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_mz_median_int, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_11, __pyx_v_tmp_val); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 690, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":691
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int-1].append(tmp_val)
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int].append(tmp_val)
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int+1].append(tmp_val)             # <<<<<<<<<<<<<<
 * 
 *             if paseftol is not False:
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 691, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_mz_median_fast_dict); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 691, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_19 = (__pyx_v_mz_median_int + 1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_11, __pyx_t_19, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 691, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_v_tmp_val); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 691, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "biosaur2/cutils.pyx":693
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int+1].append(tmp_val)
 * 
 *             if paseftol is not False:             # <<<<<<<<<<<<<<
 *                 hills_dict['hills_im_median'].append(im_median)
 * 
 */
      __pyx_t_5 = ((__pyx_v_paseftol != 0) != 0);
      if (__pyx_t_5) {

        /* "biosaur2/cutils.pyx":694
 * 
 *             if paseftol is not False:
 *                 hills_dict['hills_im_median'].append(im_median)             # <<<<<<<<<<<<<<
 * 
 *                 im_median_int = int(im_median/paseftol)
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 694, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_im_median); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_t_11); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 694, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "biosaur2/cutils.pyx":696
 *                 hills_dict['hills_im_median'].append(im_median)
 * 
 *                 im_median_int = int(im_median/paseftol)             # <<<<<<<<<<<<<<
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int-1].add(idx_1)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int].add(idx_1)
 */
        __pyx_v_im_median_int = ((int)(__pyx_v_im_median / __pyx_v_paseftol));

        /* "biosaur2/cutils.pyx":697
 * 
 *                 im_median_int = int(im_median/paseftol)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int-1].add(idx_1)             # <<<<<<<<<<<<<<
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int].add(idx_1)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int+1].add(idx_1)
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 697, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median_fast_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_19 = (__pyx_v_im_median_int - 1);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_7, __pyx_t_19, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_add); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_11 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "biosaur2/cutils.pyx":698
 *                 im_median_int = int(im_median/paseftol)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int-1].add(idx_1)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int].add(idx_1)             # <<<<<<<<<<<<<<
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int+1].add(idx_1)
 * 
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 698, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median_fast_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_im_median_int, int, 1, __Pyx_PyInt_From_int, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_add); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_11 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "biosaur2/cutils.pyx":699
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int-1].add(idx_1)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int].add(idx_1)
 *                 hills_dict['hills_im_median_fast_dict'][im_median_int+1].add(idx_1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        if (unlikely(__pyx_v_hills_dict == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 699, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_im_median_fast_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_19 = (__pyx_v_im_median_int + 1);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_7, __pyx_t_19, long, 1, __Pyx_PyInt_From_long, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_add); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_idx_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_11 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "biosaur2/cutils.pyx":693
 *             hills_dict['hills_mz_median_fast_dict'][mz_median_int+1].append(tmp_val)
 * 
 *             if paseftol is not False:             # <<<<<<<<<<<<<<
 *                 hills_dict['hills_im_median'].append(im_median)
 * 
 */
      }

      /* "biosaur2/cutils.pyx":702
 * 
 * 
 *             hills_dict['hills_intensity_array'].append(tmp_intensity)             # <<<<<<<<<<<<<<
 *             hills_dict['hills_scan_sets'].append(tmp_scans_set)
 *             hills_dict['hills_scan_lists'].append(tmp_scans_list)
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 702, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_array); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_11, __pyx_v_tmp_intensity); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":703
 * 
 *             hills_dict['hills_intensity_array'].append(tmp_intensity)
 *             hills_dict['hills_scan_sets'].append(tmp_scans_set)             # <<<<<<<<<<<<<<
 *             hills_dict['hills_scan_lists'].append(tmp_scans_list)
 *             hills_dict['hills_lengths'].append(hill_length)
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 703, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_sets); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_11, __pyx_v_tmp_scans_set); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":704
 *             hills_dict['hills_intensity_array'].append(tmp_intensity)
 *             hills_dict['hills_scan_sets'].append(tmp_scans_set)
 *             hills_dict['hills_scan_lists'].append(tmp_scans_list)             # <<<<<<<<<<<<<<
 *             hills_dict['hills_lengths'].append(hill_length)
 *             hills_dict['tmp_mz_array'].append(tmp_mz_array)
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 704, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_lists); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_11, __pyx_v_tmp_scans_list); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "biosaur2/cutils.pyx":705
 *             hills_dict['hills_scan_sets'].append(tmp_scans_set)
 *             hills_dict['hills_scan_lists'].append(tmp_scans_list)
 *             hills_dict['hills_lengths'].append(hill_length)             # <<<<<<<<<<<<<<
 *             hills_dict['tmp_mz_array'].append(tmp_mz_array)
 * 
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 705, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_lengths); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_hill_length); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_11, __pyx_t_7); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "biosaur2/cutils.pyx":706
 *             hills_dict['hills_scan_lists'].append(tmp_scans_list)
 *             hills_dict['hills_lengths'].append(hill_length)
 *             hills_dict['tmp_mz_array'].append(tmp_mz_array)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      if (unlikely(__pyx_v_hills_dict == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 706, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_tmp_mz_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_12 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_v_tmp_mz_array); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "biosaur2/cutils.pyx":661
 * 
 * 
 *         for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):             # <<<<<<<<<<<<<<
 * 
 *             hill_length = counter_hills_idx[hill_idx]
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":709
 * 
 * 
 *         hills_dict['hills_idict'] = [None] * len(hills_dict['hills_idx_array_unique'])             # <<<<<<<<<<<<<<
 *         hills_dict['hill_sqrt_of_i'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hills_intensity_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 709, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1 * ((__pyx_t_8<0) ? 0:__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_t_8; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, Py_None);
      }
    }
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 709, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idict, __pyx_t_3) < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":710
 * 
 *         hills_dict['hills_idict'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hill_sqrt_of_i'] = [None] * len(hills_dict['hills_idx_array_unique'])             # <<<<<<<<<<<<<<
 *         hills_dict['hills_intensity_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hills_scan_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 710, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1 * ((__pyx_t_8<0) ? 0:__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_t_8; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, Py_None);
      }
    }
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 710, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hill_sqrt_of_i, __pyx_t_3) < 0)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":711
 *         hills_dict['hills_idict'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hill_sqrt_of_i'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hills_intensity_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])             # <<<<<<<<<<<<<<
 *         hills_dict['hills_scan_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
 * 
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 711, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1 * ((__pyx_t_8<0) ? 0:__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_t_8; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, Py_None);
      }
    }
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 711, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_intensity_apex, __pyx_t_3) < 0)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":712
 *         hills_dict['hill_sqrt_of_i'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hills_intensity_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
 *         hills_dict['hills_scan_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])             # <<<<<<<<<<<<<<
 * 
 *     return hills_dict
 */
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 712, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_hills_dict, __pyx_n_s_hills_idx_array_unique); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1 * ((__pyx_t_8<0) ? 0:__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_t_8; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, Py_None);
      }
    }
    if (unlikely(__pyx_v_hills_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 712, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_hills_dict, __pyx_n_s_hills_scan_apex, __pyx_t_3) < 0)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "biosaur2/cutils.pyx":633
 *     hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]
 * 
 *     if len(hills_dict['hills_idx_array']):             # <<<<<<<<<<<<<<
 * 
 *         idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
 */
  }

  /* "biosaur2/cutils.pyx":714
 *         hills_dict['hills_scan_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
 * 
 *     return hills_dict             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_hills_dict);
  __pyx_r = __pyx_v_hills_dict;
  goto __pyx_L0;

  /* "biosaur2/cutils.pyx":605
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def process_hills(dict hills_dict, list data_for_analyse_tmp, float mz_step, bint paseftol, dict args, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("biosaur2.cutils.process_hills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp_intensity);
  __Pyx_XDECREF(__pyx_v_tmp_mz_array);
  __Pyx_XDECREF(__pyx_v_tmp_im_array);
  __Pyx_XDECREF(__pyx_v_tmp_scans_list);
  __Pyx_XDECREF(__pyx_v_tmp_scans);
  __Pyx_XDECREF(__pyx_v_tmp_orig_idx);
  __Pyx_XDECREF(__pyx_v_tmp_scans_set);
  __Pyx_XDECREF(__pyx_v_tmp_val);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp_hill_length);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_minl);
  __Pyx_XDECREF(__pyx_v_counter_hills_idx);
  __Pyx_XDECREF(__pyx_v_idx_sort);
  __Pyx_XDECREF(__pyx_v_orig_idx_val);
  __Pyx_XDECREF(__pyx_v_scan_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":930
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":931
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":934
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":936
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":937
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":943
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L3_error)

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":944
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 945, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 945, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":949
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 949, __pyx_L3_error)

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":950
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 951, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 951, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":955
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 955, __pyx_L3_error)

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":956
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":957
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 957, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 957, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":979
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":994
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1004
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1011
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1018
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *__pyx_freelist_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr[8];
static int __pyx_freecount_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr = 0;

static PyObject *__pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr)))) {
    o = (PyObject*)__pyx_freelist_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr[--__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr];
    memset(o, 0, sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr(PyObject *o) {
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *p = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_hill_idict_1);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr)))) {
    __pyx_freelist_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr[__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr++] = ((struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *p = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *)o;
  if (p->__pyx_v_hill_idict_1) {
    e = (*v)(p->__pyx_v_hill_idict_1, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *p = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr *)o;
  tmp = ((PyObject*)p->__pyx_v_hill_idict_1);
  p->__pyx_v_hill_idict_1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr = {
  PyVarObject_HEAD_INIT(0, 0)
  "biosaur2.cutils.__pyx_scope_struct__get_and_calc_values_for_cos_corr", /*tp_name*/
  sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr, /*tp_traverse*/
  __pyx_tp_clear_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *__pyx_freelist_8biosaur2_6cutils___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_8biosaur2_6cutils___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8biosaur2_6cutils___pyx_scope_struct_1_genexpr[--__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8biosaur2_6cutils___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_8biosaur2_6cutils___pyx_scope_struct_1_genexpr[__pyx_freecount_8biosaur2_6cutils___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8biosaur2_6cutils___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "biosaur2.cutils.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8biosaur2_6cutils___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8biosaur2_6cutils___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8biosaur2_6cutils___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8biosaur2_6cutils___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cutils(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cutils},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "cutils",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Counter, __pyx_k_Counter, sizeof(__pyx_k_Counter), 0, 0, 1, 1},
  {&__pyx_n_s_FAIMS, __pyx_k_FAIMS, sizeof(__pyx_k_FAIMS), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 0, 1, 1},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_all_exp_intensity, __pyx_k_all_exp_intensity, sizeof(__pyx_k_all_exp_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_all_idx, __pyx_k_all_idx, sizeof(__pyx_k_all_idx), 0, 0, 1, 1},
  {&__pyx_n_s_all_prevs, __pyx_k_all_prevs, sizeof(__pyx_k_all_prevs), 0, 0, 1, 1},
  {&__pyx_n_s_all_theoretical_int, __pyx_k_all_theoretical_int, sizeof(__pyx_k_all_theoretical_int), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_argsort, __pyx_k_argsort, sizeof(__pyx_k_argsort), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_average, __pyx_k_average, sizeof(__pyx_k_average), 0, 0, 1, 1},
  {&__pyx_n_s_averagineCorrelation, __pyx_k_averagineCorrelation, sizeof(__pyx_k_averagineCorrelation), 0, 0, 1, 1},
  {&__pyx_n_s_averagineExplained, __pyx_k_averagineExplained, sizeof(__pyx_k_averagineExplained), 0, 0, 1, 1},
  {&__pyx_n_s_banned_charges, __pyx_k_banned_charges, sizeof(__pyx_k_banned_charges), 0, 0, 1, 1},
  {&__pyx_n_s_banned_prev_idx_set, __pyx_k_banned_prev_idx_set, sizeof(__pyx_k_banned_prev_idx_set), 0, 0, 1, 1},
  {&__pyx_n_s_basic_id_sorted, __pyx_k_basic_id_sorted, sizeof(__pyx_k_basic_id_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_best_cor, __pyx_k_best_cor, sizeof(__pyx_k_best_cor), 0, 0, 1, 1},
  {&__pyx_n_s_best_idx_prev, __pyx_k_best_idx_prev, sizeof(__pyx_k_best_idx_prev), 0, 0, 1, 1},
  {&__pyx_n_s_best_intensity, __pyx_k_best_intensity, sizeof(__pyx_k_best_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_best_mass_diff, __pyx_k_best_mass_diff, sizeof(__pyx_k_best_mass_diff), 0, 0, 1, 1},
  {&__pyx_n_s_best_pos, __pyx_k_best_pos, sizeof(__pyx_k_best_pos), 0, 0, 1, 1},
  {&__pyx_n_s_biosaur2_cutils, __pyx_k_biosaur2_cutils, sizeof(__pyx_k_biosaur2_cutils), 0, 0, 1, 1},
  {&__pyx_kp_s_biosaur2_cutils_pyx, __pyx_k_biosaur2_cutils_pyx, sizeof(__pyx_k_biosaur2_cutils_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_bottom, __pyx_k_bottom, sizeof(__pyx_k_bottom), 0, 0, 1, 1},
  {&__pyx_n_s_c_len, __pyx_k_c_len, sizeof(__pyx_k_c_len), 0, 0, 1, 1},
  {&__pyx_n_s_candidates, __pyx_k_candidates, sizeof(__pyx_k_candidates), 0, 0, 1, 1},
  {&__pyx_n_s_ch_v, __pyx_k_ch_v, sizeof(__pyx_k_ch_v), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_charges, __pyx_k_charges, sizeof(__pyx_k_charges), 0, 0, 1, 1},
  {&__pyx_n_s_checked_id, __pyx_k_checked_id, sizeof(__pyx_k_checked_id), 0, 0, 1, 1},
  {&__pyx_n_s_checking_cos_correlation_for_car, __pyx_k_checking_cos_correlation_for_car, sizeof(__pyx_k_checking_cos_correlation_for_car), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_cos_cor, __pyx_k_cos_cor, sizeof(__pyx_k_cos_cor), 0, 0, 1, 1},
  {&__pyx_n_s_cos_cor_RT, __pyx_k_cos_cor_RT, sizeof(__pyx_k_cos_cor_RT), 0, 0, 1, 1},
  {&__pyx_n_s_cos_cor_isotopes, __pyx_k_cos_cor_isotopes, sizeof(__pyx_k_cos_cor_isotopes), 0, 0, 1, 1},
  {&__pyx_n_s_cos_corr, __pyx_k_cos_corr, sizeof(__pyx_k_cos_corr), 0, 0, 1, 1},
  {&__pyx_n_s_cos_correlation, __pyx_k_cos_correlation, sizeof(__pyx_k_cos_correlation), 0, 0, 1, 1},
  {&__pyx_n_s_cos_correlation_new, __pyx_k_cos_correlation_new, sizeof(__pyx_k_cos_correlation_new), 0, 0, 1, 1},
  {&__pyx_n_s_counter_hills_idx, __pyx_k_counter_hills_idx, sizeof(__pyx_k_counter_hills_idx), 0, 0, 1, 1},
  {&__pyx_n_s_cumsum, __pyx_k_cumsum, sizeof(__pyx_k_cumsum), 0, 0, 1, 1},
  {&__pyx_n_s_cumsum_vec, __pyx_k_cumsum_vec, sizeof(__pyx_k_cumsum_vec), 0, 0, 1, 1},
  {&__pyx_n_s_cur_intensity, __pyx_k_cur_intensity, sizeof(__pyx_k_cur_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_cur_mass_diff, __pyx_k_cur_mass_diff, sizeof(__pyx_k_cur_mass_diff), 0, 0, 1, 1},
  {&__pyx_n_s_cur_mass_diff_with_sign, __pyx_k_cur_mass_diff_with_sign, sizeof(__pyx_k_cur_mass_diff_with_sign), 0, 0, 1, 1},
  {&__pyx_n_s_cur_new_idx, __pyx_k_cur_new_idx, sizeof(__pyx_k_cur_new_idx), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_for_analyse_tmp, __pyx_k_data_for_analyse_tmp, sizeof(__pyx_k_data_for_analyse_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_defaultdict, __pyx_k_defaultdict, sizeof(__pyx_k_defaultdict), 0, 0, 1, 1},
  {&__pyx_n_s_detect_hills, __pyx_k_detect_hills, sizeof(__pyx_k_detect_hills), 0, 0, 1, 1},
  {&__pyx_n_s_dia, __pyx_k_dia, sizeof(__pyx_k_dia), 0, 0, 1, 1},
  {&__pyx_n_s_diahtol, __pyx_k_diahtol, sizeof(__pyx_k_diahtol), 0, 0, 1, 1},
  {&__pyx_n_s_diaminlh, __pyx_k_diaminlh, sizeof(__pyx_k_diaminlh), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_end_idx, __pyx_k_end_idx, sizeof(__pyx_k_end_idx), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_experimental_list, __pyx_k_experimental_list, sizeof(__pyx_k_experimental_list), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_faims_val, __pyx_k_faims_val, sizeof(__pyx_k_faims_val), 0, 0, 1, 1},
  {&__pyx_n_s_fast_array, __pyx_k_fast_array, sizeof(__pyx_k_fast_array), 0, 0, 1, 1},
  {&__pyx_n_s_fast_array_im, __pyx_k_fast_array_im, sizeof(__pyx_k_fast_array_im), 0, 0, 1, 1},
  {&__pyx_n_s_fast_dict, __pyx_k_fast_dict, sizeof(__pyx_k_fast_dict), 0, 0, 1, 1},
  {&__pyx_n_s_fast_dict_im, __pyx_k_fast_dict_im, sizeof(__pyx_k_fast_dict_im), 0, 0, 1, 1},
  {&__pyx_n_s_fi, __pyx_k_fi, sizeof(__pyx_k_fi), 0, 0, 1, 1},
  {&__pyx_n_s_flag1, __pyx_k_flag1, sizeof(__pyx_k_flag1), 0, 0, 1, 1},
  {&__pyx_n_s_flag2, __pyx_k_flag2, sizeof(__pyx_k_flag2), 0, 0, 1, 1},
  {&__pyx_n_s_flag3, __pyx_k_flag3, sizeof(__pyx_k_flag3), 0, 0, 1, 1},
  {&__pyx_n_s_fm, __pyx_k_fm, sizeof(__pyx_k_fm), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_and_calc_apex_intensity_and, __pyx_k_get_and_calc_apex_intensity_and, sizeof(__pyx_k_get_and_calc_apex_intensity_and), 0, 0, 1, 1},
  {&__pyx_n_s_get_and_calc_values_for_cos_corr, __pyx_k_get_and_calc_values_for_cos_corr, sizeof(__pyx_k_get_and_calc_values_for_cos_corr), 0, 0, 1, 1},
  {&__pyx_n_s_get_and_calc_values_for_cos_corr_2, __pyx_k_get_and_calc_values_for_cos_corr_2, sizeof(__pyx_k_get_and_calc_values_for_cos_corr_2), 0, 0, 1, 1},
  {&__pyx_n_s_get_fast_dict, __pyx_k_get_fast_dict, sizeof(__pyx_k_get_fast_dict), 0, 0, 1, 1},
  {&__pyx_n_s_get_initial_isotopes, __pyx_k_get_initial_isotopes, sizeof(__pyx_k_get_initial_isotopes), 0, 0, 1, 1},
  {&__pyx_n_s_hillValleyFactor, __pyx_k_hillValleyFactor, sizeof(__pyx_k_hillValleyFactor), 0, 0, 1, 1},
  {&__pyx_n_s_hill_idict_1, __pyx_k_hill_idict_1, sizeof(__pyx_k_hill_idict_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_idict_2, __pyx_k_hill_idict_2, sizeof(__pyx_k_hill_idict_2), 0, 0, 1, 1},
  {&__pyx_n_s_hill_idx, __pyx_k_hill_idx, sizeof(__pyx_k_hill_idx), 0, 0, 1, 1},
  {&__pyx_n_s_hill_idx_1, __pyx_k_hill_idx_1, sizeof(__pyx_k_hill_idx_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_idx_2, __pyx_k_hill_idx_2, sizeof(__pyx_k_hill_idx_2), 0, 0, 1, 1},
  {&__pyx_n_s_hill_intensity_apex_1, __pyx_k_hill_intensity_apex_1, sizeof(__pyx_k_hill_intensity_apex_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_intensity_apex_2, __pyx_k_hill_intensity_apex_2, sizeof(__pyx_k_hill_intensity_apex_2), 0, 0, 1, 1},
  {&__pyx_n_s_hill_length, __pyx_k_hill_length, sizeof(__pyx_k_hill_length), 0, 0, 1, 1},
  {&__pyx_n_s_hill_mass_accuracy, __pyx_k_hill_mass_accuracy, sizeof(__pyx_k_hill_mass_accuracy), 0, 0, 1, 1},
  {&__pyx_n_s_hill_mz_1, __pyx_k_hill_mz_1, sizeof(__pyx_k_hill_mz_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_mz_2, __pyx_k_hill_mz_2, sizeof(__pyx_k_hill_mz_2), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scan_apex_1, __pyx_k_hill_scan_apex_1, sizeof(__pyx_k_hill_scan_apex_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scan_apex_2, __pyx_k_hill_scan_apex_2, sizeof(__pyx_k_hill_scan_apex_2), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_1, __pyx_k_hill_scans_1, sizeof(__pyx_k_hill_scans_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_1_list, __pyx_k_hill_scans_1_list, sizeof(__pyx_k_hill_scans_1_list), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_1_list_first, __pyx_k_hill_scans_1_list_first, sizeof(__pyx_k_hill_scans_1_list_first), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_1_list_last, __pyx_k_hill_scans_1_list_last, sizeof(__pyx_k_hill_scans_1_list_last), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_2, __pyx_k_hill_scans_2, sizeof(__pyx_k_hill_scans_2), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_2_list, __pyx_k_hill_scans_2_list, sizeof(__pyx_k_hill_scans_2_list), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_2_list_first, __pyx_k_hill_scans_2_list_first, sizeof(__pyx_k_hill_scans_2_list_first), 0, 0, 1, 1},
  {&__pyx_n_s_hill_scans_2_list_last, __pyx_k_hill_scans_2_list_last, sizeof(__pyx_k_hill_scans_2_list_last), 0, 0, 1, 1},
  {&__pyx_n_s_hill_sqrt_of_i, __pyx_k_hill_sqrt_of_i, sizeof(__pyx_k_hill_sqrt_of_i), 0, 0, 1, 1},
  {&__pyx_n_s_hill_sqrt_of_i_1, __pyx_k_hill_sqrt_of_i_1, sizeof(__pyx_k_hill_sqrt_of_i_1), 0, 0, 1, 1},
  {&__pyx_n_s_hill_sqrt_of_i_2, __pyx_k_hill_sqrt_of_i_2, sizeof(__pyx_k_hill_sqrt_of_i_2), 0, 0, 1, 1},
  {&__pyx_n_s_hills_dict, __pyx_k_hills_dict, sizeof(__pyx_k_hills_dict), 0, 0, 1, 1},
  {&__pyx_n_s_hills_idict, __pyx_k_hills_idict, sizeof(__pyx_k_hills_idict), 0, 0, 1, 1},
  {&__pyx_n_s_hills_idx_array, __pyx_k_hills_idx_array, sizeof(__pyx_k_hills_idx_array), 0, 0, 1, 1},
  {&__pyx_n_s_hills_idx_array_unique, __pyx_k_hills_idx_array_unique, sizeof(__pyx_k_hills_idx_array_unique), 0, 0, 1, 1},
  {&__pyx_n_s_hills_im_median, __pyx_k_hills_im_median, sizeof(__pyx_k_hills_im_median), 0, 0, 1, 1},
  {&__pyx_n_s_hills_im_median_fast_dict, __pyx_k_hills_im_median_fast_dict, sizeof(__pyx_k_hills_im_median_fast_dict), 0, 0, 1, 1},
  {&__pyx_n_s_hills_intensity_apex, __pyx_k_hills_intensity_apex, sizeof(__pyx_k_hills_intensity_apex), 0, 0, 1, 1},
  {&__pyx_n_s_hills_intensity_array, __pyx_k_hills_intensity_array, sizeof(__pyx_k_hills_intensity_array), 0, 0, 1, 1},
  {&__pyx_n_s_hills_lengths, __pyx_k_hills_lengths, sizeof(__pyx_k_hills_lengths), 0, 0, 1, 1},
  {&__pyx_n_s_hills_mz_median, __pyx_k_hills_mz_median, sizeof(__pyx_k_hills_mz_median), 0, 0, 1, 1},
  {&__pyx_n_s_hills_mz_median_fast_dict, __pyx_k_hills_mz_median_fast_dict, sizeof(__pyx_k_hills_mz_median_fast_dict), 0, 0, 1, 1},
  {&__pyx_n_s_hills_scan_apex, __pyx_k_hills_scan_apex, sizeof(__pyx_k_hills_scan_apex), 0, 0, 1, 1},
  {&__pyx_n_s_hills_scan_lists, __pyx_k_hills_scan_lists, sizeof(__pyx_k_hills_scan_lists), 0, 0, 1, 1},
  {&__pyx_n_s_hills_scan_sets, __pyx_k_hills_scan_sets, sizeof(__pyx_k_hills_scan_sets), 0, 0, 1, 1},
  {&__pyx_n_s_htol, __pyx_k_htol, sizeof(__pyx_k_htol), 0, 0, 1, 1},
  {&__pyx_n_s_hvf, __pyx_k_hvf, sizeof(__pyx_k_hvf), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_i1, __pyx_k_i1, sizeof(__pyx_k_i1), 0, 0, 1, 1},
  {&__pyx_n_s_i2, __pyx_k_i2, sizeof(__pyx_k_i2), 0, 0, 1, 1},
  {&__pyx_n_s_i_sum_tmp, __pyx_k_i_sum_tmp, sizeof(__pyx_k_i_sum_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_i_val_tmp, __pyx_k_i_val_tmp, sizeof(__pyx_k_i_val_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
  {&__pyx_n_s_idx_1, __pyx_k_idx_1, sizeof(__pyx_k_idx_1), 0, 0, 1, 1},
  {&__pyx_n_s_idx_2, __pyx_k_idx_2, sizeof(__pyx_k_idx_2), 0, 0, 1, 1},
  {&__pyx_n_s_idx_end, __pyx_k_idx_end, sizeof(__pyx_k_idx_end), 0, 0, 1, 1},
  {&__pyx_n_s_idx_for_sort, __pyx_k_idx_for_sort, sizeof(__pyx_k_idx_for_sort), 0, 0, 1, 1},
  {&__pyx_n_s_idx_minl, __pyx_k_idx_minl, sizeof(__pyx_k_idx_minl), 0, 0, 1, 1},
  {&__pyx_n_s_idx_prev, __pyx_k_idx_prev, sizeof(__pyx_k_idx_prev), 0, 0, 1, 1},
  {&__pyx_n_s_idx_sort, __pyx_k_idx_sort, sizeof(__pyx_k_idx_sort), 0, 0, 1, 1},
  {&__pyx_n_s_idx_start, __pyx_k_idx_start, sizeof(__pyx_k_idx_start), 0, 0, 1, 1},
  {&__pyx_n_s_im, __pyx_k_im, sizeof(__pyx_k_im), 0, 0, 1, 1},
  {&__pyx_n_s_im_array, __pyx_k_im_array, sizeof(__pyx_k_im_array), 0, 0, 1, 1},
  {&__pyx_n_s_im_median, __pyx_k_im_median, sizeof(__pyx_k_im_median), 0, 0, 1, 1},
  {&__pyx_n_s_im_median_int, __pyx_k_im_median_int, sizeof(__pyx_k_im_median_int), 0, 0, 1, 1},
  {&__pyx_n_s_im_mz_1, __pyx_k_im_mz_1, sizeof(__pyx_k_im_mz_1), 0, 0, 1, 1},
  {&__pyx_n_s_im_mz_2, __pyx_k_im_mz_2, sizeof(__pyx_k_im_mz_2), 0, 0, 1, 1},
  {&__pyx_n_s_im_sorted, __pyx_k_im_sorted, sizeof(__pyx_k_im_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_im_to_check_fast, __pyx_k_im_to_check_fast, sizeof(__pyx_k_im_to_check_fast), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_include_factor, __pyx_k_include_factor, sizeof(__pyx_k_include_factor), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_int_val, __pyx_k_int_val, sizeof(__pyx_k_int_val), 0, 0, 1, 1},
  {&__pyx_kp_s_intensity_array, __pyx_k_intensity_array, sizeof(__pyx_k_intensity_array), 0, 0, 1, 0},
  {&__pyx_n_s_intensity_array_2, __pyx_k_intensity_array_2, sizeof(__pyx_k_intensity_array_2), 0, 0, 1, 1},
  {&__pyx_n_s_intensity_array_for_cos_corr, __pyx_k_intensity_array_for_cos_corr, sizeof(__pyx_k_intensity_array_for_cos_corr), 0, 0, 1, 1},
  {&__pyx_n_s_intensity_val, __pyx_k_intensity_val, sizeof(__pyx_k_intensity_val), 0, 0, 1, 1},
  {&__pyx_n_s_intersection, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_hill_idx, __pyx_k_isotope_hill_idx, sizeof(__pyx_k_isotope_hill_idx), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_idx, __pyx_k_isotope_idx, sizeof(__pyx_k_isotope_idx), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_number, __pyx_k_isotope_number, sizeof(__pyx_k_isotope_number), 0, 0, 1, 1},
  {&__pyx_n_s_isotopes, __pyx_k_isotopes, sizeof(__pyx_k_isotopes), 0, 0, 1, 1},
  {&__pyx_n_s_isotopes_list, __pyx_k_isotopes_list, sizeof(__pyx_k_isotopes_list), 0, 0, 1, 1},
  {&__pyx_n_s_isotopes_mass_accuracy, __pyx_k_isotopes_mass_accuracy, sizeof(__pyx_k_isotopes_mass_accuracy), 0, 0, 1, 1},
  {&__pyx_n_s_iter_candidates, __pyx_k_iter_candidates, sizeof(__pyx_k_iter_candidates), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_l_idx, __pyx_k_l_idx, sizeof(__pyx_k_l_idx), 0, 0, 1, 1},
  {&__pyx_n_s_l_r, __pyx_k_l_r, sizeof(__pyx_k_l_r), 0, 0, 1, 1},
  {&__pyx_n_s_last_idx, __pyx_k_last_idx, sizeof(__pyx_k_last_idx), 0, 0, 1, 1},
  {&__pyx_n_s_len_mz, __pyx_k_len_mz, sizeof(__pyx_k_len_mz), 0, 0, 1, 1},
  {&__pyx_n_s_local_isotopes_dict, __pyx_k_local_isotopes_dict, sizeof(__pyx_k_local_isotopes_dict), 0, 0, 1, 1},
  {&__pyx_n_s_local_res_dict, __pyx_k_local_res_dict, sizeof(__pyx_k_local_res_dict), 0, 0, 1, 1},
  {&__pyx_n_s_m_to_check, __pyx_k_m_to_check, sizeof(__pyx_k_m_to_check), 0, 0, 1, 1},
  {&__pyx_n_s_m_to_check_fast, __pyx_k_m_to_check_fast, sizeof(__pyx_k_m_to_check_fast), 0, 0, 1, 1},
  {&__pyx_kp_s_m_z_array, __pyx_k_m_z_array, sizeof(__pyx_k_m_z_array), 0, 0, 1, 0},
  {&__pyx_n_s_ma_vec, __pyx_k_ma_vec, sizeof(__pyx_k_ma_vec), 0, 0, 1, 1},
  {&__pyx_n_s_ma_vec_array, __pyx_k_ma_vec_array, sizeof(__pyx_k_ma_vec_array), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mass_diff_abs, __pyx_k_mass_diff_abs, sizeof(__pyx_k_mass_diff_abs), 0, 0, 1, 1},
  {&__pyx_n_s_mass_diff_ppm, __pyx_k_mass_diff_ppm, sizeof(__pyx_k_mass_diff_ppm), 0, 0, 1, 1},
  {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max_charge, __pyx_k_max_charge, sizeof(__pyx_k_max_charge), 0, 0, 1, 1},
  {&__pyx_kp_s_mean_inverse_reduced_ion_mobilit, __pyx_k_mean_inverse_reduced_ion_mobilit, sizeof(__pyx_k_mean_inverse_reduced_ion_mobilit), 0, 0, 1, 0},
  {&__pyx_n_s_meanfilt, __pyx_k_meanfilt, sizeof(__pyx_k_meanfilt), 0, 0, 1, 1},
  {&__pyx_n_s_min_charge, __pyx_k_min_charge, sizeof(__pyx_k_min_charge), 0, 0, 1, 1},
  {&__pyx_n_s_min_idx, __pyx_k_min_idx, sizeof(__pyx_k_min_idx), 0, 0, 1, 1},
  {&__pyx_n_s_min_idx_list, __pyx_k_min_idx_list, sizeof(__pyx_k_min_idx_list), 0, 0, 1, 1},
  {&__pyx_n_s_min_length_hill, __pyx_k_min_length_hill, sizeof(__pyx_k_min_length_hill), 0, 0, 1, 1},
  {&__pyx_n_s_min_val, __pyx_k_min_val, sizeof(__pyx_k_min_val), 0, 0, 1, 1},
  {&__pyx_n_s_minlh, __pyx_k_minlh, sizeof(__pyx_k_minlh), 0, 0, 1, 1},
  {&__pyx_n_s_mono_hills_intensity_list, __pyx_k_mono_hills_intensity_list, sizeof(__pyx_k_mono_hills_intensity_list), 0, 0, 1, 1},
  {&__pyx_n_s_mono_hills_scan_lists, __pyx_k_mono_hills_scan_lists, sizeof(__pyx_k_mono_hills_scan_lists), 0, 0, 1, 1},
  {&__pyx_kp_s_monoisotope_hill_idx, __pyx_k_monoisotope_hill_idx, sizeof(__pyx_k_monoisotope_hill_idx), 0, 0, 1, 0},
  {&__pyx_kp_s_monoisotope_idx, __pyx_k_monoisotope_idx, sizeof(__pyx_k_monoisotope_idx), 0, 0, 1, 0},
  {&__pyx_n_s_mult_val, __pyx_k_mult_val, sizeof(__pyx_k_mult_val), 0, 0, 1, 1},
  {&__pyx_n_s_mz_cur, __pyx_k_mz_cur, sizeof(__pyx_k_mz_cur), 0, 0, 1, 1},
  {&__pyx_n_s_mz_median, __pyx_k_mz_median, sizeof(__pyx_k_mz_median), 0, 0, 1, 1},
  {&__pyx_n_s_mz_median_int, __pyx_k_mz_median_int, sizeof(__pyx_k_mz_median_int), 0, 0, 1, 1},
  {&__pyx_n_s_mz_sorted, __pyx_k_mz_sorted, sizeof(__pyx_k_mz_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_mz_step, __pyx_k_mz_step, sizeof(__pyx_k_mz_step), 0, 0, 1, 1},
  {&__pyx_n_s_mz_tol, __pyx_k_mz_tol, sizeof(__pyx_k_mz_tol), 0, 0, 1, 1},
  {&__pyx_n_s_mz_val_tmp, __pyx_k_mz_val_tmp, sizeof(__pyx_k_mz_val_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_mzs_array, __pyx_k_mzs_array, sizeof(__pyx_k_mzs_array), 0, 0, 1, 1},
  {&__pyx_n_s_nIsotopes, __pyx_k_nIsotopes, sizeof(__pyx_k_nIsotopes), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_neutral_mass, __pyx_k_neutral_mass, sizeof(__pyx_k_neutral_mass), 0, 0, 1, 1},
  {&__pyx_n_s_new_index_list, __pyx_k_new_index_list, sizeof(__pyx_k_new_index_list), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nproc, __pyx_k_nproc, sizeof(__pyx_k_nproc), 0, 0, 1, 1},
  {&__pyx_n_s_numb, __pyx_k_numb, sizeof(__pyx_k_numb), 0, 0, 1, 1},
  {&__pyx_n_s_number_of_passed_isotopes, __pyx_k_number_of_passed_isotopes, sizeof(__pyx_k_number_of_passed_isotopes), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_orig_idx_array, __pyx_k_orig_idx_array, sizeof(__pyx_k_orig_idx_array), 0, 0, 1, 1},
  {&__pyx_n_s_orig_idx_val, __pyx_k_orig_idx_val, sizeof(__pyx_k_orig_idx_val), 0, 0, 1, 1},
  {&__pyx_n_s_paseftol, __pyx_k_paseftol, sizeof(__pyx_k_paseftol), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_prev_fast_dict, __pyx_k_prev_fast_dict, sizeof(__pyx_k_prev_fast_dict), 0, 0, 1, 1},
  {&__pyx_n_s_prev_fast_dict_im, __pyx_k_prev_fast_dict_im, sizeof(__pyx_k_prev_fast_dict_im), 0, 0, 1, 1},
  {&__pyx_n_s_prev_idx, __pyx_k_prev_idx, sizeof(__pyx_k_prev_idx), 0, 0, 1, 1},
  {&__pyx_n_s_process_hills, __pyx_k_process_hills, sizeof(__pyx_k_process_hills), 0, 0, 1, 1},
  {&__pyx_n_s_product, __pyx_k_product, sizeof(__pyx_k_product), 0, 0, 1, 1},
  {&__pyx_n_s_r_r, __pyx_k_r_r, sizeof(__pyx_k_r_r), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ready, __pyx_k_ready, sizeof(__pyx_k_ready), 0, 0, 1, 1},
  {&__pyx_n_s_recheck_idx, __pyx_k_recheck_idx, sizeof(__pyx_k_recheck_idx), 0, 0, 1, 1},
  {&__pyx_n_s_recheck_r_r, __pyx_k_recheck_r_r, sizeof(__pyx_k_recheck_r_r), 0, 0, 1, 1},
  {&__pyx_n_s_scan_id_val, __pyx_k_scan_id_val, sizeof(__pyx_k_scan_id_val), 0, 0, 1, 1},
  {&__pyx_n_s_scan_idx_array, __pyx_k_scan_idx_array, sizeof(__pyx_k_scan_idx_array), 0, 0, 1, 1},
  {&__pyx_n_s_scan_val, __pyx_k_scan_val, sizeof(__pyx_k_scan_val), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_smothed_intensity, __pyx_k_smothed_intensity, sizeof(__pyx_k_smothed_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_sorted_idx_array_child_process, __pyx_k_sorted_idx_array_child_process, sizeof(__pyx_k_sorted_idx_array_child_process), 0, 0, 1, 1},
  {&__pyx_n_s_sorted_idx_child_process, __pyx_k_sorted_idx_child_process, sizeof(__pyx_k_sorted_idx_child_process), 0, 0, 1, 1},
  {&__pyx_n_s_spec_idx, __pyx_k_spec_idx, sizeof(__pyx_k_spec_idx), 0, 0, 1, 1},
  {&__pyx_n_s_spec_mean_mass_accuracy, __pyx_k_spec_mean_mass_accuracy, sizeof(__pyx_k_spec_mean_mass_accuracy), 0, 0, 1, 1},
  {&__pyx_n_s_split_peaks, __pyx_k_split_peaks, sizeof(__pyx_k_split_peaks), 0, 0, 1, 1},
  {&__pyx_n_s_split_peaks_old, __pyx_k_split_peaks_old, sizeof(__pyx_k_split_peaks_old), 0, 0, 1, 1},
  {&__pyx_n_s_sqrt, __pyx_k_sqrt, sizeof(__pyx_k_sqrt), 0, 0, 1, 1},
  {&__pyx_n_s_suit_len, __pyx_k_suit_len, sizeof(__pyx_k_suit_len), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theor_total_sum, __pyx_k_theor_total_sum, sizeof(__pyx_k_theor_total_sum), 0, 0, 1, 1},
  {&__pyx_n_s_theoretical_list, __pyx_k_theoretical_list, sizeof(__pyx_k_theoretical_list), 0, 0, 1, 1},
  {&__pyx_n_s_thresh, __pyx_k_thresh, sizeof(__pyx_k_thresh), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_candidates, __pyx_k_tmp_candidates, sizeof(__pyx_k_tmp_candidates), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_hill_length, __pyx_k_tmp_hill_length, sizeof(__pyx_k_tmp_hill_length), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_im_array, __pyx_k_tmp_im_array, sizeof(__pyx_k_tmp_im_array), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_intensity, __pyx_k_tmp_intensity, sizeof(__pyx_k_tmp_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_mz_array, __pyx_k_tmp_mz_array, sizeof(__pyx_k_tmp_mz_array), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_orig_idx, __pyx_k_tmp_orig_idx, sizeof(__pyx_k_tmp_orig_idx), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_scans, __pyx_k_tmp_scans, sizeof(__pyx_k_tmp_scans), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_scans_list, __pyx_k_tmp_scans_list, sizeof(__pyx_k_tmp_scans_list), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_scans_set, __pyx_k_tmp_scans_set, sizeof(__pyx_k_tmp_scans_set), 0, 0, 1, 1},
  {&__pyx_n_s_tmp_val, __pyx_k_tmp_val, sizeof(__pyx_k_tmp_val), 0, 0, 1, 1},
  {&__pyx_n_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 1},
  {&__pyx_n_s_total_mass_diff, __pyx_k_total_mass_diff, sizeof(__pyx_k_total_mass_diff), 0, 0, 1, 1},
  {&__pyx_n_s_total_num_hills, __pyx_k_total_num_hills, sizeof(__pyx_k_total_num_hills), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_weights, __pyx_k_weights, sizeof(__pyx_k_weights), 0, 0, 1, 1},
  {&__pyx_n_s_window_width, __pyx_k_window_width, sizeof(__pyx_k_window_width), 0, 0, 1, 1},
  {&__pyx_n_s_z, __pyx_k_z, sizeof(__pyx_k_z), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 35, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 98, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 334, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 434, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 945, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "biosaur2/cutils.pyx":98
 *     ready = []
 * 
 *     charges = list(range(min_charge, max_charge + 1, 1)[::-1])             # <<<<<<<<<<<<<<
 * 
 *     for idx_1 in sorted_idx_child_process:
 */
  __pyx_slice_ = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice_)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);

  /* "biosaur2/cutils.pyx":244
 * 
 *     if not top:
 *         return 0, 0             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
  __pyx_tuple__2 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "biosaur2/cutils.pyx":311
 *     ma_vec_array = (cumsum_vec[window_width:] -
 *               cumsum_vec[:-window_width]) / window_width
 *     ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]             # <<<<<<<<<<<<<<
 *     return ma_vec
 * 
 */
  __pyx_slice__3 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__3);
  __Pyx_GIVEREF(__pyx_slice__3);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "biosaur2/cutils.pyx":16
 * 
 * charge_ban_map = {
 *     8: (1, 2, 4, ),             # <<<<<<<<<<<<<<
 *     7: (1, ),
 *     6: (1, 2, 3, ),
 */
  __pyx_tuple__6 = PyTuple_Pack(3, __pyx_int_1, __pyx_int_2, __pyx_int_4); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "biosaur2/cutils.pyx":17
 * charge_ban_map = {
 *     8: (1, 2, 4, ),
 *     7: (1, ),             # <<<<<<<<<<<<<<
 *     6: (1, 2, 3, ),
 *     5: (1, ),
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "biosaur2/cutils.pyx":18
 *     8: (1, 2, 4, ),
 *     7: (1, ),
 *     6: (1, 2, 3, ),             # <<<<<<<<<<<<<<
 *     5: (1, ),
 *     4: (1, 2, ),
 */
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_1, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "biosaur2/cutils.pyx":20
 *     6: (1, 2, 3, ),
 *     5: (1, ),
 *     4: (1, 2, ),             # <<<<<<<<<<<<<<
 *     3: (1, ),
 *     2: (1, ),
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "biosaur2/cutils.pyx":30
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 */
  __pyx_tuple__10 = PyTuple_Pack(8, __pyx_n_s_hills_dict, __pyx_n_s_idx_1, __pyx_n_s_hill_idict_1, __pyx_n_s_scan_id_val, __pyx_n_s_intensity_val, __pyx_n_s_hill_sqrt_of_i_1, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_get_and_calc_values_for_cos_corr_2, 30, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":50
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 */
  __pyx_tuple__12 = PyTuple_Pack(6, __pyx_n_s_hills_dict, __pyx_n_s_idx_1, __pyx_n_s_hill_intensity_apex_1, __pyx_n_s_hill_scan_apex_1, __pyx_n_s_int_val, __pyx_n_s_scan_val); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_get_and_calc_apex_intensity_and, 50, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 50, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":72
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation(set hill_scans_1, dict hill_idict_1, float hill_sqrt_of_i_1, set hill_scans_2, dict hill_idict_2, float hill_sqrt_of_i_2):             # <<<<<<<<<<<<<<
 * 
 *     cdef int i
 */
  __pyx_tuple__14 = PyTuple_Pack(8, __pyx_n_s_hill_scans_1, __pyx_n_s_hill_idict_1, __pyx_n_s_hill_sqrt_of_i_1, __pyx_n_s_hill_scans_2, __pyx_n_s_hill_idict_2, __pyx_n_s_hill_sqrt_of_i_2, __pyx_n_s_i, __pyx_n_s_top); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(6, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_cos_correlation, 72, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 72, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":86
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_initial_isotopes(dict hills_dict, float isotopes_mass_accuracy, list isotopes_list, dict a, int min_charge, int max_charge, float mz_step, bint paseftol, float faims_val, list sorted_idx_child_process):             # <<<<<<<<<<<<<<
 * 
 *     cdef list ready, charges, hill_scans_1_list, hill_scans_2_list, candidates, tmp_candidates
 */
  __pyx_tuple__17 = PyTuple_Pack(61, __pyx_n_s_hills_dict, __pyx_n_s_isotopes_mass_accuracy, __pyx_n_s_isotopes_list, __pyx_n_s_a, __pyx_n_s_min_charge, __pyx_n_s_max_charge, __pyx_n_s_mz_step, __pyx_n_s_paseftol, __pyx_n_s_faims_val, __pyx_n_s_sorted_idx_child_process, __pyx_n_s_ready, __pyx_n_s_charges, __pyx_n_s_hill_scans_1_list, __pyx_n_s_hill_scans_2_list, __pyx_n_s_candidates, __pyx_n_s_tmp_candidates, __pyx_n_s_idx_1, __pyx_n_s_hill_idx_1, __pyx_n_s_idx_2, __pyx_n_s_hill_idx_2, __pyx_n_s_im_to_check_fast, __pyx_n_s_hill_scans_1_list_first, __pyx_n_s_hill_scans_1_list_last, __pyx_n_s_hill_scans_2_list_first, __pyx_n_s_hill_scans_2_list_last, __pyx_n_s_charge, __pyx_n_s_isotope_number, __pyx_n_s_m_to_check_fast, __pyx_n_s_im_mz_1, __pyx_n_s_hill_mz_1, __pyx_n_s_im_mz_2, __pyx_n_s_hill_mz_2, __pyx_n_s_m_to_check, __pyx_n_s_mass_diff_abs, __pyx_n_s_cos_cor_RT, __pyx_n_s_hill_sqrt_of_i_1, __pyx_n_s_hill_sqrt_of_i_2, __pyx_n_s_banned_charges, __pyx_n_s_hill_idict_1, __pyx_n_s_hill_idict_2, __pyx_n_s_local_isotopes_dict, __pyx_n_s_hill_scans_1, __pyx_n_s_hill_scans_2, __pyx_n_s_mz_tol, __pyx_n_s__16, __pyx_n_s_neutral_mass, __pyx_n_s_tmp_intensity, __pyx_n_s_hill_intensity_apex_1, __pyx_n_s_hill_scan_apex_1, __pyx_n_s_mono_hills_scan_lists, __pyx_n_s_mono_hills_intensity_list, __pyx_n_s_all_theoretical_int, __pyx_n_s_iter_candidates, __pyx_n_s_all_exp_intensity, __pyx_n_s_hill_intensity_apex_2, __pyx_n_s_hill_scan_apex_2, __pyx_n_s_cos_corr, __pyx_n_s_number_of_passed_isotopes, __pyx_n_s_local_res_dict, __pyx_n_s_ch_v, __pyx_n_s_z); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(10, 0, 61, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_get_initial_isotopes, 86, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 86, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":228
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation_new(list theoretical_list, list experimental_list):             # <<<<<<<<<<<<<<
 *     cdef float theor_total_sum, top, i1, i2, bottom, numb, averagineExplained
 *     cdef int suit_len
 */
  __pyx_tuple__19 = PyTuple_Pack(10, __pyx_n_s_theoretical_list, __pyx_n_s_experimental_list, __pyx_n_s_theor_total_sum, __pyx_n_s_top, __pyx_n_s_i1, __pyx_n_s_i2, __pyx_n_s_bottom, __pyx_n_s_numb, __pyx_n_s_averagineExplained, __pyx_n_s_suit_len); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_cos_correlation_new, 228, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 228, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":257
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def checking_cos_correlation_for_carbon(list theoretical_list, list experimental_list, float thresh):             # <<<<<<<<<<<<<<
 * 
 *     cdef float averagineCorrelation, averagineExplained, best_cor
 */
  __pyx_tuple__21 = PyTuple_Pack(8, __pyx_n_s_theoretical_list, __pyx_n_s_experimental_list, __pyx_n_s_thresh, __pyx_n_s_averagineCorrelation, __pyx_n_s_averagineExplained, __pyx_n_s_best_cor, __pyx_n_s_best_pos, __pyx_n_s_pos); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_checking_cos_correlation_for_car, 257, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 257, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":286
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_fast_dict(np.ndarray mz_sorted, float mz_step, list basic_id_sorted):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict fast_dict
 */
  __pyx_tuple__23 = PyTuple_Pack(7, __pyx_n_s_mz_sorted, __pyx_n_s_mz_step, __pyx_n_s_basic_id_sorted, __pyx_n_s_fast_dict, __pyx_n_s_fast_array, __pyx_n_s_idx, __pyx_n_s_fm); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(3, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_get_fast_dict, 286, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 286, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":305
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def meanfilt(list data, int window_width):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray cumsum_vec, ma_vec_array
 *     cdef list ma_vec
 */
  __pyx_tuple__25 = PyTuple_Pack(5, __pyx_n_s_data, __pyx_n_s_window_width, __pyx_n_s_cumsum_vec, __pyx_n_s_ma_vec_array, __pyx_n_s_ma_vec); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_meanfilt, 305, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 305, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":318
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks(dict hills_dict, list data_for_analyse_tmp, dict args, dict counter_hills_idx, list sorted_idx_child_process, np.ndarray sorted_idx_array_child_process, int nproc, int checked_id):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */
  __pyx_tuple__27 = PyTuple_Pack(37, __pyx_n_s_hills_dict, __pyx_n_s_data_for_analyse_tmp, __pyx_n_s_args, __pyx_n_s_counter_hills_idx, __pyx_n_s_sorted_idx_child_process, __pyx_n_s_sorted_idx_array_child_process, __pyx_n_s_nproc, __pyx_n_s_checked_id, __pyx_n_s_hillValleyFactor, __pyx_n_s_min_val, __pyx_n_s_mult_val, __pyx_n_s_min_length_hill, __pyx_n_s_idx_start, __pyx_n_s_idx_end, __pyx_n_s_cur_new_idx, __pyx_n_s_idx_1, __pyx_n_s_hill_idx, __pyx_n_s_hill_length, __pyx_n_s_c_len, __pyx_n_s_l_idx, __pyx_n_s_idx_sort, __pyx_n_s_tmp_scans, __pyx_n_s_tmp_orig_idx, __pyx_n_s_new_index_list, __pyx_n_s_tmp_intensity, __pyx_n_s_smothed_intensity, __pyx_n_s_min_idx_list, __pyx_n_s_recheck_r_r, __pyx_n_s_idx, __pyx_n_s_l_r, __pyx_n_s_r_r, __pyx_n_s_include_factor, __pyx_n_s_min_idx, __pyx_n_s_end_idx, __pyx_n_s_recheck_idx, __pyx_n_s_orig_idx_val, __pyx_n_s_scan_val); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(8, 0, 37, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_split_peaks, 318, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 318, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":396
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks_old(dict hills_dict, list data_for_analyse_tmp, dict args):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */
  __pyx_tuple__29 = PyTuple_Pack(34, __pyx_n_s_hills_dict, __pyx_n_s_data_for_analyse_tmp, __pyx_n_s_args, __pyx_n_s_hillValleyFactor, __pyx_n_s_min_val, __pyx_n_s_mult_val, __pyx_n_s_min_length_hill, __pyx_n_s_idx_start, __pyx_n_s_idx_end, __pyx_n_s_cur_new_idx, __pyx_n_s_idx_1, __pyx_n_s_hill_idx, __pyx_n_s_hill_length, __pyx_n_s_c_len, __pyx_n_s_l_idx, __pyx_n_s_tmp_hill_length, __pyx_n_s_idx_minl, __pyx_n_s_idx_sort, __pyx_n_s_tmp_scans, __pyx_n_s_tmp_orig_idx, __pyx_n_s_tmp_intensity, __pyx_n_s_smothed_intensity, __pyx_n_s_min_idx_list, __pyx_n_s_recheck_r_r, __pyx_n_s_counter_hills_idx, __pyx_n_s_idx, __pyx_n_s_l_r, __pyx_n_s_r_r, __pyx_n_s_include_factor, __pyx_n_s_min_idx, __pyx_n_s_end_idx, __pyx_n_s_recheck_idx, __pyx_n_s_orig_idx_val, __pyx_n_s_scan_val); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(3, 0, 34, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_split_peaks_old, 396, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 396, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":494
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def detect_hills(list data_for_analyse_tmp, dict args, float mz_step, bint paseftol, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict hills_dict, prev_fast_dict, z
 */
  __pyx_tuple__31 = PyTuple_Pack(42, __pyx_n_s_data_for_analyse_tmp, __pyx_n_s_args, __pyx_n_s_mz_step, __pyx_n_s_paseftol, __pyx_n_s_dia, __pyx_n_s_hills_dict, __pyx_n_s_prev_fast_dict, __pyx_n_s_z, __pyx_n_s_total_num_hills, __pyx_n_s_total_mass_diff, __pyx_n_s_spec_mean_mass_accuracy, __pyx_n_s_basic_id_sorted, __pyx_n_s_all_idx, __pyx_n_s_all_prevs, __pyx_n_s_banned_prev_idx_set, __pyx_n_s_hill_mass_accuracy, __pyx_n_s_best_intensity, __pyx_n_s_mz_cur, __pyx_n_s_cur_intensity, __pyx_n_s_cur_mass_diff_with_sign, __pyx_n_s_cur_mass_diff, __pyx_n_s_best_mass_diff, __pyx_n_s_last_idx, __pyx_n_s_prev_idx, __pyx_n_s_spec_idx, __pyx_n_s_len_mz, __pyx_n_s_idx, __pyx_n_s_fm, __pyx_n_s_fi, __pyx_n_s_best_idx_prev, __pyx_n_s_idx_prev, __pyx_n_s_idx_for_sort, __pyx_n_s_mz_sorted, __pyx_n_s_im_sorted, __pyx_n_s_flag1, __pyx_n_s_flag2, __pyx_n_s_flag3, __pyx_n_s_prev_fast_dict_im, __pyx_n_s_fast_dict_im, __pyx_n_s_fast_array_im, __pyx_n_s_fast_dict, __pyx_n_s_fast_array); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(5, 0, 42, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_detect_hills, 494, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 494, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":605
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def process_hills(dict hills_dict, list data_for_analyse_tmp, float mz_step, bint paseftol, dict args, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict
 */
  __pyx_tuple__33 = PyTuple_Pack(34, __pyx_n_s_hills_dict, __pyx_n_s_data_for_analyse_tmp, __pyx_n_s_mz_step, __pyx_n_s_paseftol, __pyx_n_s_args, __pyx_n_s_dia, __pyx_n_s_dict, __pyx_n_s_mz_median, __pyx_n_s_i_sum_tmp, __pyx_n_s_mz_val_tmp, __pyx_n_s_i_val_tmp, __pyx_n_s_im_median, __pyx_n_s_tmp_intensity, __pyx_n_s_tmp_mz_array, __pyx_n_s_tmp_im_array, __pyx_n_s_tmp_scans_list, __pyx_n_s_tmp_scans, __pyx_n_s_tmp_orig_idx, __pyx_n_s_tmp_scans_set, __pyx_n_s_tmp_val, __pyx_n_s_min_length_hill, __pyx_n_s_idx_start, __pyx_n_s_idx_end, __pyx_n_s_idx_1, __pyx_n_s_hill_idx, __pyx_n_s_hill_length, __pyx_n_s_mz_median_int, __pyx_n_s_im_median_int, __pyx_n_s_tmp_hill_length, __pyx_n_s_idx_minl, __pyx_n_s_counter_hills_idx, __pyx_n_s_idx_sort, __pyx_n_s_orig_idx_val, __pyx_n_s_scan_val); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(6, 0, 34, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_biosaur2_cutils_pyx, __pyx_n_s_process_hills, 605, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* InitThreads.init */
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0
PyEval_InitThreads();
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PySet_Type_intersection.type = (PyObject*)&PySet_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_6 = PyFloat_FromDouble(0.6); if (unlikely(!__pyx_float_0_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1e6 = PyFloat_FromDouble(1e6); if (unlikely(!__pyx_float_1e6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_8biosaur2_6cutils_charge_ban_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr.tp_dictoffset && __pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr = &__pyx_type_8biosaur2_6cutils___pyx_scope_struct__get_and_calc_values_for_cos_corr;
  if (PyType_Ready(&__pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8biosaur2_6cutils___pyx_scope_struct_1_genexpr = &__pyx_type_8biosaur2_6cutils___pyx_scope_struct_1_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcutils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcutils(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cutils(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cutils(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cutils(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cutils' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cutils(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cutils", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_biosaur2__cutils) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "biosaur2.cutils")) {
      if (unlikely(PyDict_SetItemString(modules, "biosaur2.cutils", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "biosaur2/cutils.pyx":4
 * from cython.parallel cimport prange, parallel
 * cimport openmp
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * import itertools
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":6
 * import numpy as np
 * cimport numpy as np
 * import itertools             # <<<<<<<<<<<<<<
 * import math
 * from collections import Counter, defaultdict
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_itertools, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":7
 * cimport numpy as np
 * import itertools
 * import math             # <<<<<<<<<<<<<<
 * from collections import Counter, defaultdict
 * from copy import copy
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_math, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":8
 * import itertools
 * import math
 * from collections import Counter, defaultdict             # <<<<<<<<<<<<<<
 * from copy import copy
 * 
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Counter);
  __Pyx_GIVEREF(__pyx_n_s_Counter);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Counter);
  __Pyx_INCREF(__pyx_n_s_defaultdict);
  __Pyx_GIVEREF(__pyx_n_s_defaultdict);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_defaultdict);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Counter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Counter, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultdict, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "biosaur2/cutils.pyx":9
 * import math
 * from collections import Counter, defaultdict
 * from copy import copy             # <<<<<<<<<<<<<<
 * 
 * np.import_array()
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_copy);
  __Pyx_GIVEREF(__pyx_n_s_copy);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_copy);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_copy, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copy, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":11
 * from copy import copy
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * cdef dict charge_ban_map
 */
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 11, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":16
 * 
 * charge_ban_map = {
 *     8: (1, 2, 4, ),             # <<<<<<<<<<<<<<
 *     7: (1, ),
 *     6: (1, 2, 3, ),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_8, __pyx_tuple__6) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":17
 * charge_ban_map = {
 *     8: (1, 2, 4, ),
 *     7: (1, ),             # <<<<<<<<<<<<<<
 *     6: (1, 2, 3, ),
 *     5: (1, ),
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_7, __pyx_tuple__7) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":18
 *     8: (1, 2, 4, ),
 *     7: (1, ),
 *     6: (1, 2, 3, ),             # <<<<<<<<<<<<<<
 *     5: (1, ),
 *     4: (1, 2, ),
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_6, __pyx_tuple__8) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":19
 *     7: (1, ),
 *     6: (1, 2, 3, ),
 *     5: (1, ),             # <<<<<<<<<<<<<<
 *     4: (1, 2, ),
 *     3: (1, ),
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_5, __pyx_tuple__7) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":20
 *     6: (1, 2, 3, ),
 *     5: (1, ),
 *     4: (1, 2, ),             # <<<<<<<<<<<<<<
 *     3: (1, ),
 *     2: (1, ),
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_4, __pyx_tuple__9) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":21
 *     5: (1, ),
 *     4: (1, 2, ),
 *     3: (1, ),             # <<<<<<<<<<<<<<
 *     2: (1, ),
 *     1: (1, ),
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_3, __pyx_tuple__7) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":22
 *     4: (1, 2, ),
 *     3: (1, ),
 *     2: (1, ),             # <<<<<<<<<<<<<<
 *     1: (1, ),
 * }
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_2, __pyx_tuple__7) < 0) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "biosaur2/cutils.pyx":23
 *     3: (1, ),
 *     2: (1, ),
 *     1: (1, ),             # <<<<<<<<<<<<<<
 * }
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_1, __pyx_tuple__7) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_8biosaur2_6cutils_charge_ban_map);
  __Pyx_DECREF_SET(__pyx_v_8biosaur2_6cutils_charge_ban_map, ((PyObject*)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":30
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_idict_1 = hills_dict['hills_idict'][idx_1]
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_1get_and_calc_values_for_cos_corr, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_and_calc_values_for_cos_corr_2, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":50
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1):             # <<<<<<<<<<<<<<
 * 
 *     hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_3get_and_calc_apex_intensity_and_scan, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_and_calc_apex_intensity_and, __pyx_t_1) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":72
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation(set hill_scans_1, dict hill_idict_1, float hill_sqrt_of_i_1, set hill_scans_2, dict hill_idict_2, float hill_sqrt_of_i_2):             # <<<<<<<<<<<<<<
 * 
 *     cdef int i
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_5cos_correlation, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cos_correlation, __pyx_t_1) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":86
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_initial_isotopes(dict hills_dict, float isotopes_mass_accuracy, list isotopes_list, dict a, int min_charge, int max_charge, float mz_step, bint paseftol, float faims_val, list sorted_idx_child_process):             # <<<<<<<<<<<<<<
 * 
 *     cdef list ready, charges, hill_scans_1_list, hill_scans_2_list, candidates, tmp_candidates
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_7get_initial_isotopes, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_initial_isotopes, __pyx_t_1) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":228
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def cos_correlation_new(list theoretical_list, list experimental_list):             # <<<<<<<<<<<<<<
 *     cdef float theor_total_sum, top, i1, i2, bottom, numb, averagineExplained
 *     cdef int suit_len
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_9cos_correlation_new, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cos_correlation_new, __pyx_t_1) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":257
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def checking_cos_correlation_for_carbon(list theoretical_list, list experimental_list, float thresh):             # <<<<<<<<<<<<<<
 * 
 *     cdef float averagineCorrelation, averagineExplained, best_cor
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_11checking_cos_correlation_for_carbon, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_checking_cos_correlation_for_car, __pyx_t_1) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":286
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def get_fast_dict(np.ndarray mz_sorted, float mz_step, list basic_id_sorted):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict fast_dict
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_13get_fast_dict, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_fast_dict, __pyx_t_1) < 0) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":305
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def meanfilt(list data, int window_width):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray cumsum_vec, ma_vec_array
 *     cdef list ma_vec
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_15meanfilt, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_meanfilt, __pyx_t_1) < 0) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":318
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks(dict hills_dict, list data_for_analyse_tmp, dict args, dict counter_hills_idx, list sorted_idx_child_process, np.ndarray sorted_idx_array_child_process, int nproc, int checked_id):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_17split_peaks, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_split_peaks, __pyx_t_1) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":396
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def split_peaks_old(dict hills_dict, list data_for_analyse_tmp, dict args):             # <<<<<<<<<<<<<<
 * 
 *     cdef float hillValleyFactor, min_val, mult_val
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_19split_peaks_old, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_split_peaks_old, __pyx_t_1) < 0) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":494
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def detect_hills(list data_for_analyse_tmp, dict args, float mz_step, bint paseftol, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict hills_dict, prev_fast_dict, z
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_21detect_hills, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_detect_hills, __pyx_t_1) < 0) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":605
 * @cython.boundscheck(False)
 * @cython.wraparound(True)
 * def process_hills(dict hills_dict, list data_for_analyse_tmp, float mz_step, bint paseftol, dict args, bint dia=False):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8biosaur2_6cutils_23process_hills, NULL, __pyx_n_s_biosaur2_cutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_process_hills, __pyx_t_1) < 0) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "biosaur2/cutils.pyx":1
 * cimport cython             # <<<<<<<<<<<<<<
 * from cython.parallel cimport prange, parallel
 * cimport openmp
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../opt/hostedtoolcache/Python/3.10.4/x64/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init biosaur2.cutils", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init biosaur2.cutils");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
            if (unlikely(b == -1 && ((unsigned long)a) == 0-(unsigned long)a))
                return PyInt_Type.tp_as_number->nb_floor_divide(op1, op2);
            else {
                long q, r;
                q = a / b;
                r = a - q*b;
                q -= ((r != 0) & ((r ^ b) < 0));
                x = q;
            }
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_floor_divide(op1, op2);
            }
        }
                __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
                {
                    long q, r;
                    q = a / b;
                    r = a - q*b;
                    q -= ((r != 0) & ((r ^ b) < 0));
                    x = q;
                }
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                {
                    PY_LONG_LONG q, r;
                    q = lla / llb;
                    r = lla - q*llb;
                    q -= ((r != 0) & ((r ^ llb) < 0));
                    llx = q;
                }
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceFloorDivide : PyNumber_FloorDivide)(op1, op2);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
